/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more Details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * Skeleton Data Structure to AVS Geometry Transform
 *
 */

static char skelGeom_Id[] = "%W%\t%G%";

#include <stdio.h>

#include <avs/avs.h>
#include <avs/avs_data.h>
#include <avs/geom.h>
#include <avs/flow.h>

#include "Sy_Avs.h"
#include "Sy_Skeleton.h"

extern enum SyRenderMethod  _sy_avs_render_method;
extern enum SyTriangleDecor _sy_avs_triangle_decor;
extern float              _sy_avs_triangle_height;
extern float              _sy_avs_value_limit;
extern enum SyDataFocus     _sy_avs_data_focus;

static GEOMobj*      obj;
static SySkelDAC*     vsk;
static SyIpoDAC*       ipd;

static SyAvsVector*    vertex;
static SyAvsNormal*    normal;
static int           vindex, nindex, 
                     vdim, ndim;

static void Sy_AvsInsertVertex(SyAvsVector *vertex, int *vindex, float x, float y, float z);
static void Sy_AvsInsertTriStrip(SyAvsVector *vertex, int *vindex, float x, float y, float z1, float z2);

static void Sy_AvsSkel2GeomInit(void);
static void Sy_AvsSkel2GeomPrep(void);
static void Sy_AvsSkelInsert   (void);
static void Sy_AvsSkel2GeomTerm(void);
static void Sy_AvsSkel2GeomExit(void);
static void Sy_AvsCutEdge      (void);

void Sy_AvsSkel2Geom (GEOMobj *avs_obj, SySkelDAC *skel_vsk, 
		      SyIpoDAC *skel_ipd, float thres)
{
  SyWalktask wt;
  
  obj  = avs_obj;
  vsk  = skel_vsk;
  ipd  = skel_ipd;

  Sy_ResetWalktask (&wt);

  wt.user_function[WALK] = (void(*)())Sy_AvsSkelInsert;
  wt.user_function[EOSK] = (void(*)())Sy_AvsCutEdge;

  wt.user_function[ROT ] = ((_sy_avs_triangle_decor == RESIDUAL) || 
			    (_sy_avs_triangle_decor == BIRTH_VECTOR)) ?
			      (void(*)())NULL : wt.user_function[WALK];

  wt.user_function[INIT] = (void(*)())Sy_AvsSkel2GeomInit;
  wt.user_function[TRAV] = (void(*)())Sy_AvsSkel2GeomPrep;
  wt.user_function[TERM] = (void(*)())Sy_AvsSkel2GeomTerm;

  wt.user_function[FEXIT] = (void(*)())Sy_AvsSkel2GeomExit;

  Sy_SkeletonTraverse (vsk, ipd, thres, 0, &wt);
}
 
static void Sy_AvsSkel2GeomInit (void)
{

  int max_tri;
  vertex = (SyAvsVector*) NULL;
  normal = (SyAvsNormal*) NULL;
  vindex = nindex = 0;
  
  switch(_sy_avs_triangle_decor){
  case CONSTANT:
  case RADIUS:
    max_tri = 8;
    break;
  case BIRTH_VECTOR:
    max_tri = 12;
    break;
  case RESIDUAL:
    max_tri = 4;
    break;
  }
  switch(_sy_avs_render_method){
  case POLYLINE: 
    vertex = (SyAvsVector*) Sy_Malloc(4*(vsk->edim+1)*sizeof(SyAvsVector));
    vdim   = 4*(vsk->edim+1);
    break;
  case POLYTRIANGLE:
    vertex = (SyAvsVector*) Sy_Malloc(max_tri*(vsk->edim+1)*sizeof(SyAvsVector));
    normal = (SyAvsNormal*) Sy_Malloc(max_tri*(vsk->edim+1)*sizeof(SyAvsNormal));
    vdim   = ndim = max_tri*(vsk->edim+1);
    break;
  }
}

static void Sy_AvsSkel2GeomPrep (void)
{
  SyVertex  vhead, vtail;
  float   head_residual, tail_residual,
          head_radius, tail_radius;
  float   x0, y0, x1, y1;
  int     is_cut;

  vindex = 0;
  nindex = 0;

  is_cut = 0; /* Sy_IsResidualCut (syWstate.vsk, syWstate.ipd, syWstate.start.this_edge, syWstate.thres); */

  if (is_cut){

    Sy_ResidualCut (syWstate.vsk, syWstate.start.this_edge, syWstate.thres, &vtail, &vhead,
		      &tail_radius, &tail_residual, &head_radius, &head_residual);

    x0 = vtail.xden/vtail.num;
    y0 = vtail.yden/vtail.num;
  
    x1 = vhead.xden/vhead.num;
    y1 = vhead.yden/vhead.num;

    fprintf (stderr, "cut: x = %f y = %f \n", x1, y1); 
    
    switch(_sy_avs_render_method){
    case POLYLINE:      
      Sy_AvsInsertVertex (vertex, &vindex, x1, y1, 0.0); 
      break;
    case POLYTRIANGLE: 
      switch (_sy_avs_triangle_decor){
      case CONSTANT:
	Sy_AvsInsertTriStrip (vertex, &vindex, x1, y1, 0.0, _sy_avs_triangle_height);
	break;
      case RADIUS:
	switch (_sy_avs_data_focus){
	case FOREGROUND: if (!syWstate.foreground) head_radius = _sy_avs_triangle_height; break;
	case BACKGROUND: if (syWstate.foreground ) head_radius = _sy_avs_triangle_height; break;
	case ALLDATA: break;
	}
	head_radius = min(head_radius, _sy_avs_value_limit);
	Sy_AvsInsertTriStrip (vertex, &vindex, x1, y1, 0.0, head_radius);
	break;
      case RESIDUAL:
	switch (_sy_avs_data_focus){
	case FOREGROUND: if (!syWstate.foreground) 
	  head_residual = tail_residual = _sy_avs_triangle_height; break;
	case BACKGROUND: if (syWstate.foreground )
	  head_residual = tail_residual = _sy_avs_triangle_height; break;
	case ALLDATA: break;
	}
	head_residual = min(head_residual, _sy_avs_value_limit);
	tail_residual = min(tail_residual, _sy_avs_value_limit);

	fprintf (stderr, "hr = %f, tr = %f\n", head_residual, tail_residual); 
	Sy_AvsInsertTriStrip (vertex, &vindex, x1, y1, 0.0, head_residual);
	Sy_AvsInsertTriStrip (vertex, &vindex, x0, y0, 0.0, tail_residual);
	break;
      case BIRTH_VECTOR:
	break;
      }
      break;
    }
  }
}

static void Sy_AvsSkelInsert (void)
{
  SyVertex vhead, vtail;
  float x, y, x0, y0, x1, y1;
  float radius, tail_residual, head_residual;
  int   tail_vertex, head_vertex, left_point, right_point;

  vhead  = syWstate.vsk->vertex[syWstate.current.head_vertex];
  x = vhead.xden/vhead.num;
  y = vhead.yden/vhead.num;

  /* fprintf (stderr, "x = %f y = %f \n", x, y); */

/*
  syWstate.foreground = 
    Sy_SkeletonType (syWstate.vsk, syWstate.ipd, 
		     syWstate.current.head_vertex, syWstate.current.next_edge);*/

  switch(_sy_avs_render_method){
  case POLYLINE: 
    if (syWstate.foreground) Sy_AvsInsertVertex (vertex, &vindex, x, y, 0.0); 
    break;
  case POLYTRIANGLE:
    if (!syWstate.foreground) break;
    switch (_sy_avs_triangle_decor){
    case CONSTANT:
      Sy_AvsInsertTriStrip (vertex, &vindex, x, y, 0.0, _sy_avs_triangle_height);
      break;
    case RADIUS:
      radius = syWstate.vsk->regul[syWstate.current.head_vertex].radius;
      switch (_sy_avs_data_focus){
      case FOREGROUND: if (!syWstate.foreground) radius = _sy_avs_triangle_height; break;
      case BACKGROUND: if (syWstate.foreground ) radius = _sy_avs_triangle_height; break;
      case ALLDATA: break;
      }
      radius = min(radius, _sy_avs_value_limit);
      /* fprintf (stderr, "rad=%f\n", radius); */
      Sy_AvsInsertTriStrip (vertex, &vindex, x, y, 0.0, radius);
      break;
    case RESIDUAL:
      left_point    = syWstate.vsk->edge[syWstate.last.next_edge].left_point;
      right_point   = syWstate.vsk->edge[syWstate.last.next_edge].right_point;
      
/*      if ((left_point < 3) || (right_point < 3)) break; */

      tail_residual = syWstate.vsk->eresid[syWstate.last.next_edge].tail_residual;
      head_residual = syWstate.vsk->eresid[syWstate.last.next_edge].head_residual;
      
      switch (_sy_avs_data_focus){
      case FOREGROUND: if (!syWstate.foreground) 
	head_residual = tail_residual = _sy_avs_triangle_height; break;
      case BACKGROUND: if (syWstate.foreground )
	head_residual = tail_residual = _sy_avs_triangle_height; break;
      case ALLDATA: break;
      }
      head_residual = min(head_residual, _sy_avs_value_limit);
      tail_residual = min(tail_residual, _sy_avs_value_limit);

      tail_vertex   = syWstate.vsk->edge[syWstate.last.next_edge].tail_vertex;
      head_vertex   = syWstate.vsk->edge[syWstate.last.next_edge].head_vertex;
      vhead         = syWstate.vsk->vertex[head_vertex];
      vtail         = syWstate.vsk->vertex[tail_vertex];
      x0            = vtail.xden/vtail.num;
      y0            = vtail.yden/vtail.num;
      x1            = vhead.xden/vhead.num;
      y1            = vhead.yden/vhead.num;
      /* fprintf (stderr, "x0 = %f y0 = %f x1 = %f y1 = %f\n", x0, y0, x1, y1);
      fprintf (stderr, "hr = %f, tr = %f\n", head_residual, tail_residual); */
      Sy_AvsInsertTriStrip (vertex, &vindex, x0, y0, 0.0, tail_residual);
      Sy_AvsInsertTriStrip (vertex, &vindex, x1, y1, 0.0, head_residual);
    }
    break;
  case BIRTH_VECTOR:
    break;
  }
}

static void Sy_AvsCutEdge (void)
{
  SyVertex  vtmp;
  SyVertex  vhead, vtail;
  float   x0, y0, x1, y1;
  float   tail_radius, head_radius,
          tail_residual, head_residual;
  float   tri_heigth;
  int     is_cut, retval;

  int     lp, rp;

  lp = syWstate.vsk->edge[syWstate.current.next_edge].left_point;
  rp = syWstate.vsk->edge[syWstate.current.next_edge].right_point;

  /* fprintf (stderr, "left p. = %d right p. = %d \n", lp, rp);  */

  is_cut = 0; /* Sy_IsResidualCut (syWstate.vsk, syWstate.ipd, syWstate.current.next_edge, syWstate.thres); */

  if (is_cut){

    retval = Sy_ResidualCut (syWstate.vsk, syWstate.current.next_edge, syWstate.thres, &vtail, &vhead,
			       &tail_radius, &tail_residual, 
			       &head_radius, &head_residual);

    x0 = vtail.xden/vtail.num;
    y0 = vtail.yden/vtail.num;
    
    x1 = vhead.xden/vhead.num;
    y1 = vhead.yden/vhead.num;

    fprintf (stderr, "cut: x = %f y = %f | retval = %d\n", x1, y1, retval); 
    
    switch(_sy_avs_render_method){
    case POLYLINE:      
      Sy_AvsInsertVertex (vertex, &vindex, x1, y1, 0.0); 
      Sy_AvsInsertVertex (vertex, &vindex, x0, y0, 0.0); 
      break;
    case POLYTRIANGLE: 
      switch (_sy_avs_triangle_decor){
      case CONSTANT:
	Sy_AvsInsertTriStrip (vertex, &vindex, x1, y1, 0.0, _sy_avs_triangle_height);
	Sy_AvsInsertTriStrip (vertex, &vindex, x0, y0, 0.0, _sy_avs_triangle_height);
	break;
      case RADIUS:
	switch (_sy_avs_data_focus){
	case FOREGROUND: if (!syWstate.foreground) head_radius = _sy_avs_triangle_height; break;
	case BACKGROUND: if (syWstate.foreground ) head_radius = _sy_avs_triangle_height; break;
	case ALLDATA: break;
	}
	head_radius = min(head_radius, _sy_avs_value_limit);
	Sy_AvsInsertTriStrip (vertex, &vindex, x1, y1, 0.0, head_radius);
	switch (_sy_avs_data_focus){
	case FOREGROUND: if (!syWstate.foreground) tail_radius = _sy_avs_triangle_height; break;
	case BACKGROUND: if (syWstate.foreground ) tail_radius = _sy_avs_triangle_height; break;
	case ALLDATA: break;
	}
	tail_radius = min(tail_radius, _sy_avs_value_limit);
	Sy_AvsInsertTriStrip (vertex, &vindex, x0, y0, 0.0, tail_radius);
	break;
      case RESIDUAL:
	switch (_sy_avs_data_focus){
	case FOREGROUND: if (!syWstate.foreground) 
	  head_residual = tail_residual = _sy_avs_triangle_height; break;
	case BACKGROUND: if (syWstate.foreground )
	  head_residual = tail_residual = _sy_avs_triangle_height; break;
	case ALLDATA: break;
	}
	head_residual = min(head_residual, _sy_avs_value_limit);
	tail_residual = min(tail_residual, _sy_avs_value_limit);

        /* fprintf (stderr, "hr = %f, tr = %f\n", head_residual, tail_residual); */
	Sy_AvsInsertTriStrip (vertex, &vindex, x1, y1, 0.0, head_residual);
	Sy_AvsInsertTriStrip (vertex, &vindex, x0, y0, 0.0, tail_residual);
	break;
      case BIRTH_VECTOR:
	break;
      }
      break;
    }
  }
}

static void Sy_AvsSkel2GeomTerm (void)
{
  static void Sy_AvsSkelInsert (void);
  
  int i, j;

  fprintf (stderr, "now in Sy_AvsSkel2GeomTerm\n");

  switch(_sy_avs_render_method){
  case POLYLINE: 
    if (vindex <= 1) {
      vindex = nindex = 0;
      return;
    }
    GEOMadd_polyline (obj, vertex, GEOM_NULL, vindex, GEOM_COPY_DATA); 
      break;

  case POLYTRIANGLE:
    if (vindex <= 2) {
      vindex = nindex = 0;
      return;
    }

    for (i = 2; i < vindex; i+=2){
      normal[i-2].nx = normal[i-1].nx = normal[i].nx = normal[i+1].nx = 
	/* vertex[i].y   - vertex[i-1].y */ 0.0;
      normal[i-2].ny = normal[i-1].ny = normal[i].ny = normal[i+1].ny =
	/* vertex[i-1].x - vertex[i].x */ 0.0;
      normal[i-2].nz = normal[i-1].nz = normal[i].nz = normal[i+1].nz = 0.0;
    }

    switch (_sy_avs_triangle_decor){
    case CONSTANT:
    case RADIUS: /*
      for (i=0; i<vindex; i++){
	fprintf (stderr, "(v.x, v.y, v.z)=(%f %f %f)\n", vertex[i].x, vertex[i].y, vertex[i].z);
	fprintf (stderr, "(n.x, n.y, n.z)=(%f %f %f)\n", normal[i].nx, normal[i].ny, normal[i].nz);
      } */

      GEOMadd_polytriangle (obj, vertex, normal, GEOM_NULL, vindex, 
			    GEOM_COPY_DATA); 
      break;
    case RESIDUAL:
    case BIRTH_VECTOR:
      fprintf (stderr, "case BV, vindex = %d\n", vindex);
      for (i = 0; i < vindex; i+=4){
       /*  for (j=0; j<4; j++){
	  fprintf (stderr, "(v.x, v.y, v.z)=(%f %f %f)\n", vertex[i+j].x, vertex[i+j].y, vertex[i+j].z);
          fprintf (stderr, "(n.x, n.y, n.z)=(%f %f %f)\n", normal[i+j].nx, normal[i+j].ny, normal[i+j].nz);
	} */
	GEOMadd_polytriangle (obj, &vertex[i], &normal[i], GEOM_NULL, 4,
			      GEOM_COPY_DATA);
      }
      break;
    }
  }
  vindex = nindex = 0;

  /*  Sy_AvsSkelInsert (); */
}

static void Sy_AvsSkel2GeomExit (void)
{
  switch(_sy_avs_render_method){
  case POLYLINE: 
    if (vertex) free (vertex);
    break;
  case POLYTRIANGLE:
    if (vertex) free (vertex);
    if (normal) free (normal);    
    break;
  }
} 

static void Sy_AvsInsertVertex (SyAvsVector *vertex, int *vindex, float x, float y, float z)
{
  vertex[*vindex].x = x;
  vertex[*vindex].y = y;
  vertex[*vindex].z = z;
  (*vindex)++;
}

static void Sy_AvsInsertTriStrip (SyAvsVector *vertex, int *vindex, float x, float y, float z1, float z2)
{
    vertex[*vindex].x = x;
    vertex[*vindex].y = y;
    vertex[*vindex].z = z1;
    (*vindex)++;
    vertex[*vindex].x = x;
    vertex[*vindex].y = y;
    vertex[*vindex].z = z2;
    (*vindex)++;
}


