/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more Details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * Triangulation Data Structure to AVS Geometry Transform
 *
 */

static char trigeomId[] = "%W%\t%G%";

#include <stdio.h>

#include <avs/avs.h>
#include <avs/avs_data.h>
#include <avs/geom.h>
#include <avs/flow.h>

#include "Sy_Avs.h"
#include "Sy_Skeleton.h"

extern enum SyRenderMethod  _sy_avs_render_method;
extern enum SyTriangleDecor _sy_avs_triangle_decor;
extern enum SyColorMode     _sy_avs_color_mode;
extern float              _sy_avs_triangle_height;
extern float              _sy_avs_value_limit;
extern enum SyDataFocus     _sy_avs_data_focus;

static GEOMobj* obj;
static SyTriDAC*  tri;

static SyAvsVector*    vertex;
static SyAvsNormal*    normal;
static SyAvsColor*     color;
static int           vindex, nindex, cindex,
                     vdim, ndim, cdim;

static void Sy_AvsInsertVertex(SyAvsVector *vertex, int *vindex, 
			       float x, float y, float z);
static void Sy_AvsInsertTriStrip(SyAvsVector *vertex, int *vindex, 
				 float x, float y, float z1, float z2);
static void Sy_AvsInsertColorVal(SyAvsColor *color, int *cindex, 
				 float r, float g, float b);
static void Sy_AvsInsertColor(SyAvsColor *color, int *cindex, SyAvsColor atom);

static void Sy_AvsInsertColorStrip(SyAvsColor *color, int *cindex, 
				   SyAvsColor atom);
static void Sy_AvsInsertColorStripVal(SyAvsColor *color, int *cindex, 
				      float r, float g, float b);

static void Sy_AvsTri2GeomInit(void);
static void Sy_AvsTri2GeomPrep(void);
static void Sy_AvsTriInsert   (void);
static void Sy_AvsTri2GeomTerm(void);
static void Sy_AvsTri2GeomExit(void);

void Sy_AvsTri2Geom (GEOMobj *avs_obj, SyTriDAC *in_tri, 
		     SySGroupThres *sogthres)
{
  SyWalktask wt;
  
  obj  = avs_obj;
  tri  = in_tri;

  Sy_ResetWalktask (&wt);

  wt.group_thres = sogthres->cont_thres;
  wt.narr_thres  = sogthres->narr_thres;
  wt.pro_thres   = sogthres->prox_thres;

  wt.user_function[WALK] = (void(*)())Sy_AvsTriInsert;

  wt.user_function[ROT ] = (_sy_avs_color_mode == COARSE_COL)? 
                           wt.user_function[WALK] : (void(*)())NULL;

  wt.user_function[INIT] = (void(*)())Sy_AvsTri2GeomInit;
  wt.user_function[TRAV] = (void(*)())Sy_AvsTri2GeomPrep;
  wt.user_function[TERM] = (void(*)())Sy_AvsTri2GeomTerm;

  wt.user_function[FEXIT] = (void(*)())Sy_AvsTri2GeomExit;

  Sy_InitTrianTraverse (tri, &wt);

  Sy_TrianTraverse (tri, &wt);
}
 
static void Sy_AvsTri2GeomInit (void)
{

  int max_tri;

  vertex = (SyAvsVector*) NULL;
  normal = (SyAvsNormal*) NULL;
  vindex = nindex = cindex = 0;
  vdim   = ndim   = vdim   = 0;
  
  switch(_sy_avs_triangle_decor){
  case CONSTANT:
    max_tri = 8;
    break;
  }
  switch(_sy_avs_render_method){
  case POLYLINE: 
    vdim   = 4*(tri->edim+1);
    vertex = (SyAvsVector*) Sy_Malloc(vdim * sizeof(SyAvsVector));

    if (_sy_avs_color_mode == FINE_COL){
      cdim  = 4*(tri->edim+1);
      color = (SyAvsColor*) Sy_Malloc (cdim * sizeof(SyAvsColor));      
    }
    else color = (SyAvsColor*)NULL;
    break;
  case POLYTRIANGLE:
    vdim   = ndim = max_tri*(tri->edim+1);    
    vertex = (SyAvsVector*) Sy_Malloc(vdim * sizeof(SyAvsVector));
    normal = (SyAvsNormal*) Sy_Malloc(ndim * sizeof(SyAvsNormal));
    if (_sy_avs_color_mode == FINE_COL){
      cdim  = max_tri*(tri->edim+1);
      color = (SyAvsColor*) Sy_Malloc (cdim * sizeof(SyAvsColor));      
    }
    else color = (SyAvsColor*)NULL;
    break;
  }
}

static void Sy_AvsTri2GeomPrep (void)
{
  vindex = nindex = cindex = 0;
}

static void Sy_AvsTriInsert (void)
{
  float x, y, x1, y1, x2, y2;
  int   is_hardlink,
        tail_point, head_point;

  /* fprintf (stderr, "x = %f y = %f \n", x, y); */

  switch(_sy_avs_render_method){
  case POLYLINE: 
    switch (_sy_avs_color_mode){
    case COARSE_COL:
      x = wr.tri->point[wr.current.head_point].x;
      y = wr.tri->point[wr.current.head_point].y;
      Sy_AvsInsertVertex (vertex, &vindex, x, y, 0.0);
      fprintf (stderr, "id: %d, next_edge: %d, (%f,%f)\n", 
	       wr.id, wr.current.next_edge, x, y);
    break;
    case FINE_COL:
      if (wr.current.next_edge != wr.start.next_edge){
	tail_point = wr.tri->edge[wr.last.next_edge].tail_point;
	head_point = wr.tri->edge[wr.last.next_edge].head_point;
	x1 = wr.tri->point[tail_point].x;
	y1 = wr.tri->point[tail_point].y;
	x2 = wr.tri->point[head_point].x;
	y2 = wr.tri->point[head_point].y;
	is_hardlink = wr.tri->is_hardlink[wr.last.next_edge];
	Sy_AvsInsertVertex (vertex, &vindex, x1, y1, 0.0);
	Sy_AvsInsertVertex (vertex, &vindex, x2, y2, 0.0);	    
	Sy_AvsInsertColor (color, &cindex, is_hardlink ? yellow : white);
	Sy_AvsInsertColor (color, &cindex, is_hardlink ? yellow : white);
	fprintf (stderr, "id: %d, next_edge: %d, link: %d (%f,%f) to (%f,%f)\n", 
		 wr.id, wr.last.next_edge,
		 is_hardlink, x1, y1, x2, y2);
      }
      break;
    }
    break;
  case POLYTRIANGLE:
    switch (_sy_avs_triangle_decor){
    case CONSTANT:
      switch (_sy_avs_color_mode){
      case COARSE_COL:
	x = wr.tri->point[wr.current.head_point].x;
	y = wr.tri->point[wr.current.head_point].y;
	Sy_AvsInsertTriStrip (vertex, &vindex, x, y, 0.0, _sy_avs_triangle_height);
	break;
      case FINE_COL:
	if (wr.current.next_edge != 0){
	  tail_point = wr.tri->edge[wr.current.next_edge].tail_point;
	  head_point = wr.tri->edge[wr.current.next_edge].head_point;
	  x1 = wr.tri->point[tail_point].x;
	  y1 = wr.tri->point[tail_point].y;
	  x2 = wr.tri->point[head_point].x;
	  y2 = wr.tri->point[head_point].y;
	  is_hardlink = wr.tri->is_hardlink[wr.current.next_edge];
	  Sy_AvsInsertTriStrip (vertex, &vindex, x1, y1, 0.0, _sy_avs_triangle_height);
	  Sy_AvsInsertTriStrip (vertex, &vindex, x2, y2, 0.0, _sy_avs_triangle_height);
	  Sy_AvsInsertColorStrip (color, &cindex, is_hardlink ? yellow : white);
	  Sy_AvsInsertColorStrip (color, &cindex, is_hardlink ? yellow : white);
	  fprintf (stderr, "next_edge: %d, is_hardlink: %d\n", wr.current.next_edge,
		   is_hardlink);
	}
	break;
      }
      break;
    }
    break;
  }
}

static void Sy_AvsTri2GeomTerm (void)
{
  int i;

  switch(_sy_avs_render_method){
  case POLYLINE: 
    if (vindex <= 1) {
      vindex = nindex = cindex = 0;
      return;
    }
    if (_sy_avs_color_mode == FINE_COL){
      GEOMadd_disjoint_line (obj, vertex, color, vindex, GEOM_COPY_DATA); 
    }
    else{
      GEOMadd_polyline (obj, vertex, GEOM_NULL, vindex, GEOM_COPY_DATA); 
    }
    break;

  case POLYTRIANGLE:
    if (vindex <= 2) {
      vindex = nindex = cindex = 0;
      return;
    }

    for (i = 1; i < vindex; i+=2){
      normal[i-1].nx = normal[i].nx = vertex[i].y   - vertex[i-1].y;
      normal[i-1].ny = normal[i].ny = vertex[i-1].x - vertex[i].x;
      normal[i-1].nz = normal[i].nz = 0.0;
    }

    switch (_sy_avs_triangle_decor){
    case CONSTANT:
      if (_sy_avs_color_mode == FINE_COL){
	for (i = 0; i < vindex; i+=4){
	  GEOMadd_polytriangle (obj, &vertex[i], &normal[i], &color[i], 4,
				GEOM_COPY_DATA);
	}
      }
      else{
	GEOMadd_polytriangle (obj, vertex, normal, GEOM_NULL, vindex, 
			      GEOM_COPY_DATA);
      } 
      break;
    }
  }
  vindex = nindex = cindex = 0;
}

static void Sy_AvsTri2GeomExit (void)
{
  switch(_sy_avs_render_method){
  case POLYLINE: 
    if (vertex) free (vertex);
    if (color)  free (color);
    break;
  case POLYTRIANGLE:
    if (vertex) free (vertex);
    if (normal) free (normal);
    if (color)  free (color);
    break;
  }
} 

static void Sy_AvsInsertVertex (SyAvsVector *vertex, int *vindex, float x, float y, float z)
{
  vertex[*vindex].x = x;
  vertex[*vindex].y = y;
  vertex[*vindex].z = z;
  (*vindex)++;
}

static void Sy_AvsInsertTriStrip (SyAvsVector *vertex, int *vindex, float x, float y, float z1, float z2)
{
    vertex[*vindex].x = x;
    vertex[*vindex].y = y;
    vertex[*vindex].z = z1;
    (*vindex)++;
    vertex[*vindex].x = x;
    vertex[*vindex].y = y;
    vertex[*vindex].z = z2;
    (*vindex)++;
}

static void Sy_AvsInsertColorVal (SyAvsColor *color, int *cindex, float r, float g, float b)
{
  color[*cindex].r = r;
  color[*cindex].g = g;
  color[*cindex].b = b;
  (*cindex)++;
}

static void Sy_AvsInsertColor (SyAvsColor *color, int *cindex, SyAvsColor atom)
{
  color[*cindex].r = atom.r;
  color[*cindex].g = atom.g;
  color[*cindex].b = atom.b;
  (*cindex)++;
}

static void Sy_AvsInsertColorStrip (SyAvsColor *color, int *cindex, SyAvsColor atom)
{
  color[*cindex].r = atom.r;
  color[*cindex].g = atom.g;
  color[*cindex].b = atom.b;
  (*cindex)++;
  color[*cindex] = color[(*cindex)-1];
  (*cindex)++;
}

static void Sy_AvsInsertColorStripVal (SyAvsColor *color, int *cindex, float r, float g, float b)
{
  color[*cindex].r = r;
  color[*cindex].g = g;
  color[*cindex].b = b;
  (*cindex)++;
  color[*cindex] = color[(*cindex)-1];
  (*cindex)++;
}




