/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#include <stdio.h>
#include <math.h>
#include "Sy_Skeleton.h"
#include "Sy_Uparse.h"
#include "Sy_Graphics.h"

#define SY_MAXNOPT 2

static char prune_Id[] = "%W%\t%G%";

static char* usageStrings[] = {
"[options...] {infile,-} {outfile,-}",
"-verbose        : verbose mode",
"-ascii          : switch to ascii I/O",
"-binary         : switch to binary I/O",
"-timer          : show timer",
"-circ           : circularity residual",
"-bicirc         : bi-circularity residual",
"-klein          : residual according to Fernand Klein",
"-chord          : chord residual (default)",
"-pot            : potential residual",
"-threshold val  : regularization threshold",
"-flatten        : flatten operation",
"-flthres        : flatten threshold",
"-rich           : force generation of residual data for every edge",
"-infix          : fix infinity values",
"-keepnum        : preserve number of components",
"-smpot          : smoothed potentials",
"-smchord        : smoothed 'chord'",
"-smcirc         : smoothed 'circularity r.'",
"-smooth         : skeleton space -> default smoothed chord residual",
"-sigma val      : sigma for smoothing (meaning operation dependent)",
"-gauss          : trigger Gaussian smoothing (default)",
"-oliensis       : smoothing with Oliensis' kernel",
"-split          : generate border vertices",
"-bdtrans        : transform boundary",
"-automatic      : automatic scale selection",
"-thorough       : try very hard to prune spurious branches",
"-iter n         : iterate n times",
"-lo val         : lower val",
"-hi val         : upper val",
"-pyramid        : invoke skeleton-pyramid algorithm after pruning",
"-trigger val    : trigger level for skeleton-pyramid (in percent)",
"-cluster val    : clustering interval for pyramid builder (in percent)",
"-rank num       : rank order of skeleton to display",
"-outfile name   : write to file 'name' (contents op dependent)",
"-reconstruct    : display reconstructed object",
"-bg             : show background (skeletons)",
"-fg             : show foreground (skeletons)",
"-raw            : do not draw object boundaries",
"-rank num       : rank order of skeleton to display",
"-compile        : compile into a user-friendly skeleton representation",
"-xcenter coord  : ",
"-ycenter coord  : 'focus of interest'",
"-scale factor   : scale factor",
"-xupleft coord  : ",
"-yupleft coord  : upper left corner of display 'window'",
"-xwindim val    : x-dimension of display window",
"-ywindim val    : y-dimension of display window",
"-linewidth val  : set default line width",
"-idraw          : send commands for idraw to standard output",
"-display screen : display: X Windows display identifier",
"infile : file containing IPD and SyVoroDAC structures",
"outfile: file with IPD and SKELDAC structures", NULL};

static SyOption longopt[] = {
{"threshold", SY_ARG_REQ},
{"sigma",     SY_ARG_REQ},
{"iter",      SY_ARG_REQ},
{"lo",        SY_ARG_REQ},
{"hi",        SY_ARG_REQ},
{"outfile",   SY_ARG_REQ},
{"flthres",   SY_ARG_REQ},
{"xcenter",   SY_ARG_REQ},
{"ycenter",   SY_ARG_REQ},
{"scale",     SY_ARG_REQ},
{"xupleft",   SY_ARG_REQ},
{"yupleft",   SY_ARG_REQ},
{"trigger",   SY_ARG_REQ},
{"cluster",   SY_ARG_REQ},
{"rank",      SY_ARG_REQ},
{"xwindim",   SY_ARG_REQ},
{"ywindim",   SY_ARG_REQ},
{"linewidth",   SY_ARG_REQ},
{"reconstruct", SY_ARG_NONE},
{"fg",        SY_ARG_NONE},
{"bg",        SY_ARG_NONE},
{"raw",       SY_ARG_NONE},
{"thorough",  SY_ARG_NONE},
{"ascii",     SY_ARG_NONE},
{"binary",    SY_ARG_NONE},
{"verbose",   SY_ARG_NONE},
{"timer",     SY_ARG_NONE},
{"flatten",   SY_ARG_NONE},
{"circ",      SY_ARG_NONE},
{"bicirc",    SY_ARG_NONE},
{"klein",     SY_ARG_NONE},
{"chord",     SY_ARG_NONE},
{"pot",       SY_ARG_NONE},
{"infix",     SY_ARG_NONE},
{"keepnum",   SY_ARG_NONE},
{"smpot",     SY_ARG_NONE},
{"smchord",   SY_ARG_NONE},
{"smcirc",    SY_ARG_NONE},
{"smooth",    SY_ARG_NONE},
{"pyramid",   SY_ARG_NONE},
{"gauss",     SY_ARG_NONE},
{"oliensis",  SY_ARG_NONE},
{"split",     SY_ARG_NONE},
{"bdtrans",   SY_ARG_NONE},
{"rich",      SY_ARG_NONE},
{"automatic", SY_ARG_NONE},
{"compile",   SY_ARG_NONE},
{"idraw",     SY_ARG_NONE},
{"display",   SY_ARG_REQ},SY_NULL_OPTION};

enum PruneOp {PRUNE_WITH_IPD, PRUNE_WITH_TRIAN, PRUNE_WITH_IPD_PYR, 
              NO_PRUNEOP};

void Sy_OutputBirthSetHistogramNew (SySkelDAC *vsk, char *filename, 
				    float threshold, float val3d, int is3d);

main(int argc, char **argv)
{
  SyVoroDAC       *vd = NULL; 
  SyIpoDAC        *ipd = NULL;
  SyTriDAC        *tri = NULL;
  SySkelDAC       vsk;
  SyMScaleSkelDAC mvsk;
  
  SyVSkeletonSet  *vss;

  SyTriPot        *pot; 

  SyIOhandle      *iohandle      = NULL, 
                  *iohandle_ipo  = NULL,
                  *iohandle_voro = NULL,
                  *iohandle_tri  = NULL;

  char*           nopt[SY_MAXNOPT];
  char*           display;
  int             nrnopt, max_iter = 0, k, rank;
  int             is_verbose, is_ascii, is_binary, is_flatten,
                  is_timer, is_eresid, is_border, is_output, 
                  is_automatic,
                  is_infix, is_maxfix, is_smoothpoti, is_smoothcho, 
                  is_smoothcirc,
                  is_pyramid, is_reconstruct, is_thorough, is_idraw,
                  is_negative = SY_FALSE, is_compiling;

  char            hostname    [SY_HOSTNAME_LEN], 
                  win_title   [SY_TITLE_LEN], 
                  icon_title  [SY_TITLE_LEN],
                  display_info[SY_DISPLAY_STRING_LEN],
                  charbuf     [SY_CHARBUF_SIZE],
                  *th_str;

  float           uscale, pyr_th, pyr_delta, base_reg_thres;
  int             xPos, yPos, xZen, yZen;

  float           th, flatten_th, sigma, sigma_lo, sigma_hi, sigma_step;
  int             io_result, pid;
  int             no_ipo_found = 1, no_voro_found = 1, no_tri_found = 1;

  enum PruneOp    oper = NO_PRUNEOP;
  

  SyControl        ctrl;
  SyDisplayControl control;
  SyRegulControl   rcntrl;

  control.raw_mode     = SKE_OFF;
  control.display_mode = SKE_PLAIN;
  control.rank         = 0;
   
  if (!uparse_long(argv, argc, NULL, longopt, nopt, SY_MAXNOPT, &nrnopt)
      || (nrnopt < 1)){
    Sy_Usage (argv[0], usageStrings); exit(1);
  }

  gethostname (hostname, SY_HOSTNAME_LEN);     
  sprintf (win_title, "%s @ %s <%s>", argv[0], hostname, nopt[0]);
  sprintf (icon_title, "%s", nopt[0]);
  sprintf (display_info, "%s", long_option("display") ? 
	   long_argopt("display"): "default X server");
  
  display     = long_option("display") ? long_argopt("display") : (char*)NULL;

  is_binary   = long_option("binary");
  is_ascii    = long_option("ascii");
  is_verbose  = long_option("verbose");
  is_timer    = long_option("timer");
  is_thorough = long_option("thorough");
  
  sscanf (long_option("xcenter") ? long_argopt("xcenter"): "0",  "%d", &xZen);
  sscanf (long_option("ycenter") ? long_argopt("ycenter"): "0", "%d", &yZen);
  sscanf (long_option("xupleft") ? long_argopt("xupleft"): SY_DEFXSTR,"%d",
	  &xPos);
  sscanf (long_option("yupleft") ? long_argopt("yupleft"): SY_DEFYSTR,"%d",
	  &yPos);
  sscanf (long_option("scale") ? long_argopt("scale")  : "0", "%f", &uscale);
  
  Sy_SetVerbose (is_verbose);
  
  is_automatic   = long_option("automatic");
  is_eresid      = long_option("rich");
  is_infix       = long_option("infix");
  is_maxfix      = long_option("keepnum");
  is_border      = long_option("split");
  is_flatten     = long_option("flatten");
  is_smoothpoti  = long_option("smpot");
  is_smoothcho   = long_option("smchord") || long_option("smooth");
  is_output      = long_option("outfile");
  is_smoothcirc  = long_option("smcirc");
  is_pyramid     = long_option("pyramid");
  is_reconstruct = long_option("reconstruct");
  is_idraw       = long_option("idraw");
  is_compiling   = long_option("compile");
  
  if (long_option("threshold")){
    th_str = long_argopt("threshold");
    if (th_str[0] == '#'){ is_negative = SY_TRUE; th_str[0] = ' '; }
  }
  sscanf (long_option("threshold") ? long_argopt("threshold") : "3.0", "%f", &th);
  if (is_negative) th = -1.0 * th;

  sscanf (long_option("flthres") ? long_argopt("flthres"):"3.0","%f", 
	  &flatten_th);

  sscanf (long_option("trigger") ? long_argopt("trigger"):"2.0", "%f", 
	  &pyr_th);
  sscanf (long_option("cluster") ? long_argopt("cluster"):"10.0", "%f", 
	  &pyr_delta);
  sscanf (long_option("rank") ? long_argopt("rank") : "1", "%d", &rank);
  control.rank = long_option("rank") ? rank : 0;

  sscanf (long_option("sigma") ? long_argopt("sigma") :"2.0", "%f", &sigma);
  sscanf (long_option("hi")    ? long_argopt("hi")    :"50.0","%f", &sigma_hi);
  sscanf (long_option("lo")    ? long_argopt("lo")    :"1.0", "%f", &sigma_lo);
  sscanf (long_option("iter")  ? long_argopt("iter")  : "0",  "%d", &max_iter);

  control.show_boundary = long_option("raw") ? SKE_OFF : SKE_ON;  
  control.display_mode  = long_option("fg") ? SKE_FOREGROUND : 
    long_option("bg") ? SKE_BACKGROUND : SKE_PLAIN;

  sscanf (long_option("linewidth") ? long_argopt("linewidth"): "0",  "%d", 
	  &control.line_width);

  sscanf (long_option("xwindim") ? long_argopt("xwindim") : "512", "%d", 
	  &control.window_dimX);
  sscanf (long_option("ywindim") ? long_argopt("ywindim") : "512", "%d", 
	  &control.window_dimY);

  if (long_option("raw")) control.raw_mode = SKE_ON; 
  else                    control.raw_mode = SKE_OFF;

  if (long_option("xwindim") || long_option("ywindim"))
    control.set_window = SKE_ON; 
  else 
    control.set_window = SKE_OFF;

  sigma_step = 1.0;

  if (max_iter > 0){
    if (max_iter == 1) max_iter = 2;
    sigma_step = (sigma_hi - sigma_lo)/(max_iter - 1);
  }

  Sy_InitSkeleton(&vsk);
 
  rcntrl.sigma_lo              = sigma_lo;
  rcntrl.sigma_hi              = sigma_hi;
  rcntrl.sigma_step            = sigma_step;
  rcntrl.base_threshold        = 1.0;
  rcntrl.is_edge_selection     = SY_FALSE;
  rcntrl.is_pyr_trigger_hint   = SY_FALSE;
  rcntrl.is_pyr_cluster_hint   = SY_FALSE;
  rcntrl.pyr_cluster_hint      = pyr_delta;
  rcntrl.is_concave_point      = NULL;
  rcntrl.is_concave_vertex     = NULL;
  rcntrl.is_concave_vertex_0   = NULL;
  rcntrl.no_shrink_estim       = SY_FALSE;
  rcntrl.is_single_shot        = SY_TRUE;

  rcntrl.gaussian_smoothing = SY_TRUE;
  if (rcntrl.oliensis_smoothing = long_option("oliensis")){
    rcntrl.gaussian_smoothing = SY_FALSE;
  }

  ctrl.threshold            = th;
  ctrl.is_automatic         = SY_FALSE;
  ctrl.is_separate_saliency = SY_FALSE;
  ctrl.regul                = NULL;
  ctrl.branch_el_type       = SY_REAL;

  Sy_SetVerbose (is_verbose);
  sprintf (syStrBuf, "%s: loading points/Voronoi diagram from %s...", 
	   argv[0], nopt[0]);
  Sy_Message (SY_FMTNL, syStrBuf);
    
  if (nrnopt > 0){
    iohandle_ipo  = Sy_CreateIOhandle();
    iohandle_voro = Sy_CreateIOhandle();
    iohandle_tri  = Sy_CreateIOhandle();

    Sy_ManageInput (nopt[0]);
    while ((iohandle = Sy_NextIOhandle()) != NULL) {
      switch (Sy_GetIOproperty (iohandle, TYPE).type){
      case SY_IPD_STRIP:
        ipd = (SyIpoDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
        *iohandle_ipo = *iohandle;
        oper = PRUNE_WITH_IPD;
        break;
      case SY_VORO_STRIP:
        vd = (SyVoroDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
        no_voro_found = 0;
        *iohandle_voro = *iohandle;
        break;
      case SY_TRIAN_STRIP:
        tri = (SyTriDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
	oper = PRUNE_WITH_TRIAN;
	*iohandle_tri = *iohandle;
        break;
      } /* switch */
      Sy_Free(iohandle); iohandle = NULL;
    }
  }

  if (no_voro_found){
    Sy_SimpleError ("No Voronoi data found in %s\n", nopt[0]);
  }

  sprintf (syStrBuf, "%s: points:%d edges:%d vertices:%d",
	   argv[0], vd->pdim, vd->edim, vd->vdim);
  Sy_Message (SY_FMTNL, syStrBuf);

  ipd->info  = NULL;
  ipd->contourdim = 0; ipd->maxcontourdim = 0;
  vsk.pdim   = vd->pdim;
  vsk.edim   = vd->edim;
  vsk.vdim   = vd->vdim;
  vsk.point  = vd->point;
  vsk.edge   = vd->edge;
  vsk.vertex = vd->vertex;
  vsk.regul  = (SyRegul*) NULL;
  vsk.is_nfurcation_info = SY_FALSE;
  vsk.is_smcir           = SY_FALSE;

  if (is_eresid) Sy_SetResidualCompMode (RES_FULL);
  if (is_border) Sy_SetResidualBorderMode (RES_FULL);
  if (is_infix)  Sy_SetInfix (is_infix);
  if (is_maxfix) Sy_SetMaxfix (is_maxfix);

  if (is_timer) Sy_InitTimer();

  switch (oper){
  case NO_PRUNEOP:
    Sy_SimpleError ("No skeleton data found in %s\n", nopt[0]);
    break;
  case PRUNE_WITH_IPD:
    if (is_automatic){
      vsk.prune_method = SKE_SMCHO;
      th = 1.0;
      rcntrl.is_thorough           = is_thorough;
      rcntrl.smcir_term_thres      = SY_SMCIR_TERM_THRES;
      rcntrl.smcir_search_thres    = SY_SMCIR_SEARCH_THRES;
      rcntrl.is_single_shot        = SY_FALSE;
      rcntrl.pyr_cluster_delta     = pyr_delta;
      ctrl.is_automatic            = SY_TRUE;
      Sy_Message (SY_FMT, "Automatic pruning. ");
      Sy_SkelSpaceRegularize (&mvsk, &vsk, ipd, &rcntrl);
    }
    else  
    if (is_smoothpoti){
      vsk.prune_method = SKE_SMPOT;
      if (max_iter == 0){
	Sy_RegSmooth (&vsk, ipd, Sy_SmoothRegularize, is_verbose, sigma, 
		      long_option("bdtrans"), &rcntrl);
	if (is_output) 
	  Sy_OutputBirthSetHistogramNew (&vsk, long_argopt("outfile"), th, 
					 sigma, 1);
      }
      else{	
	/* rcntrl.no_shrink_estim = SY_TRUE; */
	for (k = 0; k < max_iter; k++){
	  if (vsk.regul != (SyRegul*)NULL) free(vsk.regul);
	  sigma = sigma_lo + k*sigma_step;
	  Sy_RegSmooth (&vsk, ipd, Sy_SmoothRegularize, is_verbose, sigma,
			long_option("bdtrans"), &rcntrl);
	  Sy_AddMScaleInfo (&mvsk, &vsk, sigma, max_iter, k);
	}
	if (is_output)
	  Sy_OutputSkelSpaceLog (&mvsk, long_argopt("outfile"), th);
      }/* else : max_iter > 0 */
    }
    else
    if (is_smoothcirc){
      vsk.prune_method = SKE_SMCIR;
      Sy_RegSmooth (&vsk, ipd, Sy_SmoothRegularize, is_verbose, sigma, 
		    long_option("bdtrans"), &rcntrl);
      if (is_output) 
	Sy_OutputBirthSetHistogramNew (&vsk, long_argopt("outfile"), th, 
				       sigma, 1);
    }
    else
    if (is_smoothcho){
      vsk.prune_method = SKE_SMCHO;
      if (max_iter == 0){
	Sy_RegSmooth (&vsk, ipd, Sy_SmoothRegularize, is_verbose, sigma,
		      long_option("bdtrans"), &rcntrl);
	if (is_output) 
	  Sy_OutputBirthSetHistogramm (&vsk, long_argopt("outfile"), th, 
						    sigma, 1);
      }
      else{
	for (k = 0; k < max_iter; k++){
	  if (vsk.regul != (SyRegul*)NULL) free(vsk.regul);
	  sigma = sigma_lo + k*sigma_step;
	  Sy_RegSmooth (&vsk, ipd, Sy_SmoothRegularize, is_verbose, sigma,
			long_option("bdtrans"), &rcntrl);
	  Sy_AddMScaleInfo (&mvsk, &vsk, sigma, max_iter, k);
	}
	if (is_output)
	  Sy_OutputSkelSpaceLog (&mvsk, long_argopt("outfile"), th);
      }/* else : max_iter > 0 */
    }
    else
    if (long_option("circ")) 
      Sy_Regularize (&vsk, ipd, Sy_CirResidual, is_verbose);
    else 
    if (long_option("pot")) 
      Sy_Regularize (&vsk, ipd, Sy_PotResidual, is_verbose);
    else
    if (long_option("klein")) 
      Sy_Regularize (&vsk, ipd, Sy_KleinResidual, is_verbose);
    else
    if (long_option("bicirc")) 
      Sy_Regularize (&vsk, ipd, Sy_BicirResidual, is_verbose);
    else
      Sy_Regularize (&vsk, ipd, Sy_ChordResidual, is_verbose);
    break;
  case PRUNE_WITH_TRIAN:
    if (long_option("circ")) 
      Sy_ComplexRegularize (&vsk, tri, Sy_ComplexCirResidual);
    else
    if (long_option("pot")) 
      Sy_ComplexRegularize (&vsk, tri, Sy_ComplexPotResidual);
    else
    if (long_option("klein")) 
      Sy_ComplexRegularize (&vsk, tri, Sy_ComplexKleinResidual);
    else
      Sy_ComplexRegularize (&vsk, tri, Sy_ComplexChordResidual);
    break;
  }


  if (is_flatten){
    Sy_Message (SY_FMTNL, "Flattening");
    Sy_RegFlatten (&vsk, ipd, Sy_DiamRegularize, is_verbose, flatten_th);
  }

  if (is_timer) Sy_ShowTimer();

  vsk.is_hierarchy = SY_FALSE;
  vsk.hier         = NULL;

  if (is_pyramid){
    if (long_option("threshold")){
      base_reg_thres             = th;
      rcntrl.term_threshold      = th;
    }
    else{
      base_reg_thres             = th = 1.0;
      rcntrl.term_threshold      = 0.5;
    }
    rcntrl.is_thorough           = is_thorough;
    rcntrl.base_threshold        = 1.0;
    rcntrl.search_threshold      = 1.0; 
    rcntrl.smcir_term_thres      = is_thorough ? 3.0 : SY_SMCIR_TERM_THRES;
    rcntrl.smcir_search_thres    = SY_SMCIR_SEARCH_THRES;
    rcntrl.pyr_trigger_threshold = pyr_th;
    rcntrl.pyr_cluster_delta     = pyr_delta;
    rcntrl.is_search_threshold   = SY_TRUE;
    rcntrl.is_term_threshold     = SY_TRUE;

    if (long_option("trigger")) rcntrl.is_pyr_trigger_hint = SY_FALSE;
    if (long_option("cluster")) rcntrl.is_pyr_cluster_hint = SY_FALSE;

    if (is_timer) Sy_InitTimer();
    
    Sy_Pyramid (&vsk, ipd, &rcntrl);
    
    if (is_timer) Sy_ShowTimer();    

    oper = PRUNE_WITH_IPD_PYR;
  }

  if (is_compiling){
    if (is_automatic){
      if (is_pyramid){
	Sy_RelaxPruning (&vsk, ipd, &rcntrl);
      }
      
      ctrl.regul = vsk.regul;
      vsk.regul  = NULL;
      ctrl.is_automatic = SY_TRUE;
      ctrl.is_separate_saliency = SY_TRUE;
      ctrl.threshold = 3.0;      
      if (is_eresid) Sy_SetResidualCompMode (RES_FULL);
      else           Sy_SetResidualCompMode (RES_NORMAL);

      Sy_Regularize (&vsk, ipd, Sy_ChordResidual, is_verbose);
    }

    Sy_SetBinaryIO (is_binary);
    vss = Sy_ExtractSkeletons (&vsk, ipd, &ctrl);

    if (ctrl.regul) Sy_Free(ctrl.regul);
  }

  if (nrnopt > 1){
    sprintf (syStrBuf, "Writing Voronoi skeleton(s) to %s", nopt[1]);
    Sy_Message (SY_FMTNL, syStrBuf);

    iohandle = Sy_CreateIOhandle ();

    if (is_compiling){
      Sy_SetIOproperty (iohandle, FORMAT, (SY_VOID*)((is_ascii) ? 
						     SY_ASCII:SY_BINARY));
      Sy_SetIOproperty (iohandle, TYPE, (SY_VOID*) SY_SKELETON_SET);
      Sy_SetIOproperty (iohandle, DATA, (SY_VOID*) vss);
      Sy_SetIOproperty (iohandle, IO_MODE, (SY_VOID*) SY_FREE_AFTER_WRITE);
    }
    else{
      Sy_SetIOproperty (iohandle, FORMAT, (SY_VOID*) ((is_ascii) ? 
						      SY_ASCII : SY_BINARY));
      Sy_SetIOproperty (iohandle, TYPE, (SY_VOID*) SY_SKEL_STRIP);
      Sy_SetIOproperty (iohandle, DATA, (SY_VOID*) &vsk);
      Sy_SetIOproperty (iohandle, IO_MODE, (SY_VOID*) SY_FREE_AFTER_WRITE);

      switch (oper){
      case PRUNE_WITH_IPD:
      case PRUNE_WITH_IPD_PYR:
	Sy_SetIOproperty (iohandle_ipo, FORMAT, 
			  (SY_VOID*) ((is_ascii) ? SY_ASCII : SY_BINARY));
	Sy_PutIOhandle (iohandle_ipo);
	break;
      case PRUNE_WITH_TRIAN:
	Sy_SetIOproperty (iohandle_tri, FORMAT, 
			  (SY_VOID*) ((is_ascii) ? SY_ASCII : SY_BINARY));
	Sy_PutIOhandle (iohandle_tri);
	break;
      } 
    }
    Sy_PutIOhandle (iohandle);
    Sy_ManageOutput (nopt[1]);
    Sy_Free(iohandle); 
  }
  else{      
    sprintf (syStrBuf, "%s: preparing display (th: %3.2f) on %s", argv[0],
	     th, display_info);
    Sy_Message (SY_FMTNL, syStrBuf);
    
    if ((pid = fork()) == 0){ 
      switch (oper){
      case PRUNE_WITH_IPD:
	if (is_idraw){
	  Sy_Skel2Idraw(&vsk, ipd, th, control.display_mode, stdout); 
	  Sy_DestroyIpo(ipd); Sy_Free(ipd);
	  Sy_DestroyVoronoi(vd); Sy_Free(vd);
	  Sy_DestroySkeleton(&vsk);
	  exit(0);
	}
	if (is_compiling){
	  Sy_XSkeletonSetDisp (vss, win_title, icon_title, 
			       xPos, yPos, xZen, yZen, uscale, display, control);
	}
	else{
	  Sy_XSkelDisp (&vsk, ipd, th, win_title, icon_title, 
			xPos, yPos, xZen, yZen, uscale, display, control);  
	}    
	break;
      case PRUNE_WITH_IPD_PYR:
	if (is_reconstruct){
	  Sy_XPyrReconstruct (&vsk, ipd, vsk.hier, rank, rcntrl.term_threshold,
			      win_title, icon_title,
			      xPos, yPos, xZen, yZen, uscale, display, control);
	}
	else if (is_compiling){
	  Sy_XSkeletonSetDisp (vss, win_title, icon_title, 
			       xPos, yPos, xZen, yZen, uscale, display, control);
	}
	else{
	  Sy_XPyramidDisp (&vsk, ipd, rank, win_title, icon_title,
			   xPos, yPos, xZen, yZen, uscale, display, control, 
			   vsk.hier);    
	}
	break;
      case PRUNE_WITH_TRIAN:
	Sy_XSkelDispWithTrian(&vsk, tri, th, win_title, icon_title,
			      xPos, yPos, xZen, yZen, uscale, display, control); 
	break;
      }
      exit (0);
    } /* if */
  }
  Sy_Free(iohandle_voro);
  Sy_Free(iohandle_ipo);
  Sy_Free(iohandle_tri);

  Sy_DestroyIpo(ipd); 
  Sy_Free(ipd); 
  vd->point  = NULL;
  vd->edge   = NULL;
  vd->vertex = NULL;
  Sy_DestroyVoronoi(vd); 
  Sy_Free(vd);
  Sy_DestroyTrian(tri);
  Sy_Free(tri); 
  Sy_DestroySkeleton(&vsk);
}

