/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#include <stdio.h>
#include "Sy_Skeleton.h"
#include "Sy_Uparse.h"
#include "Sy_Graphics.h"

#define SY_MAXNOPT 1

enum ShoskOp {SHOW_IPO, SHOW_VORO, SHOW_SKEL, SHOW_TRI, SHOW_SKELETON_SET, 
	      NO_SHOSKOP};

static char* usageStrings[] = {
"[options...] {infile,-}", 
"-verbose        : verbose mode",
"-threshold val  : regularization threshold for skeletons",
"-conth val      : continuity threshold (attributed triangulation)",
"-narth val      : threshold for narrowings (attributed triangulation)",
"-proxth val     : proximity threshold (attributed triangulation)",
"-rank num       : rank order of skeleton to display",
"-bold           : draw small disks instead of simple dots",
"-reseps val     : accuracy epsilon for circ./bi-circ. resid.",
"-loceps val     : accuracy epsilon for location (circ./bi-circ. resid.)",
"-ps             : generate postscript (stdout)",
"-bg             : show background (skeletons)",
"-fg             : show foreground (skeletons)",
"-raw            : do not draw object boundaries",
"-check          : consistency check (initiating points only)",
"-resolution val : point density when converting from polygon to point array",
"-factor val     : multiply input coordinates by factor 'val'",
"                  (applies only when converting from contour lists)",
"-xcenter coord  : ",
"-ycenter coord  : 'focus of interest'",
"-scale factor   : scale factor",
"-xupleft coord  : ",
"-yupleft coord  : upper left corner of display 'window'",
"-xwindim val    : x-dimension of display window",
"-ywindim val    : y-dimension of display window",
"-linewidth val  : set default line width",
"-display screen : display: X Windows display identifier",
"-idraw          : send commands for idraw to standard output",
"infile          : file containing structures of the Skeleton Project", NULL};

static SyOption longopt[] = {
{"threshold",   SY_ARG_REQ},
{"proxth",      SY_ARG_REQ},
{"narth",       SY_ARG_REQ},
{"conth",       SY_ARG_REQ},
{"rank",        SY_ARG_REQ},
{"reseps",      SY_ARG_REQ},
{"loceps",      SY_ARG_REQ},
{"xcenter",     SY_ARG_REQ},
{"ycenter",     SY_ARG_REQ},
{"scale",       SY_ARG_REQ},
{"xupleft",     SY_ARG_REQ},
{"yupleft",     SY_ARG_REQ},
{"xwindim",     SY_ARG_REQ},
{"ywindim",     SY_ARG_REQ},
{"linewidth",   SY_ARG_REQ},
{"resolution",  SY_ARG_REQ},
{"factor",      SY_ARG_REQ},
{"testout",     SY_ARG_REQ}, /* for debugging purposes only */
{"ps",          SY_ARG_NONE},
{"check",       SY_ARG_NONE},
{"verbose",     SY_ARG_NONE},
{"bold",        SY_ARG_NONE},
{"fg",          SY_ARG_NONE},
{"bg",          SY_ARG_NONE},
{"raw",         SY_ARG_NONE},
{"idraw",       SY_ARG_NONE},
{"display",     SY_ARG_REQ}, SY_NULL_OPTION};

main(int argc, char **argv)
{
  SyIpoDAC       *ipd;
  SyVoroDAC      *vd;
  SySkelDAC      *vsk;
  SyTriDAC       *tri;
  SySkContourSet *cs, *cs0;
  SyVSkeletonSet *vss;

  SyWalktask     wt;

  char*        nopt[SY_MAXNOPT];
  char*        display;
  int          nrnopt, is_verbose, is_check, is_bold;
  int          is_loceps, is_reseps, is_ps, is_idraw,
               is_negative = SY_FALSE, is_test, is_multiply;

  char         hostname    [SY_HOSTNAME_LEN], 
               win_title   [SY_TITLE_LEN], 
               icon_title  [SY_TITLE_LEN],
               display_info[SY_DISPLAY_STRING_LEN],
               charbuf     [SY_CHARBUF_SIZE],
               *th_str;

  float        th, cont_thres, prox_thres, narr_thres, loceps, reseps,
               resolution, factor;

  float        uscale;
  int          xPos, yPos, xZen, yZen;
  int          pid, rank;
 
  enum ShoskOp oper    = NO_SHOSKOP,
               suboper = NO_SHOSKOP;

  SyIOhandle       *iohandle = NULL, *iohandle_test = NULL;
  SyDisplayControl control;
  SyControl        ctrl;

  if (!uparse_long(argv, argc, NULL, longopt, nopt, SY_MAXNOPT, &nrnopt)
      || (nrnopt < 1)){
    Sy_Usage (argv[0], usageStrings); exit(1);
  }

  gethostname (hostname, SY_HOSTNAME_LEN);     
  sprintf (win_title, "%s @ %s <%s>", argv[0], hostname, nopt[0]);
  sprintf (icon_title, "%s", nopt[0]);
  sprintf (display_info, "%s", long_option("display") ? 
	   long_argopt("display"): "default X server");
  
  display     = long_option("display") ? long_argopt("display") : (char*)NULL;
  
  is_ps       = long_option("ps");
  is_check    = long_option("check");
  is_verbose  = long_option("verbose");
  is_bold     = long_option("bold");
  is_idraw    = long_option("idraw");

  sscanf (long_option("resolution") ? long_argopt("resolution"): 
	  "1.0", "%f", &resolution);

  sscanf (long_option("xcenter") ? long_argopt("xcenter"): "0",  "%d", &xZen);
  sscanf (long_option("ycenter") ? long_argopt("ycenter"): "0", "%d", &yZen);
  sscanf (long_option("xupleft") ? long_argopt("xupleft"): SY_DEFXSTR,"%d",
	  &xPos);
  sscanf (long_option("yupleft") ? long_argopt("yupleft"): SY_DEFYSTR,"%d",
	  &yPos);
  sscanf (long_option("scale")   ? long_argopt("scale")  : "0", "%f", &uscale);

  control.show_boundary = long_option("raw") ? SKE_OFF : SKE_ON;  
  control.display_mode  = long_option("fg") ? SKE_FOREGROUND : 
    long_option("bg") ? SKE_BACKGROUND : SKE_PLAIN;

  sscanf (long_option("linewidth") ? long_argopt("linewidth"): "0",  "%d", 
	  &control.line_width);

  sscanf (long_option("xwindim") ? long_argopt("xwindim") : "512", "%d", 
	  &control.window_dimX);
  sscanf (long_option("ywindim") ? long_argopt("ywindim") : "512", "%d", 
	  &control.window_dimY);

  if (long_option("raw")) control.raw_mode = SKE_ON; else control.raw_mode = SKE_OFF;

  if (long_option("xwindim") || long_option("ywindim"))
    control.set_window = SKE_ON; else control.set_window = SKE_OFF;
  
  Sy_SetVerbose (is_verbose);

  if (long_option("threshold")){
    th_str = long_argopt("threshold");
    if (th_str[0] == '#'){ is_negative = SY_TRUE; th_str[0] = ' '; }
  }
  sscanf (long_option("threshold") ? long_argopt("threshold") :"3.0","%f",&th);
  if (is_negative) th = -1.0 * th;

  sscanf (long_option("conth") ? long_argopt("conth") : "1.0 ", "%f", 
	  &cont_thres);
  sscanf (long_option("narth") ? long_argopt("narth") : "1.0 ", "%f", 
	  &narr_thres);
  sscanf (long_option("proxth")? long_argopt("proxth"): "-0.5", "%f", 
	  &prox_thres);

  sscanf (long_option("rank") ? long_argopt("rank") : "1", "%d", &rank);
  control.rank = long_option("rank") ? rank : 0;

  if (is_loceps = long_option("loceps")) 
    sscanf (long_argopt("loceps"), "%f", &loceps);   
  if (is_reseps = long_option("reseps")) 
    sscanf (long_argopt("reseps"), "%f", &reseps);   

  sscanf ((is_multiply = long_option("factor")) ? 
	  long_argopt ("factor"): "1.0", "%f", &factor);

  is_test = long_option("testout");

  ctrl.resolution  = resolution;
  ctrl.is_multiply = is_multiply;
  ctrl.mult_f      = factor;

  if (is_test){
    iohandle_test = Sy_CreateIOhandle();
  }

  if (nrnopt > 0){
    Sy_ManageInput (nopt[0]);
    while ((iohandle = Sy_NextIOhandle()) != NULL){
      switch (Sy_GetIOproperty (iohandle, TYPE).type){

      case SY_IPD_STRIP: 
        ipd = (SyIpoDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
        oper = suboper = SHOW_IPO;
        break;

      case SY_CONTOUR_UNORDERED_STRIP:
	cs0 = (SySkContourSet*) (Sy_GetIOproperty(iohandle, DATA).data);
	if (cs0->is_polygon){
	  cs = Sy_Polygon2Points (cs0, &ctrl);
	  Sy_DeleteSkContourSet (cs0); 
	  Sy_Free(cs0); cs0 = NULL;
	}
	else cs = cs0;
	  
	Sy_AddFgInfo2ContourSet(cs);
	ipd = Sy_ContourSet2Ipo(cs, NULL);
	if (ipd == NULL){
	  Sy_SimpleError ("Conversion of contour set failed. %s\n", "");
	}
	Sy_CratranBuildIcn (ipd);
	Sy_DeleteSkContourSet (cs); 
	Sy_Free(cs); cs = NULL;
	oper = suboper = SHOW_IPO;  /* proceed as usual */
	break;

      case SY_SKELETON_SET:
	vss  = (SyVSkeletonSet*) (Sy_GetIOproperty(iohandle, DATA).data);
	oper = SHOW_SKELETON_SET;
	if (is_test) *iohandle_test = *iohandle;
	break;

      case SY_VORO_STRIP:
        vd  = (SyVoroDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
	oper = SHOW_VORO;
	break;

      case SY_SKEL_STRIP:
        vsk = (SySkelDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
	oper = SHOW_SKEL;
	break;

      case SY_TRIAN_STRIP:
	tri  = (SyTriDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
	oper = suboper = SHOW_TRI;
	wt.group_thres = cont_thres;
	wt.narr_thres  = narr_thres;
	wt.pro_thres   = prox_thres;
	/* this is definitively a hack */
	tri->is_pointref = 0;
	break;
      } /* switch */
      Sy_Free(iohandle); iohandle = NULL;
    } /* while */

    switch (oper){
    case SHOW_IPO:
      sprintf (charbuf, "%s: SyIpos: %d  SyIcns: %d", 
	       argv[0], ipd->ipodim, ipd->icndim);
      Sy_Message ("%s\n", charbuf);
      if (is_check){
	fprintf (stderr, "@ consistency check ", argv[0]);
	fprintf (stderr, (Sy_CratranIpoCheck(ipd) == 0) ? 
		 "passed\n" : "failed\n");
      }
      if (is_verbose) 
	fprintf (stderr, "@ preparing display on %s...\n", display_info);

      if (is_idraw){
	Sy_Ipo2Idraw(ipd, stdout, SY_FALSE); exit(0);
      }

      if ((pid = fork()) == 0){
	if (is_bold)
	  Sy_XArcDisp (ipd, win_title, icon_title, xPos, yPos, xZen, yZen,
		      uscale, display);
	else
	  if (is_ps) 
	    Sy_Ipo2Postscript(ipd, xPos, yPos, xZen, yZen, uscale);
	  else
	    Sy_XIpoDisp (ipd, win_title, icon_title, xPos, yPos, xZen, yZen,
			uscale, display);
	exit(0);
      }
      break;
    case SHOW_TRI:
      sprintf (charbuf, "%s: points: %d  edges: %d", 
	       argv[0], tri->pdim, tri->edim);
      Sy_Message ("%s\n", charbuf);
      if (is_verbose) 
	fprintf (stderr, "@ preparing display on %s...\n", display_info);
      if ((pid = fork()) == 0){
	Sy_XTrianDisp (tri, &wt, win_title, icon_title, xPos, yPos, xZen, yZen,
		      uscale, display);
	exit(0);
      }
      break;
    case SHOW_VORO:
      sprintf (charbuf, "%s: points: %d  edges: %d vertices: %d", 
	       argv[0], vd->pdim, vd->edim, vd->vdim);
      Sy_Message ("%s\n", charbuf);

      if (is_idraw){
	Sy_Voro2Idraw(vd, stdout, xZen, yZen, uscale); exit(0);
      }

      if (is_verbose) 
	fprintf (stderr, "@ preparing display on %s...\n", display_info);
      if ((pid = fork()) == 0){
	Sy_XVoroDisp (vd, win_title, icon_title, xPos, yPos, xZen, yZen,
		     uscale, display, control);
	exit(0);
      }
      break;
    case SHOW_SKEL:
      switch (vsk->prune_method){
      case SKE_CIR:
	if (long_option("threshold")) sscanf (long_argopt("threshold"), "%f", &th); 
	else th = 3.0;
	if (is_negative) th = -1.0 * th;
	if (is_verbose) fprintf (stderr,"Circularity Residual\n");
	break;
      case SKE_BIC:
	if (long_option("threshold")) sscanf (long_argopt("threshold"), "%f", &th); 
	else th = 3.0;
	if (is_negative) th = -1.0 * th;
	if (is_verbose) fprintf (stderr,"Bi-Circularity Residual\n");
	break;
      case SKE_KLE:
	sscanf (long_option("threshold") ? long_argopt("threshold") :"3.0", "%f", &th);
	if (is_negative) th = -1.0 * th;
	if (is_verbose) fprintf (stderr,"Klein's Residual\n");
	break;
      case SKE_CHO:
	sscanf (long_option("threshold") ? long_argopt("threshold") :"3.0", "%f", &th);
	if (is_negative) th = -1.0 * th;
	if (is_verbose) fprintf (stderr,"Chord Residual\n");
	break;
      case SKE_POT:
	sscanf (long_option("threshold") ? long_argopt("threshold") :"10.0", "%f", &th);
	if (is_negative) th = -1.0 * th;
	if (is_verbose) fprintf (stderr,"Potential Residual\n");
	break;
      }

      sprintf (charbuf, "%s: points: %d  edges: %d vertices: %d", 
	       argv[0], vsk->pdim, vsk->edim, vsk->vdim);
      Sy_Message ("%s\n", charbuf);
      if (is_verbose){
	if (vsk->is_hierarchy){
	  fprintf(stderr, "Got skeleton-pyramid data.\n");
	} 
	fprintf (stderr, "@ preparing display on %s...\n", display_info);
      }
      if (is_loceps) Sy_SetLocEpsilon (loceps);
      if (is_reseps) Sy_SetResEpsilon (reseps);
      switch (suboper){
      case SHOW_IPO:
	if ((pid = fork()) == 0){
	  if (vsk->is_hierarchy){
	    if (is_idraw){
	      Sy_SkelPyr2Idraw(vsk, ipd, rank, control.display_mode, 
			       vsk->hier, stdout);
	    }
	    else{
	      Sy_XPyramidDisp (vsk, ipd, rank, win_title, icon_title,
			       xPos, yPos, xZen, yZen, uscale, display, control, 
			       vsk->hier);      
	    }
	  }
	  else{
	    if (is_idraw){
	      Sy_Skel2Idraw(vsk, ipd, th, control.display_mode, stdout); exit(0);
	    }
	    Sy_XSkelDisp (vsk, ipd, th, win_title, icon_title,
			  xPos, yPos, xZen, yZen, uscale, display, control);
	  }
	  exit(0);
	}
	break;
      case SHOW_TRI:
	if ((pid = fork()) == 0){
	  Sy_XSkelDispWithTrian (vsk, tri, th, win_title, icon_title,
				 xPos, yPos, xZen, yZen, uscale, display, control);
	  exit(0);
	}
	break;
      }
      break;
    case SHOW_SKELETON_SET:
      if (is_test){
	Sy_SetIOproperty (iohandle_test, FORMAT, (SY_VOID*) SY_ASCII);
	Sy_PutIOhandle (iohandle_test);
	Sy_ManageOutput (long_argopt("testout"));
	Sy_Free(iohandle_test);
      }
      if ((pid = fork()) == 0){
	if (is_idraw){
	  Sy_SkeletonSet2Idraw(vss, uscale, &control, stdout); exit(0);
	}
	else{
	  Sy_XSkeletonSetDisp (vss, win_title, icon_title, xPos, yPos, 
			       xZen, yZen, uscale, display, control);
	}
	exit(0);
      }
      break;
    default:
      fprintf (stderr, "@ no structures of the Skeleton Project found\n");
      exit(0);
    }
  }
  exit(0);
}



