/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

static char voro_Id[] = "%W%\t%G%";

#include <stdio.h>
#include "Sy_Skeleton.h"
#include "Sy_Uparse.h"
#include "Sy_Graphics.h"

#define SY_MAXNOPT 2

static char* usageStrings[] = {
"[options...] {infile,-} {outfile,-}",
"-ascii          : output in ascii format",
"-verbose        : verbose mode",
"-timer          : show resource usage",
"-factor val     : multiply input coordinates by factor 'val'",
"                  (applies only when converting from contour lists)",
"-resolution val : point density when converting from polygon to point array",
"-xcenter coord  : ",
"-ycenter coord  : 'focus of interest'",
"-scale factor   : scale factor (for display)",
"-xupleft coord  : ",
"-yupleft coord  : upper left corner of display 'window'",
"-display screen : X Windows display identifier",
"infile          : generators (Voronoi seed points)",
"outfile         : Voronoi diagram (DCEL)", NULL};

static SyOption longopt[] = {
{"xcenter",    SY_ARG_REQ},
{"ycenter",    SY_ARG_REQ},
{"scale",      SY_ARG_REQ},
{"xupleft",    SY_ARG_REQ},
{"yupleft",    SY_ARG_REQ},
{"resolution", SY_ARG_REQ},
{"factor",     SY_ARG_REQ},
{"ascii",      SY_ARG_NONE},
{"verbose",    SY_ARG_NONE},
{"timer",      SY_ARG_NONE},
{"display",    SY_ARG_REQ}, SY_NULL_OPTION};

main(int argc, char **argv)
{
  SyVoroDAC      vd;
  SyIpoDAC*      ipd;
  SyTriDAC*      tri;
  SySkContourSet *cs = NULL, *cs0 = NULL;
  FILE           *fpin, *fpout;
  
  int            e_mem, v_mem;
  
  char*          nopt[SY_MAXNOPT];
  char*          display;
  int            nrnopt, is_ascii, is_verbose, is_timer, is_multiply, 
                 pid;
 
  float          uscale, resolution, factor;
  int            xPos, yPos, xZen, yZen;

  char           hostname    [SY_HOSTNAME_LEN], 
                 win_title   [SY_TITLE_LEN], 
                 icon_title  [SY_TITLE_LEN],
                 display_info[SY_DISPLAY_STRING_LEN],
                 charbuf     [SY_CHARBUF_SIZE];

  int            no_ipo_found = 1, no_trian_found = 1, contour_unordered = 0;
  SyIOhandle     *iohandle = NULL, *iohandle_point = NULL;
  SyRusdelta     delta;

  SyDisplayControl  dspctrl;
  SyControl         ctrl;
  enum SyStripType  data_type;

  if (!uparse_long(argv, argc, NULL, longopt, nopt, SY_MAXNOPT, &nrnopt)
      || (nrnopt < 1)){
    Sy_Usage (argv[0], usageStrings); exit(1);
  }
  gethostname (hostname, SY_HOSTNAME_LEN);     
  sprintf (win_title, "%s @ %s <%s>", argv[0], hostname, nopt[0]);
  sprintf (icon_title, "%s", nopt[0]);
  sprintf (display_info, "%s", long_option("display") ? 
	   long_argopt("display"): "default X server");

  display    = long_option("display") ? long_argopt("display") : (char*)NULL;
  
  is_ascii   = long_option("ascii");
  is_verbose = long_option("verbose");
  is_timer   = long_option("timer");

  sscanf (long_option("resolution") ? long_argopt("resolution"): 
	  "1.0", "%f", &resolution);

  sscanf ((is_multiply = long_option("factor")) ? 
	  long_argopt ("factor"): "1.0", "%f", &factor);
  
  sscanf (long_option("xcenter") ? long_argopt("xcenter"): "0",  "%d", &xZen);
  sscanf (long_option("ycenter") ? long_argopt("ycenter"): "0", "%d", &yZen);
  sscanf (long_option("xupleft") ? long_argopt("xupleft"): SY_DEFXSTR,"%d",
	   &xPos);
  sscanf (long_option("yupleft") ? long_argopt("yupleft"): SY_DEFYSTR,"%d",
	   &yPos);
  sscanf (long_option("scale") ? long_argopt("scale")  : "0", "%f", &uscale);

  Sy_SetVerbose (is_verbose);

  ctrl.resolution  = resolution;
  ctrl.is_multiply = is_multiply;
  ctrl.mult_f      = factor;

  dspctrl.raw_mode = SKE_OFF;

  Sy_InitVoronoi (&vd);

  if (is_verbose)
    fprintf (stderr, "%s: loading SyIpos from %s...\n", argv[0], nopt[0]);

  if (nrnopt > 0){
    iohandle_point = Sy_CreateIOhandle();
    Sy_ManageInput (nopt[0]);
    while (((iohandle = Sy_NextIOhandle()) != NULL) && no_ipo_found){
      switch (data_type = Sy_GetIOproperty (iohandle, TYPE).type){
      case SY_IPD_STRIP:
        no_ipo_found = SY_FALSE;
        ipd = (SyIpoDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
	*iohandle_point = *iohandle;
        break;
      case SY_CONTOUR_UNORDERED_STRIP:
	no_ipo_found = SY_FALSE;
	contour_unordered = SY_TRUE;
	cs0 = (SySkContourSet*) (Sy_GetIOproperty(iohandle, DATA).data);
	*iohandle_point = *iohandle;
	break;
      case SY_TRIAN_STRIP:
        no_trian_found = SY_FALSE;
        tri = (SyTriDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
	*iohandle_point = *iohandle;
        break;
      } /* switch */
      Sy_Free(iohandle); iohandle = NULL;
    }
  }
  if (no_ipo_found && no_trian_found){
    Sy_SimpleError ("No Skeleton Project file found in %s\n", nopt[0]);
  }

  if (contour_unordered){
    if (cs0->is_polygon){
      cs = Sy_Polygon2Points (cs0, &ctrl);
      Sy_DeleteSkContourSet (cs0); 
      Sy_Free(cs0); cs0 = NULL;
    }
    else cs = cs0;

    Sy_AddFgInfo2ContourSet(cs);
    ipd = Sy_ContourSet2Ipo(cs, &ctrl);
    if (ipd == NULL){
      Sy_SimpleError ("Conversion of contour set failed. %s\n", "");
    }
    Sy_CratranBuildIcn (ipd);
    Sy_DeleteSkContourSet (cs); 
    Sy_Free(cs); cs = NULL;
    data_type = SY_IPD_STRIP; /* proceed as usual */
  }

  if (is_verbose){ 
    switch(data_type){
    case SY_IPD_STRIP:
      fprintf (stderr, "@%s:  points: %d  contour elements: %d\n", 
	       argv[0], ipd->ipodim, ipd->icndim);
      break;
    case SY_TRIAN_STRIP:
      fprintf (stderr, "@%s:  points: %d  edges: %d\n", 
	       argv[0], tri->pdim, tri->edim);
      break;
    }
  }

  switch(data_type){
  case SY_IPD_STRIP:
    vd.pdim  = ipd->ipodim;  
    vd.point = (SyPoint*) ipd->ipo;
    break;
  case SY_TRIAN_STRIP:
    vd.pdim  = tri->pdim;
    vd.point = tri->point;
    break;
  }

  vd.edge     = Sy_AllocEdge  (3*vd.pdim);
  vd.vertex   = Sy_AllocVertex(2*vd.pdim);
  vd.label.p_ = NULL;
  vd.label.e_ = NULL;
  vd.label.v_ = NULL;

  if (is_timer) Sy_InitTimer();

  Sy_Voronoi (&vd, is_verbose);

  if (is_timer){ 
    delta = Sy_GetTimer(1);
    fprintf (stderr, "Processing speed: %6.0f generators/sec\n", 
	     vd.pdim/delta.utime);
  }
  if (is_verbose) 
    fprintf (stderr, "\n%s: points: %d  edges: %d  vertices: %d \n",
	     argv[0], vd.pdim, vd.edim, vd.vdim);
   if (nrnopt > 1){
     
     iohandle = Sy_CreateIOhandle();
     Sy_SetIOproperty (iohandle, FORMAT, (SY_VOID*)((is_ascii) ? 
						    SY_ASCII:SY_BINARY));
     Sy_SetIOproperty (iohandle, TYPE, (SY_VOID*) SY_VORO_STRIP);
     Sy_SetIOproperty (iohandle, DATA, (SY_VOID*) &vd);
     Sy_SetIOproperty (iohandle, IO_MODE, (SY_VOID*) SY_FREE_AFTER_WRITE);

     Sy_SetIOproperty (iohandle_point, FORMAT, 
		       (SY_VOID*)((is_ascii) ? SY_ASCII:SY_BINARY));

     if (contour_unordered){
       Sy_SetIOproperty (iohandle_point, TYPE, (SY_VOID*) SY_IPD_STRIP);
       Sy_SetIOproperty (iohandle_point, DATA, (SY_VOID*) ipd);
       Sy_SetIOproperty (iohandle_point, IO_MODE, (SY_VOID*)
			 SY_FREE_AFTER_WRITE);
     }

     Sy_PutIOhandle (iohandle_point);
     Sy_PutIOhandle (iohandle);
     Sy_Free(iohandle_point); 
     Sy_Free(iohandle);

     sprintf (charbuf, "@ writing Voronoi Diagram to %s...\n", nopt[1]);
     Sy_Message ("%s", charbuf);

     Sy_ManageOutput (nopt[1]);
   }
   else {
     Sy_Free(iohandle_point); 

     sprintf (charbuf, "@ %s: preparing display on %s\n", argv[0],
	      display_info);
     Sy_Message ("%s", charbuf);
     
     if ((pid = fork()) == 0){
       Sy_XVoroDisp (&vd, win_title, icon_title, xPos, yPos, xZen, yZen,
		     uscale, display, dspctrl);
       exit(0);
     }
    
     Sy_DestroyIpo (ipd); vd.point = NULL;
     Sy_DestroyVoronoi (&vd);
     Sy_Free (ipd);
  }
}

