
/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Division of Applied Sciences,
 *          Harvard University, Cambridge, Massachusetts. 
	    (rlo@hrl.harvard.edu)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1995 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#ifndef _SY_EXPORT_H
#define _SY_EXPORT_H 1

#ifdef  __cplusplus
extern "C" {
#endif

typedef enum _SyBranchNodeType { SY_INTERNAL, SY_END, SY_NILNODE } 
  SyBranchNodeType;

typedef enum _SyBranchElementType {SY_REAL, 
                                   SY_RATIONAL, 
				   SY_INTEGER, 
				   SY_NILBRANCH
} SyBranchElementType;

typedef struct _SyRealPoint {
  float x, y;
} SyRealPoint;

typedef struct _SyIntPoint {
  int x, y;
} SyIntPoint;

typedef struct _SyRationalPoint {
  double xnum, ynum, 
         xden, yden;
} SyRationalPoint;

union SyGenericPoint {
  SyIntPoint      np;
  SyRationalPoint qp;
  SyRealPoint     rp;
};

typedef struct _SyVSkeletonRef {
  int left_contour, right_contour,
      left_point, right_point;
} SyVSkeletonRef;

typedef struct _SyVSkeletonBndryOffset {
  float xoff, 
        yoff;
} SyVSkeletonBndryOffset;

/* NOTE on array cell allocation for all arrays in 
   the data structures listed below: 
   N array elements are stored in array cells
   array[1]...array[N], array[0] remains void... */

 /* for computing an 'energy' term for particular skeleton
    constituents */
typedef struct _SyVSkeletonEnergy {
  float node,
        len,
        curv,
        dcurv,
        medial,
        recerr;

  float extra0, extra1, extra2;
  float total;
} SyVSkeletonEnergy;

typedef struct _SyVSkeletonBranch {
  SyBranchElementType etype;

  int        id, skel_id, user_id;
  int        hierarchy_level;
  SyBoolean  is_foreground,
             has_boundary;
  float      strength, 
             max_strength, min_strength, mean_strength, var_strength,
             length,
             max_radius, min_radius, mean_radius, var_radius;
             
  int        tail_branch, head_branch,
             tail_node, head_node;

  /* maxelementdim, maxaxisdim... etc. are used for internal 
     allocation purposes -- e.g. maxaxisdim denotes the maximal
     number of element of type union SyGenericPoint that the
     array axis[...] can accomodate. axisdim denotes the 
     actual number of elements that contains valid data. */

  int        elementdim, maxelementdim;

  int        axisdim, maxaxisdim, 
             radiusdim, maxradiusdim,
             refdim, maxrefdim,
             bndrydim, maxbndrydim;

  union SyGenericPoint   *axis;        /* array of axis points */
  float                  *radius;      /* array of local radii */
  SyVSkeletonRef         *ref;         /* array of boundary references */
  SyVSkeletonBndryOffset *left_bndry,  /* array of boundary offset vectors 
                                          for left boundary */
                         *right_bndry; /* array of boundary offset vectors 
                                          for right boundary */

  SyVSkeletonEnergy      energy;
} SyVSkeletonBranch;

typedef struct _SyVSkeletonNode {
  SyBranchElementType etype;
  SyBranchNodeType    ntype;

  int       id, skel_id, user_id;
  int       hierarchy_level;
  SyBoolean is_foreground;
  float     strength,
            radius;
  int       rank, maxrank;      /* number of incident branches */
  int       *inbranch;          /* references to all incident branches */
  union SyGenericPoint loc; 
  SyVSkeletonEnergy energy;
} SyVSkeletonNode;

typedef struct _SyVSkeleton {
  int               id, user_id;
  int               branchdim, nodedim,
                    maxbranchdim, maxnodedim;
  SyBoolean         is_foreground;
  SyVSkeletonBranch *branch;   /* array of skeleton branches */
  SyVSkeletonNode   *node;     /* array of skeleton nodes    */
  SyVSkeletonEnergy energy;
} SyVSkeleton;

typedef struct _SyVSkeletonSet {
  int               maxdim, dim;
  SyBoolean         is_foreground;   /* true if data solely consists of 
                                        foreground skeletons */
  SyVSkeleton       *skel;           /* array of skeletons */
} SyVSkeletonSet;

SyVSkeletonSet* Sy_CopyVSkeletonSet (SyVSkeletonSet* vss); 
/* returns a "deep" copy of 'vss', i.e. including all substructures */

void Sy_DestroySkeletonSet (SyVSkeletonSet* vss); 
/* frees all memory allocated by the substructures 'vss' is pointing
   to -- do not forget to run free on 'vss' ! */

SyVSkeletonSet *Sy_ReadSkeletonSet (FILE *stream);
void Sy_WriteSkeletonSet (FILE *stream, SyVSkeletonSet *vss, 
			  int is_ascii /* bool variable: 0 = no, 1 = yes */ );
#ifdef  __cplusplus
}
#endif
#endif /* _SY_EXPORT_H */



