/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include "Sy_Graphics.h"
#include "vinclude.h"

#include "GraphicsGems.h"

extern void digline(int x1, int y1, int x2, int y2, 
		    void (*dotproc)(int,int));

static unsigned char *ibuf;
static int      ncols, nrows;

void xvdot(int x, int y)
{
  ibuf[x + ncols * y] = 255;
}

void Sy_DigitalLine(int x1, int y1, int x2, int y2)
{
  digline (x1, y1, x2, y2, xvdot); 
}

void Sy_Ipo2Khoros(SyIpoDAC *ipd, struct xvimage *img2)
{
  int      tail_ipo, head_ipo;
  int      i;
  int      x1, x2, y1, y2;

  nrows = (int )img2->col_size;
  ncols = (int )img2->row_size;
  ibuf  = (unsigned char*) img2->imagedata;

  for (i = 1; i <= ipd->ipodim; i++){
    ibuf[ipd->ipo[i].x + (int)SY_XCORR + 
	 (ipd->ipo[i].y + (int)SY_YCORR) * ncols] = 255;
  }
  for (i = 1; i <= ipd->icndim; i++) {
    if ((tail_ipo = ipd->icn[i].tail_ipo) && 
	(head_ipo = ipd->icn[i].head_ipo)){
      x1 = ipd->ipo[tail_ipo].x + SY_XCORR;
      y1 = ipd->ipo[tail_ipo].y + SY_YCORR; 
      x2 = ipd->ipo[head_ipo].x + SY_XCORR;
      y2 = ipd->ipo[head_ipo].y + SY_YCORR;
      digline (x1, y1, x2, y2, xvdot); 
    }
  }
}

void Sy_Voro2Khoros(SyVoroDAC *vd, struct xvimage *img2)
{

  SyVertex vhead, vtail;
  float    x1, y1, x2, y2;
  int      i;

  nrows = (int )img2->col_size;
  ncols = (int )img2->row_size;
  ibuf  = (unsigned char*) img2->imagedata;

  Sy_SetSimpleViewport(ncols-1, nrows-1);
  Sy_SetVectorProc(Sy_DigitalLine);

  for (i = 1; i <= vd->pdim; i++){
    ibuf[vd->point[i].x + (int)SY_XCORR + 
	 (vd->point[i].y + (int)SY_YCORR) * ncols] = 255;
  }

  for (i = 1; i <= vd->edim; i++) {
    vtail = vd->vertex[vd->edge[i].tail_vertex];
    vhead = vd->vertex[vd->edge[i].head_vertex];

    x1 = vtail.xden / vtail.num + SY_XCORR;
    y1 = vtail.yden / vtail.num + SY_YCORR;
    x2 = vhead.xden / vhead.num + SY_XCORR;
    y2 = vhead.yden / vhead.num + SY_YCORR;

    switch (vd->edge[i].rank) {
    case 0:
      Sy_DrawClippedLine(x1, y1, vd->edge[i].dirx, 
			 vd->edge[i].diry); break;
    case 1: 
      Sy_DrawClippedRay(x1, y1, vd->edge[i].dirx, 
			vd->edge[i].diry); break;
    default:
      /*
	 if ((x1 < 0) || (x1 > ncols-1) || (x2 < 0) || (x2 > ncols-1) ||
	 (y1 < 0) || (y1 > nrows-1) || (y2 < 0) || (y2 > nrows-1)) break;
	 digline (x1, y1, x2, y2, xvdot); 
	 */
      Sy_DrawClippedVector (x1, y1, x2, y2);
      break;
    }
  }
}


void Sy_Skel2Khoros(SySkelDAC *vsk, SyIpoDAC *ipd, float thres, 
		    enum SyDisplayMode display_mode, struct xvimage *img2)
{
  SyVertex   vhead, vtail;
  int	   i;
  float    x1, y1, x2, y2;
  int      conn_flag;
  int      conn[4];
  int      tail_vertex, head_vertex, 
           left_point, right_point,
           tail_ipo, head_ipo;
  int      tail_flag, head_flag;
  int      is_residual_cut, is_skel, is_foreground, is_render;

  nrows = (int )img2->col_size;
  ncols = (int )img2->row_size;
  ibuf  = (unsigned char*) img2->imagedata;

  Sy_SetSimpleViewport(ncols-1, nrows-1);
  Sy_SetVectorProc(Sy_DigitalLine);

  for (i = 1; i <= vsk->edim; i++) {
    
    tail_vertex = vsk->edge[i].tail_vertex;
    head_vertex = vsk->edge[i].head_vertex;
    left_point  = vsk->edge[i].left_point;
    right_point = vsk->edge[i].right_point;

    is_foreground  = Sy_SkeletonType (vsk, ipd, tail_vertex, i);

    if ((is_foreground && (display_mode == SKE_BACKGROUND)) ||
        (!is_foreground && (display_mode == SKE_FOREGROUND))) continue;

    tail_flag   = vsk->regul[tail_vertex].birth[2] > thres;
    head_flag   = vsk->regul[head_vertex].birth[2] > thres;

    conn[0]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));

    conn[1]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));

    conn[2]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].head_icn != 0));

    conn[3]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].head_icn != 0));

    conn_flag = (conn[0] || conn[1] || conn[2] || conn[3]);

    
    is_residual_cut = vsk->is_eresid && 
                      ((vsk->eresid[i].tail_residual > thres) || 
                       (vsk->eresid[i].head_residual > thres));

    is_skel         = tail_flag && head_flag;

    is_render = (is_residual_cut || is_skel) && (!conn_flag || vsk->is_border);

    if (is_render){

      if (!(tail_flag && head_flag)){
	Sy_SimpleResidualCut (vsk, i, thres, &vtail, &vhead);
      }
      else{
	vtail = vsk->vertex[vsk->edge[i].tail_vertex];
	vhead = vsk->vertex[vsk->edge[i].head_vertex];
      }

      x1 = vtail.xden / vtail.num + SY_XCORR;
      y1 = vtail.yden / vtail.num + SY_YCORR;
      x2 = vhead.xden / vhead.num + SY_XCORR;
      y2 = vhead.yden / vhead.num + SY_YCORR;

      switch (vsk->edge[i].rank) {
      case 0:
	Sy_DrawClippedLine(x1, y1, vsk->edge[i].dirx, 
			   vsk->edge[i].diry); break;
      case 1: 
	Sy_DrawClippedRay(x1, y1, vsk->edge[i].dirx, 
			  vsk->edge[i].diry); break;
      default:
/*	if ((x1 < 0) || (x1 > ncols-1) || (x2 < 0) || (x2 > ncols-1) ||
	    (y1 < 0) || (y1 > nrows-1) || (y2 < 0) || (y2 > nrows-1)) break;
	digline (x1, y1, x2, y2, xvdot); */
	Sy_DrawClippedVector (x1, y1, x2, y2);
	break;
      }
    }
  }
}


void Sy_SkelPyr2Khoros(SySkelDAC *vsk, SyIpoDAC *ipd, int rank, 
		       enum SyDisplayMode display_mode, 
		       int *pyrlab, struct xvimage *img2)
{
  SyVertex   vhead, vtail;
  int	   i;
  float    x1, y1, x2, y2;
  int      conn_flag;
  int      conn[4];
  int      tail_vertex, head_vertex, 
           left_point, right_point,
           tail_ipo, head_ipo;
  int      tail_flag, head_flag;
  int      is_residual_cut, is_skel, is_foreground, is_render;

  nrows = (int )img2->col_size;
  ncols = (int )img2->row_size;
  ibuf  = (unsigned char*) img2->imagedata;

  Sy_SetSimpleViewport(ncols-1, nrows-1);
  Sy_SetVectorProc(Sy_DigitalLine);

  for (i = 1; i <= vsk->edim; i++) {
    
    tail_vertex = vsk->edge[i].tail_vertex;
    head_vertex = vsk->edge[i].head_vertex;
    left_point  = vsk->edge[i].left_point;
    right_point = vsk->edge[i].right_point;

    is_foreground  = Sy_SkeletonType (vsk, ipd, tail_vertex, i);

    if ((is_foreground && (display_mode == SKE_BACKGROUND)) ||
        (!is_foreground && (display_mode == SKE_FOREGROUND))) continue;

    conn[0]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));

    conn[1]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));

    conn[2]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].head_icn != 0));

    conn[3]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].head_icn != 0));

    conn_flag = (conn[0] || conn[1] || conn[2] || conn[3]);

    is_render = (pyrlab[i] > 0) && (rank >= pyrlab[i]) && 
      (!conn_flag || vsk->is_border);

    if (is_render){

      vtail = vsk->vertex[vsk->edge[i].tail_vertex];
      vhead = vsk->vertex[vsk->edge[i].head_vertex];

      x1 = vtail.xden / vtail.num + SY_XCORR;
      y1 = vtail.yden / vtail.num + SY_YCORR;
      x2 = vhead.xden / vhead.num + SY_XCORR;
      y2 = vhead.yden / vhead.num + SY_YCORR;

      switch (vsk->edge[i].rank) {
      case 0:
	Sy_DrawClippedLine(x1, y1, vsk->edge[i].dirx, 
			   vsk->edge[i].diry); break;
      case 1: 
	Sy_DrawClippedRay(x1, y1, vsk->edge[i].dirx, 
			  vsk->edge[i].diry); break;
      default:
/*	if ((x1 < 0) || (x1 > ncols-1) || (x2 < 0) || (x2 > ncols-1) ||
	    (y1 < 0) || (y1 > nrows-1) || (y2 < 0) || (y2 > nrows-1)) break;
	digline (x1, y1, x2, y2, xvdot); */      
	Sy_DrawClippedVector (x1, y1, x2, y2);
	break;
      }
    }
  }
}






