/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/* 
 * all-purpose memory allocation
 *
 */

#ifdef NOX
int _Xdebug = 0; /* for linking with libImLoad.a in ../loadimage 
                    when -DNOX option is set */
#endif

#if !defined(convex)
#include <memory.h>
#if defined (bsdi) || defined (__bsdi__)
#else
#include <malloc.h>
#endif
#endif

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include "Sy_Idraw.h"

SyLabel* Sy_AllocLabel (int nr_elem)
{
  int      page, label_mem;
  SyLabel* label;
  
  page          = getpagesize();
  label_mem    = ((++nr_elem * sizeof(SyLabel))/page+1)*page;
  label         = (SyLabel*) Sy_Malloc ((unsigned)label_mem);
  (void) memset ((char*) label, 0, label_mem);
  return (label);
}

int* Sy_AllocInt (int nr_elem)
{
  int  page, int_mem;
  int* iptr;
  
  page    = getpagesize();
  int_mem = ((++nr_elem * sizeof(int))/page+1)*page;
  iptr    = (int*) Sy_Malloc ((unsigned)int_mem);
  (void) memset ((char*) iptr, 0, int_mem);
  return (iptr);
}

SyReference* Sy_AllocRef (int nr_elem)
{
  int  page, ref_mem;
  SyReference* ref;
  
  page    = getpagesize();
  ref_mem = ((++nr_elem * sizeof(int))/page+1)*page;
  ref     = (SyReference*) Sy_Malloc ((unsigned)ref_mem);
  (void) memset ((char*) ref, 0, ref_mem);
  return (ref);
}

/**********************************************************************
 * function 
 * void Sy_AllocCratranEnvironment (ima, buff, table, stack, util, info)
 * 
 * purpose: memory allocation for Sy_Cratran-environment,
 *          sets defined initial state
 *
 * param:   [inout] crack   - pointer to Sy_Cratran image structure
 *          [inout] buff    - pointer to Sy_Cratran line buffer
 *          [inout] table   - pointer to Sy_Cratran table
 *          [inout] stack   - pointer to Sy_Cratran stacks
 *          [inout] util    - pointer to auxiliary structures
 *          [inout] info    - pointer to info - structure
 *********************************************************************/
void Sy_AllocCratranEnvironment (SyImage_t *ima, SyCratranBuff *buff, 
				 SyCratranTable *table, SyCratranStack *stack,
				 SyCratranUtil *util, SyCratranInfo *info,
				 SyCratranControl *cntl)
{
  static SyCratranBuff   buff_0  = {{0,0,0,0},0};
  static SyCratranTable  table_0 = {0,0,0};
  static SyCratranStack  stack_0 = {0,0,0,0,0,0,0};
  static SyCratranUtil   util_0  = {0,0,0,0,0};
  
  static int            ones16_0[16] = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
  
  unsigned              alloc_bytes;
  int                   i;
  
  *buff  = buff_0;
  *table = table_0;
  *stack = stack_0;
  *util  = util_0;
	
  if (cntl){
    util->invert_image = cntl->invert_image;
    util->reverse_bits = cntl->reverse_bits;
  }
  
  buff->width = table->width = stack->width = util->width = ima->width;
  
  alloc_bytes = ima->width%8 ?  2*sizeof(int) * (ima->width + 8) :
                                2*sizeof(int) *  ima->width;
  
  info->page_size = getpagesize();
  
  for (i = 0; i < 4; i++){
    buff->line[i]  = (int *) Sy_Malloc(alloc_bytes);
    (void) memset ((char*) buff->line[i], 0, (int) alloc_bytes);
  }
  
  table->arc_   = (SyCratranArc*) Sy_Malloc (2 * ima->width * 
					     sizeof(SyCratranArc));
  table->x_     = (int *) Sy_Malloc (alloc_bytes);
  
  stack->x_     = (SyCarcMemo*) Sy_Malloc (ima->width * 
					   sizeof(SyCarcMemo));
  stack->arc_   = (int *) Sy_Malloc (alloc_bytes);
  stack->index_ = (int *) Sy_Malloc (alloc_bytes);

  util->offset     = (int *) Sy_Malloc (alloc_bytes);
  util->mask       = (int *) Sy_Malloc (alloc_bytes);
  util->blank_line = (char*) Sy_Malloc (ima->linebytes);
  
  switch(ima->depth){
  case  1: (void) memset ((char*)util->blank_line, 
			  util->invert_image ? 255 : 0, ima->linebytes);
			  break;
  case  8:
  case 24: (void) memset ((char*)util->blank_line, util->invert_image ? 
			  255 : 0, ima->linebytes);
			  break;
  }
  
  util->ones16     = ones16_0;
}

void Sy_FreeCratranEnvironment (SyImage_t *ima, SyCratranBuff *buff, 
				SyCratranTable *table, SyCratranStack *stack,
				SyCratranUtil *util, SyCratranInfo *info,
				SyCratranControl *cntl)
{
  int i;

  for (i = 0; i < 4; i++){
    if (buff->line[i]) {
      free(buff->line[i]);
      buff->line[i] = NULL;
    }
  }

  if (table->arc_) {
    free(table->arc_); table->arc_ = NULL;
  }
  if (table->x_) {
    free(table->x_); table->x_ = NULL;
  }

  if (stack->x_) {
    free(stack->x_); stack->x_ = NULL;
  }
  if (stack->arc_) {
    free(stack->arc_); stack->arc_ = NULL;
  }
  if (stack->index_) {
    free(stack->index_); stack->index_ = NULL;
  }
  
  if (util->offset){
    free(util->offset); util->offset = NULL;
  }
  if (util->mask){
    free(util->mask); util->mask = NULL;
  }
  if (util->blank_line){
    free(util->blank_line); util->blank_line = NULL;
  }
}

/**********************************************************************
 * function int Sy_AllocCratranHittab (hit, nr_icns)
 * 
 * purpose: memory allocation for hit-table
 *
 * param:   [inout] hit      - pointer to int
 *          [input] nr_icns  - nr of SyIcns
 *********************************************************************/
int Sy_AllocCratranHittab (int **hit, int nr_icns)
{
  int    page, hit_mem;
  
  page     = getpagesize();
  hit_mem  = ((++nr_icns * sizeof(int))/page + 1) * page;
  *hit     = (int*) Sy_Malloc ((unsigned)hit_mem);
  (void) memset ((char*)*hit, 0, hit_mem);
  return (hit_mem);
}

void Sy_FreeCratranHittab (int **hit)
{
  if (*hit){ free(*hit); *hit = NULL;}
}

/*
 * memory allocation for the Voronoi Diagram routine
 *
 */


#define SY_MIN_STORAGE  8192

/**********************************************************************
 * function 
 * void Sy_AllocVoroEnvironment (vd, comb, hull, support, stack, info)
 * 
 * purpose: memory allocation for VD-environment,
 *          sets defined initial state
 *
 * param:   [inout] vd      - pointer to Voronoi-diagram data
 *          [inout] comb    - pointer to `comb`
 *          [inout] hull    - pointer to `hull`
 *          [inout] support - pointer to `support`-structure
 *          [inout] stack   - pointer to stack
 *          [inout] info    - pointer to general information
 *********************************************************************/

void Sy_AllocVoroEnvironment (SyVoroDAC *vd, SyComb *comb, SyHull *hull,
			      SySupport *support, SyVoroStack *stack, 
			      SyGenInfo *info)
{
  static SyComb	      comb_0  = {0,0,0,0,0,0,0};
  static SyHull       hull_0  = {0,0,0,0,0,0,0,0,0,0};
  static SyVoroStack  stack_0 = {0,0,0,0,0,0,0};
  static SySupport    supp_0  = {0,0,0,0};

  *comb    = comb_0;
  *hull    = hull_0;
  *support = supp_0;
  *stack   = stack_0;

  info->page_size = 2*getpagesize();
  comb->dim       = 2*  info->page_size;
  hull->left_dim  =     info->page_size;
  hull->right_dim =     info->page_size;
  hull->dim       = 2 * info->page_size;
  stack->dim      = 2 * info->page_size;
  comb->max_edges = comb->dim/sizeof(SyEdge) - 1;
  info->max_lines = stack->dim/sizeof(int) - 1;

  comb->elem  = (SyEdge*) Sy_Malloc (comb->dim);
  hull->elem  = (int*)  Sy_Malloc (hull->dim);
  hull->left  = (int*)  Sy_Malloc (hull->left_dim);
  hull->right = (int*)  Sy_Malloc (hull->right_dim);
  stack->e_   = (int*)  Sy_Malloc (stack->dim);
  stack->v_   = (int*)  Sy_Malloc (stack->dim);
  stack->h_   = (int*)  Sy_Malloc (stack->dim);
}


/**********************************************************************
 * function void Sy_ReallocVoroComb (comb, info)
 * 
 * purpose: memory (`comb`) reallocation
 *
 * param:   [inout] comb    - pointer to `comb`
 *          [inout] info    - pointer to general information
 *********************************************************************/
void Sy_ReallocVoroComb (SyComb *comb, SyGenInfo *info)
{
  comb->dim += ((comb->n +1)* sizeof(SyEdge) + 2*info->page_size);
  comb->max_edges  = comb->dim / sizeof(SyEdge)-1; 
  comb->elem = (SyEdge*) Sy_Realloc ((char*)comb->elem, comb->dim);
#ifdef DEBUG 
  Sy_Message(SY_FMTNL, "`comb' structure extended");
#endif
}


/**********************************************************************
 * function void Sy_ReallocVoroHullStack (hull, stack, info)
 * 
 * purpose: memory (`hull` & `stack`) reallocation
 *
 * param:   [inout] hull   -  pointer to `hull`-structure
 *          [inout] stack  -  pointer to stack 
 *          [inout] info   -  pointer to general information
 *********************************************************************/
void Sy_ReallocVoroHullStack (SyHull *hull, SyVoroStack *stack, 
			      SyGenInfo *info)
{
  int i, j, offset, hi_index;

  hull->left_dim  +=     info->page_size;
  hull->right_dim +=     info->page_size;
  
  stack->dim      += 2 * info->page_size;
  info->max_lines  = hull->left_dim / sizeof(int) -1;
  
  hull->left  = (int*) Sy_Realloc ((char*)hull->left,  hull->left_dim);
  hull->right = (int*) Sy_Realloc ((char*)hull->right, hull->right_dim);
  stack->e_   = (int*) Sy_Realloc ((char*)stack->e_, stack->dim);
  stack->v_   = (int*) Sy_Realloc ((char*)stack->v_, stack->dim);
  stack->h_   = (int*) Sy_Realloc ((char*)stack->h_, stack->dim);
  
  hull->dim += 2 * info->page_size;

  hull->elem = (int*) Sy_Realloc ((char*)hull->elem, hull->dim);
  offset     = (hull->dim/sizeof(int) - (hull->right_index - hull->left_index))/2;
  hi_index   = hull->dim/sizeof(int) - offset;

  for (i = hull->right_index, j = hi_index; i >= hull->left_index; i--, j--){
    hull->elem[j] = hull->elem[i];
  }
  hull->right_index = hi_index;
  hull->left_index  = j+1;

#ifdef DEBUG
  Sy_Message(SY_FMTNL, "`stack' & `hull' structures extended");
#endif
}


SyPoint* Sy_AllocPoint (int nrpoints)
{
  SyPoint*  point;
  int     page, p_mem;
  
  page   = getpagesize();
  p_mem  = ((++nrpoints * sizeof(SyPoint))/page + 1) * page;
  point  = (SyPoint*) Sy_Malloc ((unsigned)p_mem);
  return (point);
}



SyEdge* Sy_AllocEdge (int nredges)
{
  SyEdge*  edge;
  int    page, e_mem;
  
  page      = getpagesize();
  e_mem     = ((++nredges * sizeof(SyEdge))/page + 1) * page;
  edge      = (SyEdge*) Sy_Malloc ((unsigned)e_mem);
  return (edge);
}

SyVertex* Sy_AllocVertex (int nrvertices)
{
  SyVertex* vertex;
  int     page, v_mem;
        
  page    = getpagesize();
  v_mem   = ((++nrvertices * sizeof(SyVertex))/page + 1) * page;
  vertex  = (SyVertex*) Sy_Malloc ((unsigned)v_mem);
  return (vertex);
}


void Sy_FreeVoroEnvironment (SyVoroDAC *vd, SyComb *comb, SyHull *hull, 
			     SySupport *support, SyVoroStack *stack, 
			     SyGenInfo *info)
{
  static SyComb	      comb_0  = {0,0,0,0,0,0,0};
  static SyHull       hull_0  = {0,0,0,0,0,0,0,0,0,0};
  static SyVoroStack  stack_0 = {0,0,0,0,0,0,0};
  static SySupport    supp_0  = {0,0,0,0};

  Sy_Free (comb->elem);
  Sy_Free (hull->left);
  Sy_Free (hull->right);
  Sy_Free (hull->elem); 
  Sy_Free (stack->e_);
  Sy_Free (stack->v_);
  Sy_Free (stack->h_);
  
  *comb    = comb_0;
  *hull    = hull_0;
  *support = supp_0;
  *stack   = stack_0;
  
  info->max_lines = 0;
}


SyPoint* Sy_ReallocPoint (SyPoint *point, int nrpoints)
{
  SyPoint*  tmppoint;
  int     page, p_mem;
  
  page      = getpagesize();
  p_mem     = ((++nrpoints * sizeof(SyPoint))/page + 1) * page;
  tmppoint  = (SyPoint*) Sy_Realloc (point, (unsigned)p_mem);
  return (tmppoint);
}

SyEdge* Sy_ReallocEdge (SyEdge *edge, int nredges)
{
  SyEdge*  tmpedge;
  int    page, e_mem;
        
  page      = getpagesize();
  e_mem     = ((++nredges * sizeof(SyEdge))/page + 1) * page;
  tmpedge   = (SyEdge*) Sy_Realloc ((char*)edge, (unsigned)e_mem);
  return (tmpedge);
}

SyVertex* Sy_ReallocVertex (SyVertex *vertex, int nrvertices)
{
  SyVertex* tmpvertex;
  int     page, v_mem;
  
  page      = getpagesize();
  v_mem     = ((++nrvertices * sizeof(SyVertex))/page + 1) * page;
  tmpvertex = (SyVertex*) Sy_Realloc ((char*) vertex, (unsigned)v_mem);
  return (tmpvertex);
}

/*
 * memory allocation for regularization routines
 */

/**********************************************************************
 * function int Sy_AllocRegpot (pot, nr_elem)
 * 
 * purpose: memory allocation for SyRegpot structure
 *
 * param:   [inout] pot      - pointer to SyRegpot structure
 *          [input] nr_elem  - nr of elements
 *********************************************************************/

#define SY_POT_NR_OF_OBJECTS 7000

void Sy_AllocRegpot (SyRegpot *pot, int nr_elem)
{
  int    page, regpot_mem, obj_mem;
  
  page           = getpagesize();
  regpot_mem     = ((++nr_elem * sizeof(SyRegpotElement))/page+1)*page;
  pot->elem_     = (SyRegpotElement*) Sy_Malloc ((unsigned)regpot_mem);
  (void) memset ((char*) pot->elem_, 0, regpot_mem);
  obj_mem        = ((SY_POT_NR_OF_OBJECTS * sizeof(SyObjTab))/page+1)*page;
  pot->obj_tab_  = (SyObjTab*) Sy_Malloc((unsigned)obj_mem);
  pot->obj_closed_ = (int*) Sy_Malloc (SY_POT_NR_OF_OBJECTS * sizeof(int));
  pot->obj_dim   = SY_POT_NR_OF_OBJECTS;
  pot->dim       = nr_elem;
  pot->count     = 0;
  pot->obj_count = 0;
}

void Sy_FreeRegpot (SyRegpot *pot)
{
  Sy_Free(pot->elem_);       pot->elem_       = NULL;
  Sy_Free(pot->obj_tab_);    pot->obj_tab_    = NULL;
  Sy_Free(pot->obj_closed_); pot->obj_closed_ = NULL;

  pot->obj_dim   = 0;
  pot->dim       = 0;
  pot->count     = 0;
  pot->obj_count = 0;
}

SyRegul* Sy_AllocRegul (int nr_elem)
{
  int    page, regul_mem;
  SyRegul* ptr;
  
  page        = getpagesize();
  regul_mem   = ((++nr_elem * sizeof(SyRegul))/page+1)*page;
  ptr         = (SyRegul*) Sy_Malloc ((unsigned)regul_mem);
  memset ((char*) ptr, 0, regul_mem);
  return (ptr);
}

SyEresid* Sy_AllocEresid (int nr_elem)
{
  SyEresid*   eresid;
  int       page, eresid_mem;
  
  page        = getpagesize();
  eresid_mem  = ((++nr_elem * sizeof(SyEresid))/page+1)*page;
  eresid      = (SyEresid*) Sy_Malloc ((unsigned)eresid_mem);
  (void) memset ((char*) eresid, 0, eresid_mem);
  return (eresid);
}

/*
 *
 * memory allocation for the Del. Triangulation
 * 
 */

SyTriEdge* Sy_AllocTriEdge (int nredges)
{
  SyTriEdge*  edge;
  int       page, e_mem;
  
  page      = getpagesize();
  e_mem     = ((++nredges * sizeof(SyTriEdge))/page + 1) * page;
  edge      = (SyTriEdge*) Sy_Malloc ((unsigned)e_mem);
  (void) memset ((char*)edge, 0, e_mem);	
  return (edge);
}

SyTriReg* Sy_AllocTriRegul (int nredges)
{
  SyTriReg*   regul;
  int       page, r_mem;
  
  page      = getpagesize();
  r_mem     = ((++nredges * sizeof(SyTriReg))/page + 1) * page;
  regul     = (SyTriReg*) Sy_Malloc ((unsigned)r_mem);
  (void) memset ((char*)regul, 0, r_mem);	
  return (regul);
}

SyEdgeRef* Sy_AllocEdgeRef (int nredges)
{
  SyEdgeRef   *eref;
  int       page, e_mem;
  
  page      = getpagesize();
  e_mem     = ((++nredges * sizeof(SyEdgeRef))/page + 1) * page;
  eref      = (SyEdgeRef*) Sy_Malloc ((unsigned)e_mem);
  (void) memset ((char*)eref, 0, e_mem);	
  return (eref);
}

SY_P_LABEL* Sy_AllocPointLabel (int nrpoints)
{
  SY_P_LABEL* plabel;
  int      page, plabel_mem;
  
  page        = getpagesize();
  plabel_mem  = ((++nrpoints * sizeof(SY_P_LABEL))/page+1)*page;
  plabel      = (SY_P_LABEL*) Sy_Malloc ((unsigned)plabel_mem);
  (void) memset ((char*)plabel , 0, plabel_mem);
  return (plabel);
}

SY_E_LABEL* Sy_AllocEdgeLabel (int nredges)
{
  SY_E_LABEL* elabel;
  int      page, elabel_mem;
  
  page        = getpagesize();
  elabel_mem  = ((++nredges * sizeof(SY_E_LABEL))/page+1)*page;
  elabel      = (SY_E_LABEL*) Sy_Malloc ((unsigned)elabel_mem);
  (void) memset ((char*)elabel , 0, elabel_mem);
  return (elabel);
}

SY_V_LABEL* Sy_AllocVertexLabel (int nrvertices)
{
  SY_V_LABEL* vlabel;
  int      page, vlabel_mem;
  
  page        = getpagesize();
  vlabel_mem  = ((++nrvertices * sizeof(SY_V_LABEL))/page+1)*page;
  vlabel      = (SY_V_LABEL*) Sy_Malloc ((unsigned)vlabel_mem);
  (void) memset ((char*)vlabel , 0, vlabel_mem);
  return (vlabel);
}

Boolean* Sy_AllocBoolean (int nrelem)
{
  Boolean  *bool;
  int      page, bool_mem;
  
  page      = getpagesize();
  bool_mem  = ((++nrelem* sizeof(Boolean))/page+1)*page;
  bool      = (Boolean*) Sy_Malloc ((unsigned)bool_mem);
  (void) memset ((char*)bool , 0, bool_mem);
  return (bool);
}


SyIpo* Sy_AllocIpo (int nr_ipos)
{
  int    page, ipo_mem;
  SyIpo* ipo;
  page         = getpagesize();
  ipo_mem      = ((++nr_ipos * sizeof(SyIpo))/page + 1) * page;
  ipo          = (SyIpo*) Sy_Malloc (ipo_mem);
  memset ((char*)ipo, 0, ipo_mem);
  return (ipo);
}

SyIpl* Sy_AllocIpl (int nr_ipos)
{
  int    page, ipl_mem;
  SyIpl  *ipl;
  page         = getpagesize();
  ipl_mem      = ((++nr_ipos * sizeof(SyIpl))/page + 1) * page;
  ipl          = (SyIpl*) Sy_Malloc (ipl_mem);
  memset ((char*)ipl, 0, ipl_mem);
  return (ipl);
}

SyIcn* Sy_AllocIcn (int nr_ipos)
{
  int    page, icn_mem;
  SyIcn  *icn;
  page         = getpagesize();
  icn_mem      = ((++nr_ipos * sizeof(SyIcn))/page + 1) * page;
  icn          = (SyIcn*) Sy_Malloc (icn_mem);
  memset ((char*)icn, 0, icn_mem);
  return (icn);
}

SySkContourIndex* Sy_AllocContourIndex (int nr_ipos)
{
  int    page, ci_mem;
  SySkContourIndex *ci;
  page     = getpagesize();
  ci_mem   = ((++nr_ipos * sizeof(SySkContourIndex))/page + 1) * page;
  ci       = (SySkContourIndex*) Sy_Malloc (ci_mem);
  memset ((char*)ci, 0, ci_mem);
  return ci;
}

SyIdrawPoint* Sy_AllocIdrawPoint (int nr_ipos)
{
  int    page, ipo_mem;
  SyIdrawPoint* ipo;
  page         = getpagesize();
  ipo_mem      = ((++nr_ipos * sizeof(SyIdrawPoint))/page + 1) * page;
  ipo          = (SyIdrawPoint*) Sy_Malloc (ipo_mem);
  memset ((char*)ipo, 0, ipo_mem);
  return (ipo);
}

SyIpo* Sy_ReallocIpo (SyIpo* old, int nr_ipos)
{
  int    page, ipo_mem;
  SyIpo* ipo;
  page         = getpagesize();
  ipo_mem      = ((++nr_ipos * sizeof(SyIpo))/page + 1) * page;
  ipo          = (SyIpo*) Sy_Realloc (old, ipo_mem);
  return (ipo);
}

SyIpl* Sy_ReallocIpl (SyIpl* old, int nr_ipos)
{
  int    page, ipl_mem;
  SyIpl  *ipl;
  page         = getpagesize();
  ipl_mem      = ((++nr_ipos * sizeof(SyIpl))/page + 1) * page;
  ipl          = (SyIpl*) Sy_Realloc (old, ipl_mem);
  return (ipl);
}

SyIcn* Sy_ReallocIcn (SyIcn* old, int nr_ipos)
{
  int    page, icn_mem;
  SyIcn  *icn;
  page         = getpagesize();
  icn_mem      = ((++nr_ipos * sizeof(SyIcn))/page + 1) * page;
  icn          = (SyIcn*) Sy_Realloc (old, icn_mem);
  return (icn);
}

SyIdrawPoint* Sy_ReallocIdrawPoint (SyIdrawPoint* old, int nr_ipos)
{
  int    page, ipo_mem;
  SyIdrawPoint* ipo;
  page         = getpagesize();
  ipo_mem      = ((++nr_ipos * sizeof(SyIdrawPoint))/page + 1) * page;
  ipo          = (SyIdrawPoint*) Sy_Realloc (old, ipo_mem);
  return (ipo);
}

static double sy_alloc_bytes = 0;

SY_VOID* Sy_Malloc(int nbytes)
{
  SY_VOID *ptr;
  ptr = malloc (nbytes);
  if (ptr == NULL){
    sprintf (syStrBuf, "Sy_Malloc failed while allocating %d bytes", nbytes);
    Sy_SimpleError (SY_FMTNL, syStrBuf);
    exit(2);
  }
  memset(ptr, 0, nbytes);
  sy_alloc_bytes += nbytes;
  return(ptr);
}

SY_VOID* Sy_Realloc(SY_VOID* old, int nbytes)
{
  SY_VOID *ptr;
  ptr = realloc(old, nbytes);
  if (ptr == NULL){
    sprintf (syStrBuf, "Sy_Realloc failed while allocating %d bytes", nbytes);
    Sy_SimpleError (SY_FMTNL, syStrBuf);
    exit(2);
  }
  sy_alloc_bytes += nbytes;
  return(ptr);
}

void Sy_Free(SY_VOID *ptr) {if (ptr) free(ptr);}












