/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

static char sy_ClipId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"

static int win_width, win_height;
static void (*Sy_VectorProc)(int, int, int, int);

void Sy_SetSimpleViewport (int width, int height)
{
  win_width  = width;
  win_height = height;
}

void Sy_SetVectorProc (void(*funcptr)(int x0, int y0, int x1, int y1))
{
  Sy_VectorProc = funcptr;
}

/**********************************************************************
 * function Sy_DrawClippedVector (x0, y0, x1, y1)
 * 
 * purpose: draws a vector given by the points P=(x0,y0) 
 *          and Q=(x1,y1)
 *
 * param:   [input] x0, y0, x1, y1 
 *********************************************************************/
void Sy_DrawClippedVector(int x0, int y0, int x1, int y1)
{
  SyBoolean is_p0_outside, is_p1_outside;
  
  is_p0_outside = (x0<0)||(x0>win_width)||(y0<0)||(y0>win_height);
  is_p1_outside = (x1<0)||(x1>win_width)||(y1<0)||(y1>win_height);
  
  if (is_p0_outside && is_p1_outside){
    return;
    Sy_DrawClippedLine (x0, y0, x1-x0, y1-y0);}
  else if (is_p0_outside){
    Sy_DrawClippedRay (x1, y1, x0-x1, y0-y1);}
  else if (is_p1_outside){
    Sy_DrawClippedRay (x0, y0, x1-x0, y1-y0);}
  else{
    (*Sy_VectorProc) (x0, y0, x1, y1);}    
}

/**********************************************************************
 * function Sy_DrawClippedRay (x0, y0, dirx, diry)
 * 
 * purpose: draws a ray starting at P=(x0,y0) 
 *          with direction given by R=(dirx,diry)
 *
 * param:   [input] x0, y0, dirx, diry   
 *********************************************************************/
void Sy_DrawClippedRay (int x0, int y0, int dirx, int diry)
{
  float	px = -1.0, py = -1.0, qx = -1.0, qy = -1.0; 
  int     p[8], k = 0;
  float	temp1, temp2;
  
  if ((x0<0)||(x0>win_width)||(y0<0)||(y0>win_height)){
    /* Sy_DrawClippedLine (x0, y0, dirx, diry); */
    return;
  }  
  if (diry != 0) {
    temp1 = (float) dirx / (float) diry;
    px = -y0 * temp1 + (float) x0;
    qx = (win_height - y0) * temp1 + (float) x0;
  }
  if (dirx != 0) {
    temp2 = (float) diry / (float) dirx;
    py = -x0 * temp2 + (float) y0;
    qy = (win_width - x0) * temp2 + (float) y0;
  }
  if ((px >= 0) & (px <= win_width)) {
    p[k] = px;
    p[k+1] = 0;
    k += 2;
  }
  if ((qx >= 0) & (qx <= win_width)) {
    p[k] = qx;
    p[k+1] = win_height;
    k += 2;
  }
  if ((py >= 0) & (py <= win_height)) {
    p[k] = 0;
    p[k+1] = py;
    k += 2;
  }
  if ((qy >= 0) & (qy <= win_height)) {
    p[k] = win_width;
    p[k+1] = qy;
  }
  
  if ((dirx * (p[0] - x0) + diry * (p[1] - y0)) >= 0) {
    (*Sy_VectorProc) (x0, y0, p[0], p[1]);} 
  else{
    (*Sy_VectorProc) (x0, y0, p[2], p[3]);}
}

/**********************************************************************
 * function Sy_DrawClippedLine (x0, y0, dirx, diry)
 * 
 * purpose: draw_line draws a line passing through  
 *          P=(x0,y0) with the direction given by R=(dirx,diry)
 * 
 * input:   [input] x0, y0, dirx, diry    
 *********************************************************************/
void Sy_DrawClippedLine (int x0, int y0, int dirx, int diry)
{
  float	px = -1.0, py = -1.0, qx = -1.0, qy = -1.0;
  int   p[8], k = 0;
  float	temp1, temp2;
  
  if (diry != 0) {
    temp1 = (float) dirx / (float) diry;
    px = -y0 * temp1 + (float) x0;
    qx = (win_height - y0) * temp1 + (float) x0;
  }
  if (dirx != 0) {
    temp2 = (float) diry / (float) dirx;
    py = -x0 * temp2 + (float) y0;
    qy = (win_width - x0) * temp2 + (float) y0;
  }
  if ((px >= 0) & (px <= win_width)) {
    p[k] = px;
    p[k+1] = 0;
    k += 2;
  }
  if ((qx >= 0) & (qx <= win_width)) {
    p[k] = qx;
    p[k+1] = win_height;
    k += 2;
  }
  if ((py >= 0) & (py <= win_height)) {
    p[k] = 0;
    p[k+1] = py;
    k += 2;
  }
  if ((qy >= 0) & (qy <= win_height)) {
    p[k] = win_width;
    p[k+1] = qy;
  }
  
  (*Sy_VectorProc) (p[0], p[1], p[2], p[3]);
}









