/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

static char sy_CompressId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"

typedef struct _SyComprTable {
  int *p_,
      *e_,
      *v_,
      *c_;
  int p_ptr,
      e_ptr,
      v_ptr,
      c_ptr;
} SyComprTable;

static void Sy_FixTriReference(SyTriDAC *tri, int ref_edge, int ref_point, SyTriThres *trith);
static void Sy_FixSkeReference(SySkelDAC *vsk, SyIpoDAC *ipd, int ref_edge, int ref_vertex, float thres);

void Sy_TrianPack (SyTriDAC *tri, SyTriThres trith)
{
  SyComprTable table;

  int head_point, tail_point,
      ref_edge, ref_point;

  int i, is_linked;

  Sy_Message (SY_FMTNL, "compressing triangulation");  

  if (tri->regdim == 0) return;

  table.e_ptr = table.p_ptr = 0;

  table.p_ = Sy_AllocInt (tri->pdim);
  table.e_ = Sy_AllocInt (tri->edim);
  tri->label.p_ = Sy_AllocPointLabel (tri->pdim);

  table.p_[1] = 1; table.p_[2] = 2;
  tri->label.p_[1] = tri->label.p_[2] = 1;
  
  /* build compression tables */

  for (i = 1; i <= tri->edim; i++){

    is_linked = ((tri->regul[i].cont >= trith.cont_thres) &&
		 (tri->regul[i].prox >= trith.prox_thres) &&
		 (tri->regul[i].narr >= trith.narr_thres));
  
    if (!is_linked){
      continue;
    }

    tail_point = tri->edge[i].tail_point;
    head_point = tri->edge[i].head_point;

    if (tri->label.p_[tail_point] == 0){
      Sy_FixTriReference (tri, ref_edge = i, ref_point = tail_point, &trith);
      tri->label.p_[tail_point] = 1;
    }
    if (tri->label.p_[head_point] == 0){
      Sy_FixTriReference (tri, ref_edge = i, ref_point = head_point, &trith);
      tri->label.p_[head_point] = 1;
    }
    
    table.e_[i] = ++table.e_ptr;                /* edge-table entry        */
  }

  /* compress point array */

  for (i = 1; i <= tri->pdim; i++){
    if (tri->label.p_[i] != 0){
      table.p_[i] = ++table.p_ptr;
      tri->point[table.p_ptr] = tri->point[i];
    }
  }

  /* compress edge array */

  for (i = 1; i <= tri->edim; i++){
    if (table.e_[i] != 0){
      tri->edge[table.e_[i]] = tri->edge[i];
    }
  }

  if (tri->erefdim > 0){
    for (i = 1; i <= tri->erefdim; i++){
      if (table.e_[i] != 0){
        tri->eref[table.e_[i]] = tri->eref[i];
      }
    }
    tri->erefdim = table.e_ptr;
  }

  if (tri->regdim > 0){
    for (i = 1; i <= tri->regdim; i++){
      if (table.e_[i] != 0){
	tri->regul[table.e_[i]] = tri->regul[i];
      }
    }
    tri->regdim = table.e_ptr;
  }

  if (tri->is_linkinfo){
    for (i = 1; i <= tri->edim; i++){
      if (table.e_[i] != 0){
	tri->is_hardlink[table.e_[i]] = tri->is_hardlink[i];
      }
    }
  }

  tri->edim = table.e_ptr;
  tri->pdim = table.p_ptr;

  /* modify references */

  for (i = 1; i <= table.e_ptr; i++){

    tri->edge[i].head_edge   = table.e_[tri->edge[i].head_edge];
    tri->edge[i].tail_edge   = table.e_[tri->edge[i].tail_edge];
    tri->edge[i].head_point  = table.p_[tri->edge[i].head_point];
    tri->edge[i].tail_point  = table.p_[tri->edge[i].tail_point];      
  }

  free ((char*) table.e_);
  free ((char*) table.p_);
  free ((char*) tri->label.p_);

  Sy_Message (SY_FMTNL, " done ");

  return;
}

static void Sy_FixTriReference (SyTriDAC *tri, int ref_edge, int ref_point, SyTriThres *trith)
{
  int is_linked,
      last_edge,
      next_edge;

  last_edge = ref_edge;

  next_edge = (tri->edge[ref_edge].tail_point == ref_point) ? 
              tri->edge[ref_edge].tail_edge: tri->edge[ref_edge].head_edge;

  if (next_edge == 0){
    return;
  }

  while ((next_edge != ref_edge) && (next_edge != 0)){

    is_linked = ((tri->regul[next_edge].cont >= trith->cont_thres) &&
		 (tri->regul[next_edge].prox >= trith->prox_thres) &&
		 (tri->regul[next_edge].narr >= trith->narr_thres));
  
    if (is_linked){
      if (tri->edge[last_edge].tail_point == ref_point){
	tri->edge[last_edge].tail_edge = next_edge;
      }
      else{
	tri->edge[last_edge].head_edge = next_edge;
      }
      last_edge = next_edge;
    }

    if (ref_point == tri->edge[next_edge].tail_point){
      next_edge = tri->edge[next_edge].tail_edge;
    }
    else{
      next_edge = tri->edge[next_edge].head_edge;
    }
  }

  if (last_edge == ref_edge){
    if (tri->edge[ref_edge].tail_point == ref_point){
      tri->edge[ref_edge].tail_edge = 0;
    }
    else{
      tri->edge[ref_edge].head_edge = 0;
    }
  }
  else{
    if (tri->edge[last_edge].tail_point == ref_point){
      tri->edge[last_edge].tail_edge = ref_edge;
    }
    else{
      tri->edge[last_edge].head_edge = ref_edge;
    }
  }
}

static Boolean _SKE_retain_border = 1;

void Sy_SetRetainBorder (Boolean bool)
{
  _SKE_retain_border = bool;
}

void Sy_SkeletonPack (SySkelDAC *vsk, SyIpoDAC *ipd, float thres)
{
  SyComprTable table;

  int head_vertex, tail_vertex,
      ref_edge, ref_vertex,
      left_point, right_point;

  int i, is_connected;

  Sy_Message (SY_FMTNL, "compressing skeleton");

  table.e_ptr = table.v_ptr = 0;

  table.v_ = Sy_AllocInt (vsk->vdim);
  table.e_ = Sy_AllocInt (vsk->edim);
  vsk->label.v_ = Sy_AllocVertexLabel (vsk->vdim);
  
  /* build compression tables */

  for (i = 1; i <= vsk->edim; i++){

    if ((vsk->regul[vsk->edge[i].head_vertex].birth[2] < thres) ||
        (vsk->regul[vsk->edge[i].tail_vertex].birth[2] < thres)) continue;
   
    if (!_SKE_retain_border){

      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      
      if (Sy_IpoConnected(ipd, left_point, right_point)) continue;
    }

    tail_vertex = vsk->edge[i].tail_vertex;
    head_vertex = vsk->edge[i].head_vertex;

    if (vsk->label.v_[tail_vertex] == 0){
      Sy_FixSkeReference (vsk, ipd, ref_edge=i, ref_vertex=tail_vertex, thres);
      vsk->label.v_[tail_vertex] = 1;
    }
    if (vsk->label.v_[head_vertex] == 0){
      Sy_FixSkeReference (vsk, ipd, ref_edge=i, ref_vertex=head_vertex, thres);
      vsk->label.v_[head_vertex] = 1;
    }
    
    table.e_[i] = ++table.e_ptr;                /* edge-table entry        */
  }

  /* compress vertex array */

  for (i = 1; i <= vsk->vdim; i++){
    if (vsk->label.v_[i] != 0){
      table.v_[i] = ++table.v_ptr;
      vsk->vertex[table.v_ptr] = vsk->vertex[i];
      vsk->regul[table.v_ptr]  = vsk->regul[i];
    }
  }

  /* compress edge array */

  for (i = 1; i <= vsk->edim; i++){
    if (table.e_[i] != 0){
      vsk->edge[table.e_[i]] = vsk->edge[i];
    }
  }

  if (vsk->erdim > 0){
    for (i = 1; i <= vsk->erdim; i++){
      if (table.e_[i] != 0){
	vsk->eresid[table.e_[i]] = vsk->eresid[i];
      }
    }
    vsk->erdim = table.e_ptr;
  }

  vsk->edim = table.e_ptr;
  vsk->vdim = table.v_ptr;


  /* modify references */

  for (i = 1; i <= table.e_ptr; i++){

    vsk->edge[i].head_edge    = table.e_[vsk->edge[i].head_edge];
    vsk->edge[i].tail_edge    = table.e_[vsk->edge[i].tail_edge];
    vsk->edge[i].head_vertex  = table.v_[vsk->edge[i].head_vertex];
    vsk->edge[i].tail_vertex  = table.v_[vsk->edge[i].tail_vertex];
  }

  free ((char*) table.e_);
  free ((char*) table.v_);
  free ((char*) vsk->label.v_);

  Sy_Message (SY_FMTNL, " done ");

  return;
}

static void Sy_FixSkeReference (SySkelDAC *vsk, SyIpoDAC *ipd, int ref_edge, int ref_vertex, float thres)
{
  int is_skel,
      left_point, right_point,
      last_edge, next_edge;

  last_edge = ref_edge;

  next_edge = (vsk->edge[ref_edge].tail_vertex == ref_vertex) ? 
              vsk->edge[ref_edge].tail_edge: vsk->edge[ref_edge].head_edge;

  if (next_edge == 0){
    return;
  }

  while ((next_edge != ref_edge) && (next_edge != 0)){

    is_skel = ((vsk->regul[vsk->edge[next_edge].head_vertex].birth[2] > thres) &&
              (vsk->regul[vsk->edge[next_edge].tail_vertex].birth[2]  > thres));
   
    if (is_skel && !_SKE_retain_border){

      left_point  = vsk->edge[next_edge].left_point;
      right_point = vsk->edge[next_edge].right_point;
      
      if (Sy_IpoConnected(ipd, left_point, right_point)) is_skel = 0;
    }

    if (is_skel){
      if (vsk->edge[last_edge].tail_vertex == ref_vertex){
	vsk->edge[last_edge].tail_edge = next_edge;
      }
      else{
	vsk->edge[last_edge].head_edge = next_edge;
      }
      last_edge = next_edge;
    }

    if (ref_vertex == vsk->edge[next_edge].tail_vertex){
      next_edge = vsk->edge[next_edge].tail_edge;
    }
    else{
      next_edge = vsk->edge[next_edge].head_edge;
    }
  }

  if (last_edge == ref_edge){
    if (vsk->edge[ref_edge].tail_vertex == ref_vertex){
      vsk->edge[ref_edge].tail_edge = 0;
    }
    else{
      vsk->edge[ref_edge].head_edge = 0;
    }
  }
  else{
    if (vsk->edge[last_edge].tail_vertex == ref_vertex){
      vsk->edge[last_edge].tail_edge = ref_edge;
    }
    else{
      vsk->edge[last_edge].head_edge = ref_edge;
    }
  }
}

void Sy_CleanupIpo (SyIpoDAC *ipd, SyRegpot *pot)
{
  SyComprTable table;

  int head_ipo, tail_ipo,
      head_icn, tail_icn;
  int i, num_del;
  float max_len, len_thres;

  Sy_Message (SY_FMTNL, "cleaning initiating points"); 
  Sy_CleanupIpoSelect (ipd, pot); 

  table.p_ptr = table.c_ptr = 0;
  table.p_ = Sy_AllocInt (ipd->ipodim);
  table.c_ = Sy_AllocInt (ipd->icndim);

  /* compute threshold */

  for (i = 1, max_len = 0.0; i <= pot->obj_count; i++){
    max_len = max (max_len, pot->obj_tab_[i].len);
  }
  
  len_thres = ipd->info->clthreshold * max_len;

  sprintf (syStrBuf, "max. boundary length: %f (len. th.: %f)", 
	   max_len, len_thres);
  Sy_Message (SY_FMTNL, syStrBuf);

  /* build compression table */

  for (i = 1, num_del = 0; i <= ipd->ipodim; i++){

    if (pot->obj_tab_[pot->elem_[i].label].len < len_thres){
      num_del++; continue;
    }
    table.p_[i] = ++table.p_ptr;                /* point table entry */
  }

  if (num_del <= 0) {
    Sy_Message (SY_FMTNL, "no initiating points deleted");
    Sy_Free (table.p_); Sy_Free (table.c_);
    return;
  }

  /* index SyIcn array */

  for (i = 1; i <= ipd->icndim; i++){
    if (table.p_[ipd->icn[i].tail_ipo] == 0 &&
	table.p_[ipd->icn[i].head_ipo] == 0) continue;
    table.c_[i] = ++table.c_ptr;
  }

  /* transform SyIcn array */

  for (i = 1; i <= ipd->icndim; i++){
    if (table.c_[i] != 0){
      ipd->icn[i].tail_ipo = table.p_[ipd->icn[i].tail_ipo];
      ipd->icn[i].head_ipo = table.p_[ipd->icn[i].head_ipo];
    }
  }

  /* transform SyIpl array */

  for (i = 1; i <= ipd->ipodim; i++){
    if (table.p_[i] != 0){
      ipd->ipl[i].tail_icn = table.c_[ipd->ipl[i].tail_icn];
      ipd->ipl[i].head_icn = table.c_[ipd->ipl[i].head_icn];
    }
  }

  /* compress SyIpo/SyIpl arrays */

  for (i = 1; i <= ipd->ipodim; i++){
    if (table.p_[i] != 0){
      ipd->ipo[table.p_[i]] = ipd->ipo[i];
      ipd->ipl[table.p_[i]] = ipd->ipl[i];
    }
  }

  /* compress SyIcn array */

  for (i = 1; i <= ipd->icndim; i++){
    if (table.c_[i] != 0){
      ipd->icn[table.c_[i]] = ipd->icn[i];
    }
  }

  ipd->ipodim = table.p_ptr;
  ipd->icndim = table.c_ptr;

  Sy_Free (table.p_); Sy_Free (table.c_);
  Sy_Message (SY_FMTNL, "compression done");
}
