/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 *
 * High level input/output for skeletons etc...
 *
 */

static char sy_HilevIOId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include "Sy_Lang.h"
#include "Sy_Yacc.h"

static char sy_file_signature[] =
"/* Skeleton Project File */ \n\n";

int Sy_GenLoader (SyIoToken *tok);

void Sy_Strncat (char *dest, char *s1, char *s2, int maxlen);

typedef struct _SyHandlerpair {
  int(*ascii_iohandler) ();
  int(*binary_iohandler)();
} SyHandlerpair;

typedef struct _IOatom {
  char *keyword;
  enum SyStripType tag;
  SyHandlerpair    inp, out;
} SyIOatom;

typedef struct _SyIOmaster {
  SyIOatom* handle;
  int       nhandles;
} SyIOmaster;

enum OpMode {INPUT, OUTPUT, NOP_MODE};

#define SY_MAX_DATA_NR 2048

static SyIOhandle databoard[SY_MAX_DATA_NR];

static int            ndata = 0;
static int            entry = 0;
static int            hptr  = 0;
static int            contourset_entry = -1;
static int            ncontourdata = 0;
static enum OpMode    lastmode = NOP_MODE;
 
static SyIOmaster iomaster;

static char outputmode[3]="w";

static SyIOatom handles[] = 
  {
   "contour",       SY_IPD_STRIP,                 {NULL, NULL}, {NULL, NULL},
   "voronoi",       SY_VORO_STRIP,                {NULL, NULL}, {NULL, NULL},
   "vskeleton",     SY_SKEL_STRIP,                {NULL, NULL}, {NULL, NULL},
   "triangulation", SY_TRIAN_STRIP,               {NULL, NULL}, {NULL, NULL},
   "pointset",      SY_POINT_STRIP,               {NULL, NULL}, {NULL, NULL}, 
   "edgeset",       SY_EDGE_STRIP,                {NULL, NULL}, {NULL, NULL}, 
   "vertexset",     SY_VERTEX_STRIP,              {NULL, NULL}, {NULL, NULL},
   "labelset",      SY_LABEL_STRIP,               {NULL, NULL}, {NULL, NULL},
   "pointlabel",    SY_P_LABEL_STRIP,             {NULL, NULL}, {NULL, NULL}, 
   "edgelabel",     SY_E_LABEL_STRIP,             {NULL, NULL}, {NULL, NULL},
   "vertexlabel",   SY_V_LABEL_STRIP,             {NULL, NULL}, {NULL, NULL}, 
   "nref",          SY_NREF_STRIP,                {NULL, NULL}, {NULL, NULL}, 
   "trianpoint",    SY_TRIPO_STRIP,               {NULL, NULL}, {NULL, NULL}, 
   "trianedge",     SY_TRIEDGE_STRIP,             {NULL, NULL}, {NULL, NULL},
   "trianvertex",   SY_TRIVERTEX_STRIP,           {NULL, NULL}, {NULL, NULL}, 
   "contour_unord", SY_CONTOUR_UNORDERED_STRIP,   {NULL, NULL}, {NULL, NULL}, 
   "skeleton_set",  SY_SKELETON_SET,              {NULL, NULL}, {NULL, NULL}, 
   NULL,            SY_NULL_STRIP,                {NULL, NULL}, {NULL, NULL},
   NULL,            SY_NULL_STRIP,                {NULL, NULL}, {NULL, NULL},
   NULL,            SY_NULL_STRIP,                {NULL, NULL}, {NULL, NULL},
   NULL,            SY_NULL_STRIP,                {NULL, NULL}, {NULL, NULL},
   NULL,            SY_NULL_STRIP,                {NULL, NULL}, {NULL, NULL},
   NULL,            SY_NULL_STRIP,                {NULL, NULL}, {NULL, NULL},
   NULL,            SY_NULL_STRIP,                {NULL, NULL}, {NULL, NULL},
   NULL,            SY_NULL_STRIP,                {NULL, NULL}, {NULL, NULL},
   NULL,            SY_NULL_STRIP,                {NULL, NULL}, {NULL, NULL},
   NULL,            SY_NULL_STRIP,                {NULL, NULL}, {NULL, NULL},
   NULL,            SY_NULL_STRIP,                {NULL, NULL}, {NULL, NULL},
   NULL,            SY_NULL_STRIP,                {NULL, NULL}, {NULL, NULL},
   NULL,            SY_NULL_STRIP,                {NULL, NULL}, {NULL, NULL},
 };
   
static SySkeHeader null_val = {"SY_NIL", "SY_NIL", "SY_NIL", 
			       SY_NONE, 0.0, 0.0, 0}; 

static void Sy_InitIOmaster (void)
{
  int i;
  iomaster.handle = handles;
  
  iomaster.nhandles = sizeof(handles)/sizeof(SyIOatom) - 1;
   
  iomaster.handle[SY_IPD_STRIP ].inp.ascii_iohandler     = Sy_Ipo_fscanf;  
  iomaster.handle[SY_IPD_STRIP ].inp.binary_iohandler    = Sy_Ipo_fread;  
  iomaster.handle[SY_VORO_STRIP].inp.ascii_iohandler     = Sy_Voro_fscanf;
  iomaster.handle[SY_VORO_STRIP].inp.binary_iohandler    = Sy_Voro_fread;
  iomaster.handle[SY_SKEL_STRIP].inp.ascii_iohandler     = Sy_Skel_fscanf;
  iomaster.handle[SY_SKEL_STRIP].inp.binary_iohandler    = Sy_Skel_fread;
  iomaster.handle[SY_TRIAN_STRIP].inp.ascii_iohandler    = Sy_Trian_fscanf;
  iomaster.handle[SY_TRIAN_STRIP].inp.binary_iohandler   = Sy_Trian_fread;

  iomaster.handle[SY_CONTOUR_UNORDERED_STRIP].inp.ascii_iohandler  
                                                         = Sy_Contour_fscanf;

  iomaster.handle[SY_IPD_STRIP ].out.ascii_iohandler     = Sy_Ipo_fprintf;  
  iomaster.handle[SY_IPD_STRIP ].out.binary_iohandler    = Sy_Ipo_fwrite;  
  iomaster.handle[SY_VORO_STRIP].out.ascii_iohandler     = Sy_Voro_fprintf;
  iomaster.handle[SY_VORO_STRIP].out.binary_iohandler    = Sy_Voro_fwrite;
  iomaster.handle[SY_SKEL_STRIP].out.ascii_iohandler     = Sy_Skel_fprintf;
  iomaster.handle[SY_SKEL_STRIP].out.binary_iohandler    = Sy_Skel_fwrite;
  iomaster.handle[SY_TRIAN_STRIP].out.ascii_iohandler    = Sy_Trian_fprintf;
  iomaster.handle[SY_TRIAN_STRIP].out.binary_iohandler   = Sy_Trian_fwrite;

  iomaster.handle[SY_CONTOUR_UNORDERED_STRIP].out.ascii_iohandler  = 
  iomaster.handle[SY_CONTOUR_UNORDERED_STRIP].out.binary_iohandler = 
                                                     Sy_Ipo2ContourSet_fprintf;

  iomaster.handle[SY_SKELETON_SET].out.ascii_iohandler  = 
  iomaster.handle[SY_SKELETON_SET].out.binary_iohandler = 
    Sy_SkeletonSet_fprintf;

  Sy_AvlTreeInit(&sy_yacc_tree);
  Sy_AvlTreeInit(&sy_func_tree);

  Sy_AddIoObject(&sy_func_tree, "callback");
  Sy_AddDataSameObject(&sy_func_tree, "Sy_GenLoader", 
		       SY_LOADFUNC_T, Sy_GenLoader);

  /* ------------------ hack to handle skeleton sets -------------------- */

  Sy_AddDataSameObject(&sy_func_tree, "Sy_ReadIncList", SY_LOADFUNC_T, 
		       Sy_ReadIncList);
  Sy_AddDataSameObject(&sy_func_tree, "Sy_ReadSymmAxis", SY_LOADFUNC_T, 
		       Sy_ReadSymmAxis);
  Sy_AddDataSameObject(&sy_func_tree, "Sy_ReadRadiusList", SY_LOADFUNC_T, 
		       Sy_ReadRadiusList);
  Sy_AddDataSameObject(&sy_func_tree, "Sy_ReadReferences", SY_LOADFUNC_T, 
		       Sy_ReadReferences);
  Sy_AddDataSameObject(&sy_func_tree, "Sy_ReadBndryOffsetVectors", 
		       SY_LOADFUNC_T, Sy_ReadBndryOffsetVectors);

  Sy_AddDataSameObject(&sy_func_tree, "Sy_Load_skeletonnode", 
		       SY_LOADFUNC_T, Sy_Load_skeletonnode);

  Sy_AddDataSameObject(&sy_func_tree, "Sy_Load_skeletonbranch", 
		       SY_LOADFUNC_T, Sy_Load_skeletonbranch);

  Sy_AddDataSameObject(&sy_func_tree, "Sy_Load_skeleton", 
		       SY_LOADFUNC_T, Sy_Load_skeleton);
  /* ------------------------------------------------------------------ */

  Sy_AddIoObject(&sy_yacc_tree, "world");
 
  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;
  sy_yacc_iotoken.func_tree = sy_func_tree;

  for (i = 0; i < SY_MAX_DATA_NR; i++){ 
    databoard[i].index   = i;
    databoard[i].val.numbytes = 0;
  }
    
}

typedef union _SyFakeHeader {
  SyFileHeader header;
  char          fill[SY_HEADER_SIZE];
} SyFakeHeader;

/********************* Generic Loader Routine ****************/

int Sy_GenLoader (SyIoToken *tok)
{
  int is_sync;
  SyDataObject *dobj;

  SyFakeHeader   buff;
  SyFileHeader  *header;
  SySkeHeader    val;
  int            handleptr;
  unsigned int   input_io_mode = 0; /* temporary */
  int            curentry;
  enum SyStripType tag;

#ifdef SY_DEBUG
  printf ("Entering %s...\n", tok->funcname);
#endif

  header = &(buff.header);

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "format");
  if (dobj == NULL){
    Sy_Error ("Sy_HilevIO", "Sy_GenLoader", "missing format info...\n", "");  
  }
  if (strcmp(dobj->t.s_val, "ascii")==0) val.format = SY_ASCII;
  else                                   val.format = SY_BINARY;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "type");
  if (dobj == NULL){
    Sy_Error ("Sy_HilevIO", "Sy_GenLoader", "missing type info...\n", "");  
  }
  strncpy (val.type, dobj->t.s_val, SY_TYPE_LEN);

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "x_extent");
  if (dobj != NULL){
    val.x_extent = dobj->t.f_val;
  }
  else val.x_extent = 0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "y_extent");
  if (dobj != NULL){
    val.y_extent = dobj->t.f_val;
  }
  else val.y_extent = 0;
  
  for (handleptr = 0; handleptr < iomaster.nhandles; handleptr++){
    if (strcmp(iomaster.handle[handleptr].keyword, val.type) == 0){
      tag = iomaster.handle[handleptr].tag;
      break;
    }
  }
  
  if (handleptr >= iomaster.nhandles){
    Sy_Error ("Sy_HilevIO", "Sy_ManageInput", "unknown data type %s\n",
	      val.type);
  }
  
  if (iomaster.handle[handleptr].tag == SY_CONTOUR_UNORDERED_STRIP){
    if (contourset_entry < 0){
      contourset_entry = entry; curentry = entry;
      databoard[curentry].data = NULL; /* no data attached yet */ 
    }
    else{
      curentry = contourset_entry;
    }
    ncontourdata++;
  }
  else curentry = entry;

  databoard[curentry].tag     = iomaster.handle[handleptr].tag;
  databoard[curentry].val     = val;
  databoard[curentry].io_mode = SY_PRESERVE_DATA;
  
  is_sync = Sy_SyncWithStream(yyin, SY_SYNC_COOKIE_BEGIN);
  
  if (is_sync){

    switch (val.format){
    case SY_ASCII:
      if (iomaster.handle[tag].inp.ascii_iohandler == NULL){
	Sy_Error ("Sy_HilevIO", "Sy_ManageInput", 
		  "missing ascii IO routine for type %s\n", val.type); 
      }
      iomaster.handle[tag].inp.ascii_iohandler(yyin, 
					       &databoard[curentry].data, 
					       input_io_mode);
      break;
    case SY_BINARY:
      if (iomaster.handle[tag].inp.binary_iohandler == NULL){
	Sy_Error ("Sy_HilevIO", "Sy_ManageInput",
		  "missing binary IO routine for type %s\n", val.type); 
      }
      iomaster.handle[tag].inp.binary_iohandler(yyin, 
						&databoard[curentry].data, 
						input_io_mode);
      break;
    }

    if (tag != SY_CONTOUR_UNORDERED_STRIP){
      if (entry >= SY_MAX_DATA_NR){
	fprintf (stderr, "max number of IO-handles (> %d) exceeded\n",
		 SY_MAX_DATA_NR);
      }
      else{
	entry++; ndata++;
      }
    }
    else {
      if (ncontourdata == 1) ndata++;
    }

  }/* if sync */  
  else{ return is_sync; }

  is_sync = Sy_SyncWithStream(yyin, SY_SYNC_COOKIE_END);

#ifdef SY_DEBUG
  printf ("Exiting %s...\n", tok->funcname);
#endif
  return is_sync;  
}
/*************************************************************/

void Sy_ManageInput (char *filename)
{
  Sy_ResetIOhandler();
  Sy_InitIOmaster();
  entry  = ndata = 0; 
  
  sy_yacc_iotoken.infilename = filename;
  yyin = (strcmp(filename, "-")==0)? stdin : fopen (filename, "r");
  if (yyin == NULL) 
    Sy_Error ("Sy_HilevIO", "Sy_ManageInput", "open failed on file %s\n", 
	      filename);

  sy_yacc_iotoken.in = yyin;
  yyparse();
  fclose (yyin);   

  Sy_DeleteIoObjectTree(&sy_yacc_tree);
  Sy_DeleteIoObjectTree(&sy_func_tree);

  { /* hack to handle skeleton sets */ 
    SyVSkeletonSet *vss = NULL;
    vss = Sy_GetCurSkeletonSetPtr();
    if (vss != NULL){
      databoard[entry].tag     = SY_SKELETON_SET;
      databoard[entry].data    = (char*) vss;
      databoard[entry].io_mode = SY_PRESERVE_DATA;  

      Sy_ClearCurSkeletonSetPtr ();

      if (entry >= SY_MAX_DATA_NR){
	fprintf (stderr, "max number of IO-handles (> %d) exceeded\n",
		 SY_MAX_DATA_NR);
      }
      else{
	entry++; ndata++;
      }

    }
  }
      
}

static Sy_ManageOutputSlave (FILE *stream)
{
  enum SyStripType tag;
  int              i;

  if (stream == NULL) return;

  Sy_InitIOmaster();
  
  fprintf (stream, "%s", sy_file_signature);

  for (i= 0; i < entry; i++){

    if (databoard[i].tag != SY_CONTOUR_UNORDERED_STRIP &&
        databoard[i].tag != SY_SKELETON_SET){
      fprintf (stream, "object %s {\n", 
	       iomaster.handle[databoard[i].tag].keyword);
  
      switch (databoard[i].val.format){
      case SY_ASCII:  fprintf (stream, "format = ascii;\n"); break;
      case SY_BINARY: fprintf (stream, "format = binary;\n"); break;
      }
    
      fprintf (stream, "type = %s;\n", 
	       iomaster.handle[databoard[i].tag].keyword);
      fprintf (stream, "x_extent = %7.2f;\n", databoard[i].val.x_extent);
      fprintf (stream, "y_extent = %7.2f;\n", databoard[i].val.y_extent);
      fprintf (stream, "numbytes = %d;\n",    databoard[i].val.numbytes);
    }

    tag = databoard[i].tag;

    if (databoard[i].val.format == SY_ASCII){
      if (iomaster.handle[tag].out.ascii_iohandler == NULL){
	Sy_Error ("Sy_HilevIO", "Sy_ManageOutputSlave", 
		   "missing ascii IO routine for type %s\n",
		   iomaster.handle[tag].keyword); 
      }
      iomaster.handle[tag].out.ascii_iohandler(stream, databoard[i].data, 
					       databoard[i].io_mode);
    }
    else{
      if (iomaster.handle[tag].out.binary_iohandler == NULL){
	Sy_Error ("Sy_HilevIO", "Sy_ManageOutputSlave",
		   "missing binary IO routine for type %s\n",
		   iomaster.handle[tag].keyword); 
      }
      iomaster.handle[tag].out.binary_iohandler(stream, databoard[i].data, 
						databoard[i].io_mode);
    }

    if (tag != SY_CONTOUR_UNORDERED_STRIP &&
        tag != SY_SKELETON_SET){
      Sy_Wrendload (stream, SY_FALSE);
      fprintf (stream, "}\n"); 
    }
  }
  entry = ndata = 0;  
  hptr = -1;
}

void Sy_ManageOutput (char *filename)
{
  FILE *stream;

  stream = (strcmp(filename, "-")==0)? stdout : fopen (filename, outputmode);
  if (stream == NULL) 
    Sy_Error ("Sy_HilevIO", "Sy_ManageOutput", "open failed on file %s\n", 
	      filename);

  Sy_ManageOutputSlave (stream);

  fclose(stream);
}

void Sy_ManageOutputStream (FILE *stream)
{
  Sy_ManageOutputSlave (stream);
}

SyIOhandle* Sy_NextIOhandle(void)
{
  SyIOhandle* ioh = NULL;

  hptr++;
  if (hptr >= ndata){ 
    return ((SyIOhandle*) NULL);
  }
  else{
    ioh = (SyIOhandle*) Sy_Malloc (sizeof(SyIOhandle));
    *ioh = databoard[hptr];
    return ioh;
  }
}

SyIOhandle* Sy_GetIOhandle(int index)
{
  SyIOhandle* ioh = NULL;

  if (index >= ndata || index < 0){ 
    return ((SyIOhandle*) NULL);
  }
  else{
    ioh = (SyIOhandle*) Sy_Malloc (sizeof(SyIOhandle));
    *ioh = databoard[index];
    return ioh;
  }
}

void Sy_PutIOhandle (SyIOhandle *handle)
{ 
  if (lastmode != OUTPUT){
    Sy_ResetIOhandler();
    lastmode = OUTPUT;
  }
  if (entry >= SY_MAX_DATA_NR){
    fprintf (stderr, "max number of IO-handles (> %d) exceeded\n",
	             SY_MAX_DATA_NR);
    return;
  }
  databoard[entry] = *handle;
  entry++;
}

void Sy_ResetIOhandler (void)
{  
  int i;

  hptr   = -1;
  entry  = 0;
  ndata  = 0;

  memset (databoard, 0, sizeof(SyIOhandle)*SY_MAX_DATA_NR);
  for (i = 0; i < SY_MAX_DATA_NR; i++){ 
    Sy_ResetIOhandle (&databoard[i]);
  }

}

void Sy_SetOutputMode (char *str)
{
  strncpy (outputmode, str, 2);
}

void Sy_SetObject (SyIOhandle *handle, char *object)
{
  strncpy (handle->val.object, object, SY_OBJECT_LEN);
}

void Sy_SetParent (SyIOhandle *handle, char *parent)
{
  strncpy (handle->val.parent, parent, SY_PARENT_LEN);
}

void Sy_SetFormat (SyIOhandle *handle, enum SySkeDataFormat format)
{
  handle->val.format = format;
}

void Sy_SetXExtent (SyIOhandle *handle, float x_extent)
{
  handle->val.x_extent = x_extent;
}

void Sy_SetYExtent (SyIOhandle *handle, float y_extent)
{
  handle->val.y_extent = y_extent;
}

void Sy_SetNumbytes (SyIOhandle *handle, int numbytes)
{
  handle->val.numbytes = numbytes;
}

void Sy_SetTag (SyIOhandle *handle, enum SyStripType tag)
{
  handle->tag = tag;
}

void Sy_SetData (SyIOhandle *handle, char *data)
{
  handle->data = data;
}

char* Sy_GetObject (SyIOhandle *handle)
{
  char* retval;
  retval = (char*) Sy_Malloc ((SY_OBJECT_LEN+1)*sizeof(char));
  strncpy (retval, handle->val.object, SY_OBJECT_LEN);
  return(retval);
}

char* Sy_GetParent (SyIOhandle *handle)
{
  char* retval;
  retval = (char*) Sy_Malloc ((SY_PARENT_LEN+1)*sizeof(char));
  strncpy (retval, handle->val.parent, SY_PARENT_LEN);
  return(retval);
}

char* Sy_GetType (SyIOhandle *handle)
{
  char* retval;
  retval = (char*) Sy_Malloc ((SY_TYPE_LEN+1)*sizeof(char));
  strncpy (retval, handle->val.type, SY_TYPE_LEN);
  return(retval);
}

enum SyStripType SKE_get_tag (SyIOhandle *handle)
{
  return (handle->tag);
}

enum SySkeDataFormat SKE_get_format (SyIOhandle *handle)
{
  return(handle->val.format);
}

float Sy_GetXExtent (SyIOhandle *handle)
{
  return (handle->val.x_extent);
}

float Sy_GetYExtent (SyIOhandle *handle)
{
  return (handle->val.y_extent);
}

int Sy_GetNumBytes (SyIOhandle *handle)
{
  return (handle->val.numbytes);
}

char* Sy_GetData (SyIOhandle *handle)
{
  return (handle->data);
}

SyIOhandle* Sy_CreateIOhandle (void)
{
  SyIOhandle* retval;
  retval = (SyIOhandle*) Sy_Malloc (sizeof(SyIOhandle));
  Sy_ResetIOhandle (retval);
  return (retval);
}

void Sy_DestroyIOhandle (SyIOhandle *handle)
{
  if (handle != NULL) free(handle);
}

void Sy_ResetIOhandle (SyIOhandle *handle)
{
  handle->tag     = SY_NULL_STRIP;
  handle->data    = (char*) NULL;
  handle->io_mode = SY_PRESERVE_DATA;
  handle->val     = null_val;
}

void Sy_SetIOproperty (SyIOhandle *handle, enum SyIOpropertyId property, 
		       SY_VOID *val)
{
  switch (property){
  case OBJECT  : strncpy (handle->val.object, val, SY_OBJECT_LEN); break;
  case PARENT  : strncpy (handle->val.parent, val, SY_PARENT_LEN); break;
  case FORMAT  : handle->val.format   = (enum SySkeDataFormat) val; break;
  case TYPE    : handle->tag          = (enum SyStripType)  val; break;
  case DATA    : handle->data         = val;                      break;
  case X_EXTENT: handle->val.x_extent = (float) (int) val;        break;
  case Y_EXTENT: handle->val.y_extent = (float) (int) val;        break;
  case NUMBYTES: handle->val.numbytes = (int) val;                break;
  case IO_TYPE : strncpy (outputmode, val, 2);                    break;
  case IO_MODE : handle->io_mode      = (unsigned int) val;       break;  
  }
}

SyIOpropertyValue  Sy_GetIOproperty (SyIOhandle *handle, 
				     enum SyIOpropertyId property)
{  
  SyIOpropertyValue retval;

  switch (property){
  case OBJECT  : 
    retval.object = (char*) Sy_Malloc ((SY_OBJECT_LEN+1)*sizeof(char));
    strncpy (retval.object, handle->val.object, SY_OBJECT_LEN);
    break;
  case PARENT  : 
    retval.parent = (char*) Sy_Malloc ((SY_PARENT_LEN+1)*sizeof(char));
    strncpy (retval.parent, handle->val.parent, SY_PARENT_LEN);
    break;
  case IO_TYPE : 
    retval.io_type   = (char*) Sy_Malloc (8*sizeof(char));
    strncpy (retval.io_type, outputmode, 7);
    break;
  case IO_MODE : retval.io_mode  = handle->io_mode;      break;
  case FORMAT  : retval.format   = handle->val.format;   break;
  case TYPE    : retval.type     = handle->tag;          break;
  case DATA    : retval.data     = handle->data;         break;
  case X_EXTENT: retval.x_extent = handle->val.x_extent; break;
  case Y_EXTENT: retval.y_extent = handle->val.y_extent; break;
  case NUMBYTES: retval.numbytes = handle->val.numbytes; break;
  case IO_INDEX: retval.index    = handle->index;        break;
  }
  return (retval);
}

void Sy_Strncat (char *dest, char *s1, char *s2, int maxlen)
{
  int   rlen, blen;
  char* buf;

  blen = strlen(s1) + strlen(s2);
  buf  = (char*) Sy_Malloc ((blen+2) * sizeof(char));
  rlen = (blen > maxlen) ? maxlen : blen;
  strcpy (buf, s1);
  strcat (buf, s2);
  strncpy (dest, buf, rlen);
  if (rlen < blen){
    Sy_Error ("Sy_HilevIO", "Sy_Strncat", "String truncated: %s\n", buf);
  }
  free(buf);
} 
