/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Division of Applied Sciences,
 *          Harvard University, Cambridge, Massachusetts (rlo@hrl.harvard.edu)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1995 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#include "Sy_AvlTree.h"
#include "Sy_Lang.h"

SyIoObject *sy_global_obj_mem = NULL;

char *Sy_Strcpy (char *oldstr)
{
  char *newstr;
  int len;
  len = strlen(oldstr)+1;
  newstr = (char*) malloc (sizeof(char)*len);
  memcpy(newstr, oldstr, len);
  return newstr;
}

int Sy_SyncWithStream(FILE *stream, char* cookie)
{
  char *buf, ch;
  int len, found = 0, bufptr = 0;
  len = strlen(cookie);
  buf = (char*) malloc (sizeof(char)*(len+1));
  strcpy(buf, cookie);
  while (!feof(stream) && !found){
    ch = getc(stream);
    if (ch == buf[bufptr]) bufptr++; else bufptr = 0;
    if (bufptr >= len) { found = 1; break; }
  }
  free (buf);
#ifdef SY_DEBUG
    printf (found ? "synchronized\n" : "could not synchronize\n");
#endif
  return found;
}

SyIoObject *Sy_CreateIoObject(char *str, int in_id)
{
  SyIoObject *tmp;
  tmp = (SyIoObject*) malloc (sizeof(SyIoObject));
  tmp->id = in_id;
  Sy_AvlTreeInit (&tmp->tree);
  tmp->name = Sy_Strcpy(str);
  return tmp;
}

SyDataObject *Sy_CreateDataObject(char *str, int in_id, 
				SyIoDataType ty, void *val)
{
  SyDataObject *tmp;
  tmp = (SyDataObject*) malloc (sizeof(SyDataObject));
  tmp->id   = in_id;
  tmp->name = Sy_Strcpy(str);
  tmp->tag  = ty;
  switch (ty){
  case SY_INT_T:      tmp->t.i_val    = *(int*)val;   break;
  case SY_FLOAT_T:    tmp->t.f_val    = *(float*)val; break;
  case SY_DOUBLE_T:   tmp->t.d_val    = *(double*)val; break;
  case SY_PTR_T:      tmp->t.p_val    = val; break;
  case SY_STR_T:      tmp->t.s_val    = Sy_Strcpy((char*)val); break;
  case SY_LOADFUNC_T: tmp->t.func_val = val; break;
  }
  return tmp;
}

int Sy_DataObjectInfo (void *din)
{
  SyDataObject *d;

  if (din == NULL){
    fprintf (stderr, "Warning: no information on nil pointer.\n");
    return;
  }

  d = (SyDataObject*) din;
  printf ("data %s (id:%d), val: ", d->name, d->id);
  switch (d->tag){
  case SY_INT_T:      printf ("%d",  d->t.i_val); break;
  case SY_FLOAT_T:    printf ("%f",  d->t.f_val); break;
  case SY_DOUBLE_T:   printf ("%lf", d->t.d_val); break;
  case SY_PTR_T:      printf ("%x",  d->t.s_val); break;
  case SY_STR_T:      printf ("%s",  d->t.p_val); break;
  case SY_LOADFUNC_T: printf ("func at %x", d->t.func_val); break;
  }
  printf ("\n");
  return(1);
}

int Sy_IoObjCompare (void* l, void* r)
{
  return strcmp((char*)l, (char*)r);
}

int Sy_DataObjCompare (void* l, void* r)
{
  return strcmp((char*)l, (char*)r);
}

int Sy_DataObjDelete(void* d)
{
  SyDataObject *dobj;
  dobj = (SyDataObject*) d;
#ifdef DEBUG
  printf ("deleting data: %s\n", dobj->name);
#endif
  free(dobj->name);
  switch (dobj->tag){
  case SY_STR_T: free(dobj->t.s_val); break;
  } 
  free(dobj);
  return (1);
}

int Sy_IoObjDelete (void* l)
{
  SyIoObject *lo;
  lo = (SyIoObject*) l;
#ifdef DEBUG
  printf ("deleting: %s\n", lo->name);
#endif
  if (lo->tree != NULL){
    Sy_AvlTreeMung(&lo->tree, Sy_DataObjPurge);   
  }
  free(lo->name);
  free(lo);
  return (1);
}

int Sy_DataObjPurge (void *data, void *key)
{
  Sy_DataObjDelete(data);
  if (key != NULL){
    free ((char*)key);
  }
  return(1);
}

int Sy_IoObjPurge (void *l, void *key)
{
  SyIoObject *lo;
  lo = (SyIoObject*) l;
#ifdef DEBUG
  printf ("deleting: %s\n", lo->name);
#endif
  if (lo->tree != NULL){
    Sy_AvlTreeMung(&lo->tree, Sy_DataObjPurge);   
  }
  free(lo->name);
  free(lo);
  if (key != NULL){
    free ((char*)key);
  }
  return(1);
}

int Sy_IoObjPrint (void* l)
{
  SyIoObject *lo;
  lo = (SyIoObject*) l;
  printf ("object: %s\n", lo->name);
  return (1);
}

int Sy_IoObjInfo (void* l)
{
  SyIoObject *lo;
  lo = (SyIoObject*) l;
  printf ("object: %s\n", lo->name);
  if (lo->tree != NULL){
    Sy_AvlTreeTrav(&lo->tree, Sy_DataObjectInfo);
  }
  return (1);
}

int Sy_AddIoObject (SyAvlTree **objtree, char *objname)
{
  SyIoObject *objptr;
  char *tmpkey;
  int new_node;

  objptr = Sy_CreateIoObject(objname, 1);
  tmpkey = Sy_Strcpy(objname);

  new_node = 
    Sy_AvlTreeAdd(objtree, Sy_IoObjCompare, objptr, tmpkey, Sy_IoObjDelete);

  if (!new_node) free(tmpkey);

  sy_global_obj_mem = 
    (SyIoObject*) Sy_AvlTreeSrch (objtree, Sy_IoObjCompare, objname);
  return (1);
}

int Sy_AddDataObject (SyAvlTree **objtree, char *objname, 
		      char *dataname, SyIoDataType ty, void *val)
{
  SyIoObject   *ret;
  SyDataObject *dobj;
  char *tmpkey;
  int new_node;
  
  ret = (SyIoObject*) Sy_AvlTreeSrch (objtree, Sy_IoObjCompare, objname);

  if (ret == NULL){
    fprintf (stderr, "Warning: Object `%s' not found...\n", objname);
    return (0);
  }
  sy_global_obj_mem = ret;

  dobj = Sy_CreateDataObject(dataname, 1, ty, val);
  tmpkey = Sy_Strcpy(dataname);
  
  new_node = Sy_AvlTreeAdd(&ret->tree, Sy_DataObjCompare, dobj, tmpkey, 
			   Sy_DataObjDelete);

  if (!new_node) free (tmpkey);

  return(1);
}

int Sy_AddDataSameObject (SyAvlTree **objtree, char *dataname, 
			  SyIoDataType ty, void *val)
{
  SyDataObject *dobj;
  char* tmpkey;
  int new_node;

  if (sy_global_obj_mem == NULL){
    fprintf (stderr, "Warning: Last object not found...\n");
    return (0);
  }
  dobj = Sy_CreateDataObject(dataname, 1, ty, val);
  tmpkey = Sy_Strcpy(dataname);
  
  new_node = Sy_AvlTreeAdd(&sy_global_obj_mem->tree, Sy_DataObjCompare, 
			   dobj, tmpkey, Sy_DataObjDelete);
  if (!new_node) free(tmpkey);

  return(1);
}

SyIoObject* Sy_GetIoObject (SyAvlTree **objtree, char *objname)
{
  SyIoObject   *ret;
  ret = (SyIoObject*) Sy_AvlTreeSrch(objtree, Sy_IoObjCompare, objname);
  if (ret == NULL){
    fprintf (stderr, "Warning: Object `%s' not found...\n", objname);
    return NULL;
  }
  sy_global_obj_mem = ret;
  return ret;
}

SyDataObject* Sy_GetDataObject (SyAvlTree **objtree, char *objname, 
		      char *dataname)
{
  SyIoObject   *ret;
  SyDataObject *dret;

  ret = (SyIoObject*) Sy_AvlTreeSrch (objtree, Sy_IoObjCompare, objname);

  if (ret == NULL){
    fprintf (stderr, "Warning: Object `%s' not found...\n", objname);
    return NULL;
  }

  sy_global_obj_mem = ret;

  dret = (SyDataObject*) Sy_AvlTreeSrch (&ret->tree, Sy_DataObjCompare, 
					 dataname);
  if (dret == NULL){
#ifdef DEBUG
    fprintf (stderr, "Warning: Data `%s' in object `%s' not found...\n", 
	     dataname, objname);
#endif
    return NULL;
  }
  return dret;
}

SyDataObject* Sy_GetDataSameObject (SyAvlTree **objtree, char *dataname)
{
  SyDataObject *dret;

  if (sy_global_obj_mem == NULL){
    fprintf (stderr, "Warning: Last object not found...\n");
    return NULL;
  }

  dret = (SyDataObject*) Sy_AvlTreeSrch (&sy_global_obj_mem->tree, 
					 Sy_DataObjCompare, 
					 dataname);
  if (dret == NULL){
#ifdef DEBUG
    fprintf (stderr, "Warning: Data `%s' in object `%s' not found...\n", 
	     dataname, sy_global_obj_mem->name);
#endif
    return NULL;
  }
  return dret;
}

void Sy_DeleteIoObjectTree(SyAvlTree **objtree)
{
  Sy_AvlTreeMung(objtree, Sy_IoObjPurge);
}


void Sy_IoObjectTreeInfo (SyAvlTree **objtree)
{
  Sy_AvlTreeTrav (objtree, Sy_IoObjInfo);
}

void Sy_Wrbegload(FILE *out, char* callee, int no_cookie, int is_ascii)
{
  if (no_cookie){
    fprintf (out, "beginloadseq (%s)\n", callee);
  }
  else{
    fprintf (out, is_ascii ? "beginloadseq (%s)\n%s\n" : "beginloadseq (%s)\n%s", 
	     callee, SY_SYNC_COOKIE_BEGIN);
  }
}

void Sy_Wrendload(FILE *out, int no_cookie)
{
  if (no_cookie){
    fprintf (out, "\n%s\nendloadseq;\n");
  }
  else {
    fprintf (out, "\n%s\nendloadseq;\n", SY_SYNC_COOKIE_END);
  }
}

#if defined(linux) || defined (LinuxArchitecture) 
#ifndef HAS_YYERROR
void yyerror (char* errstring){
  fprintf(stderr, "%s\n", errstring);
}
#endif
#endif
