/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 *
 * regularization routines for Voronoi skeletons
 *
 */

static char sy_RegularId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"

enum SyScanDirection {FORWARD, BACKWARD};

enum SyResidualCompMode     residualCompMode   = RES_NORMAL;
enum SyResidualBorderMode   residualBorderMode = RES_NOBORDER;

Boolean is_residual_fix = 0,
        is_maxfix       = 0;

void Sy_CirResidualBase (SySkelDAC *vsk, SyIpoDAC *ipd, SyRegpot *pot, 
			 int is_verbose, enum SyPruneMethod flavor);

void Sy_IpoTrace (SyIpoDAC *ipd, SyRegpot *pot, int start_ipo, 
		  int start_icn, int nr, enum SyScanDirection dir);

void Sy_QuickIpoTrace(SyIpoDAC *ipd, SyRegpot *pot, int start_ipo, 
		      int start_icn, int nr, enum SyScanDirection dir);
 
void Sy_ComPot(SyIpoDAC *ipd, SyRegpot *pot, int is_verbose, 
	       void (*TraceMethod)());

static void Sy_Insert(SySkelDAC *vsk, int vindex, float dev, float rad);

void Sy_Regularize (SySkelDAC *vsk, SyIpoDAC *ipd, 
		    void (*regul_method)(), int is_verbose)
{
  SyRegpot  pot;
  int i;

  if (residualBorderMode == RES_BORDER){
    Sy_SkelBorder (vsk, ipd, is_verbose);
  }
  else{
    vsk->is_border = 0;
  }

  Sy_Message (SY_FMT, "Pruning: ");

  Sy_AllocRegpot (&pot, ipd->ipodim);

  if (vsk->regul != NULL) Sy_Free(vsk->regul);
  vsk->regul = Sy_AllocRegul (vsk->vdim);

  if (residualCompMode == RES_FULL){ 
    if (vsk->eresid != NULL) Sy_Free(vsk->eresid);
    vsk->eresid    = Sy_AllocEresid (vsk->edim);
    vsk->erdim     = vsk->edim;
    vsk->is_eresid = SY_TRUE;
  }
  else{ 
    if (vsk->eresid != NULL) Sy_Free(vsk->eresid);
    vsk->eresid    = (SyEresid*) NULL;
    vsk->erdim     = 0;
    vsk->is_eresid = SY_FALSE;
  }

  Sy_ComPot (ipd, &pot, is_verbose, Sy_IpoTrace);

  Sy_Message (SY_FMT, "residual = ");

  (*regul_method) (vsk, ipd, &pot, is_verbose);

  if (is_residual_fix) Sy_Infix  (vsk, ipd);
  if (is_maxfix)       Sy_Maxfix (vsk, ipd);

  if (ipd->rpe != NULL) Sy_Free(ipd->rpe);
  ipd->rpe = pot.elem_; pot.elem_ = NULL;
  Sy_FreeRegpot(&pot);
}

void Sy_ComPot (SyIpoDAC *ipd, SyRegpot *pot, int is_verbose, 
		void (*TraceMethod) ())
{ 
  void ipo_trace();
  int i, obj_count = 0;

  for (i = 1; i <= ipd->ipodim; i++){
    if (pot->elem_[i].label == 0){
      (*TraceMethod)(ipd, pot, i, ipd->ipl[i].tail_icn, ++obj_count, FORWARD);
    }
  }
  if (syIsVerbose) 
    sprintf (syStrBuf, "(tracked %d contour[s]), ", obj_count);
  Sy_Message (SY_FMT, syStrBuf);
}

void Sy_QuickIpoTrace (SyIpoDAC *ipd, SyRegpot *pot, int start_ipo, 
		       int start_icn, int nr, enum SyScanDirection dir)
{
  int   cur_ipo, cur_icn; 
  int   next_ipo, next_icn, dlen;
  float cur_len;
  SyIpo   cur_pos, next_pos;
  
  cur_ipo = start_ipo;
  cur_pos = ipd->ipo[cur_ipo];
  cur_len = 0.0;

  do{
    pot->elem_[cur_ipo].val   = cur_len;
    pot->elem_[cur_ipo].label = nr;

    if (ipd->icn[cur_icn].tail_ipo == cur_ipo){ 
      next_ipo = ipd->icn[cur_icn].head_ipo;
    }
    else{
      next_ipo = ipd->icn[cur_icn].tail_ipo;
    }

    if (ipd->ipl[next_ipo].tail_icn == cur_icn){
      next_icn = ipd->ipl[next_ipo].head_icn;
    }
    else{
      next_icn = ipd->ipl[next_ipo].tail_icn;
    }
    next_pos = ipd->ipo[next_ipo];

    dlen = SY_SQUARE(cur_pos.x-next_pos.x) + SY_SQUARE(cur_pos.y-next_pos.y);

    switch (dlen){
       case 1 :  cur_len++; break;
       case 2 :  cur_len += SY_SQRT2; break;
       default:  cur_len += sqrt ((float) dlen);
    }

    cur_ipo = next_ipo;
    cur_icn = next_icn;
    cur_pos = next_pos;

  } while (pot->elem_[cur_ipo].label == 0);

  pot->obj_count        = nr;
  pot->obj_tab_[nr].len = cur_len;
}

void Sy_IpoTrace (SyIpoDAC *ipd, SyRegpot *pot, int start_ipo, 
		  int start_icn, int nr, enum SyScanDirection dir)
{
  int   cur_ipo, cur_icn; 
  int   next_ipo, next_icn, dlen;
  float cur_len;
  SyIpo   cur_pos, next_pos;
  
  cur_ipo = start_ipo;
  cur_icn = start_icn;
  cur_pos = ipd->ipo[cur_ipo];
  cur_len = 0.0;

  do{
    pot->elem_[cur_ipo].val   = (dir == FORWARD) ? cur_len: -cur_len;
    pot->elem_[cur_ipo].label = nr;

    if (ipd->icn[cur_icn].tail_ipo == cur_ipo){ 
      next_ipo = ipd->icn[cur_icn].head_ipo;
    }
    else{
      next_ipo = ipd->icn[cur_icn].tail_ipo;
    }

    if (ipd->ipl[next_ipo].tail_icn == cur_icn){
      next_icn = ipd->ipl[next_ipo].head_icn;
    }
    else{
      next_icn = ipd->ipl[next_ipo].tail_icn;
    }
    next_pos = ipd->ipo[next_ipo];

    dlen = SY_SQUARE(cur_pos.x-next_pos.x) + SY_SQUARE(cur_pos.y-next_pos.y);

    switch (dlen){
       case 1 :  cur_len++; break;
       case 2 :  cur_len += SY_SQRT2; break;
       default:  cur_len += sqrt ((float) dlen);
    }
    cur_ipo = next_ipo;
    cur_icn = next_icn;
    cur_pos = next_pos;

  } while ((pot->elem_[cur_ipo].label == 0) && (next_ipo != 0));

  pot->obj_count     = nr;

  if (dir == FORWARD){
    pot->obj_tab_[nr].len = cur_len;
    pot->obj_closed_[nr]  = SY_TRUE;
  }
  else{                
    pot->obj_tab_[nr].len = 0.0;
    pot->obj_closed_[nr]  = SY_FALSE;
  }

  if ((dir == FORWARD) && (next_ipo == 0)){
    if (start_icn == ipd->ipl[start_ipo].tail_icn){
      start_icn = ipd->ipl[start_ipo].head_icn;
    }
    else{
      start_icn = ipd->ipl[start_ipo].tail_icn;
    }
    Sy_IpoTrace (ipd, pot, start_ipo, start_icn, nr, BACKWARD);
  }
}

void Sy_CorrResidEdge (SySkelDAC *vsk, SyIpoDAC *ipd, SyRegpot *pot, 
		       SyVertex tvertex, SyVertex hvertex, SyVector tvec0, 
		       SyVector hvec0, int left_point, int right_point, 
		       int ref_point, int *tcorr, int *hcorr, float *dist, 
		       int *is_mindev)
{
  SyVector evec, tvec, hvec, flow, normal, hivec, lovec;
  int    cur_ipo, o_point, tail_ipo, head_ipo;
  float  dist0;
  int    hidist = 0, lodist = 0;
  int    tail_icn, head_icn;
  int    label;
  double hiproj, loproj, refproj;
  double hinorm, lonorm;

  *is_mindev = 0;

  evec.dirx = hvertex.xden/hvertex.num - tvertex.xden/tvertex.num;
  evec.diry = hvertex.yden/hvertex.num - tvertex.yden/tvertex.num;

  cur_ipo   = ref_point;
  if (cur_ipo == left_point) o_point = right_point; else o_point = left_point;

  head_icn  = ipd->ipl[cur_ipo].head_icn;
  tail_icn  = ipd->ipl[cur_ipo].tail_icn;
  label     = pot->elem_[cur_ipo].label;
  
  if (head_icn != 0){    
    head_ipo = (ipd->icn[head_icn].tail_ipo == cur_ipo) ? 
      ipd->icn[head_icn].head_ipo : ipd->icn[head_icn].tail_ipo;
  } 
  else{
    head_ipo   = 0;
    *is_mindev = 1;
    *tcorr = *hcorr = 0;
    return;
  }

  if (tail_icn != 0){    
    tail_ipo = (ipd->icn[tail_icn].tail_ipo == cur_ipo) ? 
      ipd->icn[tail_icn].head_ipo : ipd->icn[tail_icn].tail_ipo;
  } else {
    tail_ipo = 0;
    *is_mindev = 1;
    *tcorr = *hcorr = 0;
    return;
  }

  dist0 = SY_ALV(pot->elem_[left_point].val - pot->elem_[right_point].val); 
  dist0 = min(dist0, pot->obj_tab_[label].len - dist0); 
  *dist = dist0 = SY_ALV(dist0);

  if ((tail_ipo == o_point) || (head_ipo == o_point)){
    *tcorr = *hcorr = 0;
    return;
  }
  
  if (head_ipo != 0){
    hidist = SY_ALV(pot->elem_[head_ipo].val - pot->elem_[o_point].val); 
    hidist = min(hidist, pot->obj_tab_[label].len - hidist); 
    hidist = SY_ALV(hidist); 
  } else hidist = MAXINT;

  if (tail_ipo != 0){
    lodist = SY_ALV(pot->elem_[tail_ipo].val - pot->elem_[o_point].val); 
    lodist = min(lodist, pot->obj_tab_[label].len - lodist); 
    lodist = SY_ALV(lodist); 
  } else lodist = MAXINT;
  
  if (lodist < hidist){
    lovec.dirx =  vsk->point[tail_ipo].x - vsk->point[cur_ipo].x;
    lovec.diry =  vsk->point[tail_ipo].y - vsk->point[cur_ipo].y;
    hivec.dirx =  vsk->point[head_ipo].x - vsk->point[cur_ipo].x;
    hivec.diry =  vsk->point[head_ipo].y - vsk->point[cur_ipo].y;
   }
  else{
    lovec.dirx =  vsk->point[head_ipo].x - vsk->point[cur_ipo].x;
    lovec.diry =  vsk->point[head_ipo].y - vsk->point[cur_ipo].y;
    hivec.dirx =  vsk->point[tail_ipo].x - vsk->point[cur_ipo].x;
    hivec.diry =  vsk->point[tail_ipo].y - vsk->point[cur_ipo].y;
  }    

  lonorm = sqrt(SY_SQUARE(lovec.dirx)+SY_SQUARE(lovec.diry));
  hinorm = sqrt(SY_SQUARE(hivec.dirx)+SY_SQUARE(hivec.diry));

  /* test tail vertex */

  if ((evec.dirx * tvec0.dirx + evec.diry * tvec0.diry) < 0){
    flow.dirx = - evec.dirx; flow.diry = - evec.diry;
  } else flow = evec;
  
  normal.dirx = tvec0.diry; normal.diry = - tvec0.dirx;

  if ((flow.dirx * normal.dirx + flow.diry * normal.diry) < 0){
    normal.dirx = - normal.dirx; normal.diry = - normal.diry;
  }
  loproj = lovec.dirx * normal.dirx + lovec.diry * normal.diry;
  hiproj = hivec.dirx * normal.dirx + hivec.diry * normal.diry;

  if (SY_SGN(loproj) != SY_SGN(hiproj)){
    *tcorr = loproj < hiproj;
  }
  else{
    refproj = loproj;
    loproj  = (lovec.dirx * tvec0.dirx + lovec.diry * tvec0.diry)/lonorm;
    hiproj  = (hivec.dirx * tvec0.dirx + hivec.diry * tvec0.diry)/hinorm; 
    *tcorr  = (refproj > 0) ? loproj > hiproj : loproj < hiproj;
  }

  /* test head vertex */

  if ((evec.dirx * hvec0.dirx + evec.diry * hvec0.diry) < 0){
    flow.dirx = - evec.dirx; flow.diry = - evec.diry;

  } else flow = evec;

  normal.dirx = hvec0.diry; normal.diry = - hvec0.dirx;

  if ((flow.dirx * normal.dirx + flow.diry * normal.diry) < 0){
    normal.dirx = - normal.dirx; normal.diry = - normal.diry;
  }

  loproj = lovec.dirx * normal.dirx + lovec.diry * normal.diry;
  hiproj = hivec.dirx * normal.dirx + hivec.diry * normal.diry;

  if (SY_SGN(loproj) != SY_SGN(hiproj)){
    *hcorr = loproj < hiproj;
  }
  else{
    refproj = loproj;
    loproj  = (lovec.dirx * hvec0.dirx + lovec.diry * hvec0.diry)/lonorm;
    hiproj  = (hivec.dirx * hvec0.dirx + hivec.diry * hvec0.diry)/hinorm; 
    *hcorr  = (refproj > 0) ? loproj > hiproj : loproj < hiproj;
  }
}

void Sy_CorrResidRay (SySkelDAC *vsk, SyIpoDAC *ipd, SyRegpot *pot, 
		      SyVertex tvertex, SyVertex hvertex, SyVector tvec0, 
		      int left_point, int right_point, int ref_point, 
		      int *tcorr, float *dist, int *is_mindev)
{
  SyVector evec, tvec, flow, normal, hivec, lovec;
  int    cur_ipo, o_point, tail_ipo, head_ipo;
  float  dist0;
  int    hidist = 0, lodist = 0;
  int    tail_icn, head_icn;
  int    label;
  double hiproj, loproj, refproj;
  double hinorm, lonorm;

  *is_mindev = 0;

  evec.dirx = hvertex.xden/hvertex.num - tvertex.xden/tvertex.num;
  evec.diry = hvertex.yden/hvertex.num - tvertex.yden/tvertex.num;
  
  cur_ipo   = ref_point;
  if (cur_ipo == left_point) o_point = right_point; else o_point = left_point;

  head_icn = ipd->ipl[cur_ipo].head_icn;
  tail_icn = ipd->ipl[cur_ipo].tail_icn;
  label    = pot->elem_[cur_ipo].label;
  
  if (head_icn != 0){    
    head_ipo = (ipd->icn[head_icn].tail_ipo == cur_ipo) ? 
      ipd->icn[head_icn].head_ipo : ipd->icn[head_icn].tail_ipo;
  }
  else{
    head_ipo   = 0;
    *is_mindev = 1;
    *tcorr     = 0;
    return;
  }

  if (tail_icn != 0){    
    tail_ipo = (ipd->icn[tail_icn].tail_ipo == cur_ipo) ? 
      ipd->icn[tail_icn].head_ipo : ipd->icn[tail_icn].tail_ipo;
  }
  else{
    tail_ipo   = 0;
    *is_mindev = 1;
    *tcorr     = 0;
    return;
  }
  
  dist0 = SY_ALV(pot->elem_[left_point].val - pot->elem_[right_point].val); 
  dist0 = min(dist0, pot->obj_tab_[label].len - dist0); 
  *dist = dist0 = SY_ALV(dist0);

  if ((tail_ipo == o_point) || (head_ipo == o_point)){
    *tcorr = 0;
    return;
  }
  
  if (head_ipo != 0){
    hidist = SY_ALV(pot->elem_[head_ipo].val - pot->elem_[o_point].val); 
    hidist = min(hidist, pot->obj_tab_[label].len - hidist); 
    hidist = SY_ALV(hidist); 
  } else hidist = MAXINT;

  if (tail_ipo != 0){
    lodist = SY_ALV(pot->elem_[tail_ipo].val - pot->elem_[o_point].val); 
    lodist = min(lodist, pot->obj_tab_[label].len - lodist); 
    lodist = SY_ALV(lodist); 
  } else lodist = MAXINT;

  if (lodist < hidist){
    lovec.dirx =  vsk->point[tail_ipo].x - vsk->point[cur_ipo].x;
    lovec.diry =  vsk->point[tail_ipo].y - vsk->point[cur_ipo].y;
    hivec.dirx =  vsk->point[head_ipo].x - vsk->point[cur_ipo].x;
    hivec.diry =  vsk->point[head_ipo].y - vsk->point[cur_ipo].y;
   }
  else{
    lovec.dirx =  vsk->point[head_ipo].x - vsk->point[cur_ipo].x;
    lovec.diry =  vsk->point[head_ipo].y - vsk->point[cur_ipo].y;
    hivec.dirx =  vsk->point[tail_ipo].x - vsk->point[cur_ipo].x;
    hivec.diry =  vsk->point[tail_ipo].y - vsk->point[cur_ipo].y;
  }    

  lonorm = sqrt(SY_SQUARE(lovec.dirx)+SY_SQUARE(lovec.diry));
  hinorm = sqrt(SY_SQUARE(hivec.dirx)+SY_SQUARE(hivec.diry));

  /* test tail vertex */

  if ((evec.dirx * tvec0.dirx + evec.diry * tvec0.diry) < 0){
    flow.dirx = - evec.dirx; flow.diry = - evec.diry;
  } else flow = evec;
  
  normal.dirx = tvec0.diry; normal.diry = - tvec0.dirx;

  if ((flow.dirx * normal.dirx + flow.diry * normal.diry) < 0){
    normal.dirx = - normal.dirx; normal.diry = - normal.diry;
  }

  loproj = lovec.dirx * normal.dirx + lovec.diry * normal.diry;
  hiproj = hivec.dirx * normal.dirx + hivec.diry * normal.diry;

  if (SY_SGN(loproj) != SY_SGN(hiproj)){
    *tcorr = loproj < hiproj;
  }
  else{
    refproj = loproj;
    loproj  = (lovec.dirx * tvec0.dirx + lovec.diry * tvec0.diry)/lonorm;
    hiproj  = (hivec.dirx * tvec0.dirx + hivec.diry * tvec0.diry)/hinorm; 
    *tcorr  = (refproj > 0) ? loproj > hiproj : loproj < hiproj;
  }
}

void Sy_CirResidual (SySkelDAC *vsk, SyIpoDAC *ipd, SyRegpot *pot, 
		     int is_verbose)
{
  Sy_CirResidualBase (vsk, ipd, pot, is_verbose, SKE_CIR);
}

void Sy_BicirResidual (SySkelDAC *vsk, SyIpoDAC *ipd, SyRegpot *pot, 
		       int is_verbose)
{
  Sy_CirResidualBase (vsk, ipd, pot, is_verbose, SKE_BIC);
}


void Sy_CirResidualBase (SySkelDAC *vsk, SyIpoDAC *ipd, SyRegpot *pot, 
			 int is_verbose, enum SyPruneMethod flavor)
{ 
  int    i, corrcount = 0; 
  SyPoint  lpoint, rpoint; 
  SyVertex tvertex, hvertex; 
  float  dist, dev, tdev, hdev;
  double trad, hrad; 
  double tper, hper; 
  double norm, dweight;
  SyVector tvec[2], hvec[2]; 
  int    left_point, right_point; 
  int    tail_vertex, head_vertex; 
  int    left_label, right_label; 
  double targ, harg; 
  int    is_tcorr, is_hcorr, is_mindev; 

  Sy_Message (SY_FMTNL, (flavor == SKE_CIR)? 
	      "Circularity Residual": "Bi-Circularity Residual");

  vsk->prune_method = flavor;
  norm    = (flavor == SKE_CIR) ? 1.0 : SY_FRAC_2_PI;
  dweight = (flavor == SKE_CIR) ? 1.0 : SY_PI_2;

  vsk->point[0].x = vsk->point[0].y = 0;

  for (i = 1; i <= vsk->vdim; i++){
    vsk->regul[i].birth[0] = SY_NEG_INFTY;
    vsk->regul[i].birth[1] = SY_NEG_INFTY;
    vsk->regul[i].birth[2] = SY_NEG_INFTY;
  }

  for (i = 1; i <= vsk->edim; i++){

    switch(vsk->edge[i].rank){
    case 2:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      tail_vertex = vsk->edge[i].tail_vertex;
      head_vertex = vsk->edge[i].head_vertex;
      left_label  = pot->elem_[left_point].label;
      right_label = pot->elem_[right_point].label;

      lpoint  = vsk->point[left_point];
      rpoint  = vsk->point[right_point];
      tvertex = vsk->vertex[tail_vertex];
      hvertex = vsk->vertex[head_vertex];
      
      tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
      tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
      tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
      tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;
      hvec[0].dirx = lpoint.x - hvertex.xden/hvertex.num;
      hvec[0].diry = lpoint.y - hvertex.yden/hvertex.num;
      hvec[1].dirx = rpoint.x - hvertex.xden/hvertex.num;
      hvec[1].diry = rpoint.y - hvertex.yden/hvertex.num;

      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));
      hrad = sqrt(SY_SQUARE(hvec[0].dirx) + SY_SQUARE(hvec[0].diry));

      if (left_label != right_label){

 	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = SY_POS_INFTY;
	  vsk->eresid[i].head_residual = SY_POS_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = 0.0;
	  vsk->eresid[i].same_boundary = 0;
	}

	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	Sy_Insert(vsk, head_vertex, SY_POS_INFTY, hrad);
      }
      else{	
	
	Sy_CorrResidEdge (vsk, ipd, pot, tvertex, hvertex, 
			  tvec[0], hvec[0], 
			  left_point, right_point, left_point,
			  &is_tcorr, &is_hcorr, &dist, &is_mindev);

	if (is_mindev){
	  Sy_CorrResidEdge (vsk, ipd, pot, tvertex, hvertex, 
			    tvec[1], hvec[1], 
			    left_point, right_point, right_point,
			    &is_tcorr, &is_hcorr, &dist, &is_mindev);
	}

        if (is_tcorr) corrcount++;
	if (is_hcorr) corrcount++;

	targ = (tvec[0].dirx*tvec[1].dirx + tvec[0].diry*tvec[1].diry)/
	       (trad*trad);

	harg = (hvec[0].dirx*hvec[1].dirx + hvec[0].diry*hvec[1].diry)/
               (hrad*hrad);

        if (targ >  1.0) targ =  1.0;
        if (targ < -1.0) targ = -1.0;
        if (harg >  1.0) harg =  1.0;
        if (harg < -1.0) harg = -1.0;
  
	if (is_mindev){

	  tdev = norm * (dweight*dist - trad*acos(targ));
	  hdev = norm * (dweight*dist - hrad*acos(harg));

          dev  = min(tdev, hdev);
	  
          Sy_Insert (vsk, tail_vertex, tdev = dev, trad);
          Sy_Insert (vsk, head_vertex, hdev = dev, hrad);
        }
        else{
	  tper = trad * ((is_tcorr) ? (2*SY_PI - acos(targ)) : acos(targ));
	  hper = hrad * ((is_hcorr) ? (2*SY_PI - acos(harg)) : acos(harg));

	  Sy_Insert (vsk, tail_vertex, tdev = norm *(dweight*dist-tper), trad);
	  Sy_Insert (vsk, head_vertex, hdev = norm *(dweight*dist-hper), hrad);	  
	}

	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = tdev;
	  vsk->eresid[i].head_residual = hdev;
	  vsk->eresid[i].tail_corr     = is_tcorr;
	  vsk->eresid[i].head_corr     = is_hcorr;
	  vsk->eresid[i].dist          = dist;
	  vsk->eresid[i].same_boundary = 1;
	}
      }
      break;
      case 1:
        left_point  = vsk->edge[i].left_point;
        right_point = vsk->edge[i].right_point;
        head_vertex = vsk->edge[i].head_vertex;
	tail_vertex = vsk->edge[i].tail_vertex;
	left_label  = pot->elem_[left_point].label;
	right_label = pot->elem_[right_point].label;
	
	lpoint  = vsk->point[left_point];
	rpoint  = vsk->point[right_point];
	tvertex = vsk->vertex[tail_vertex];
        hvertex = vsk->vertex[head_vertex];
      
	tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
	tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
	tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
	tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;

	trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));

	if (left_label != right_label){

	  if (residualCompMode == RES_FULL){
	    vsk->eresid[i].tail_residual = SY_POS_INFTY;
	    vsk->eresid[i].head_residual = SY_POS_INFTY;
	    vsk->eresid[i].tail_corr     = 0;
	    vsk->eresid[i].head_corr     = 0;
	    vsk->eresid[i].dist          = 0.0;
	    vsk->eresid[i].same_boundary = 0;
	  }
	  
	  Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	}
	else{	
	
	  Sy_CorrResidRay (vsk, ipd, pot, tvertex, hvertex, tvec[0], 
			   left_point, right_point, left_point, 
			   &is_tcorr, &dist, &is_mindev);
	  

	  if (is_tcorr) corrcount++;

	  targ = (tvec[0].dirx*tvec[1].dirx + tvec[0].diry*tvec[1].diry)/
	         (trad*trad);

	  if (targ >  1.0) targ =  1.0;
	  if (targ < -1.0) targ = -1.0;

	  tper = trad * ((is_tcorr) ? (2*SY_PI - acos(targ)) : acos(targ));
	  dev  = norm * (dweight*dist-tper);
	  
	  if (residualCompMode == RES_FULL){
	    vsk->eresid[i].tail_residual = dev;
	    vsk->eresid[i].head_residual = SY_NEG_INFTY;
	    vsk->eresid[i].tail_corr     = is_tcorr;
	    vsk->eresid[i].head_corr     = 0;
	    vsk->eresid[i].dist          = dist;
	    vsk->eresid[i].same_boundary = 1;
	  }

	  Sy_Insert (vsk, tail_vertex, dev, trad);
	  break;
	}
    }
  }
  sprintf (syStrBuf, "%d corrections ", corrcount);
  Sy_Message (SY_FMTNL, syStrBuf);
}

void Sy_KleinResidual (SySkelDAC *vsk, SyIpoDAC *ipd, SyRegpot *pot, 
		       int is_verbose)
{ 
  int    i, corrcount = 0; 
  SyPoint  lpoint, rpoint; 
  SyVertex tvertex, hvertex; 
  double trad, hrad;
  float  dist, dev; 
  double tper, hper; 
  SyVector tvec[2], hvec[2]; 
  int    left_point, right_point; 
  int    tail_vertex, head_vertex; 
  int    left_label, right_label; 
  double targ, harg; 

  if (is_verbose) Sy_Message (SY_FMTNL, "Method: Klein's Residual ");

  vsk->prune_method = SKE_KLE;
  vsk->point[0].x = vsk->point[0].y = 0;

  for (i = 1; i <= vsk->vdim; i++){
    vsk->regul[i].birth[0] = SY_NEG_INFTY;
    vsk->regul[i].birth[1] = SY_NEG_INFTY;
    vsk->regul[i].birth[2] = SY_NEG_INFTY;
  }

  for (i = 1; i <= vsk->edim; i++){
    switch(vsk->edge[i].rank){
    case 2:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      tail_vertex = vsk->edge[i].tail_vertex;
      head_vertex = vsk->edge[i].head_vertex;
      left_label  = pot->elem_[left_point].label;
      right_label = pot->elem_[right_point].label;

      lpoint  = vsk->point[left_point];
      rpoint  = vsk->point[right_point];
      tvertex = vsk->vertex[tail_vertex];
      hvertex = vsk->vertex[head_vertex];
      
      tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
      tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
      tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
      tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;
      hvec[0].dirx = lpoint.x - hvertex.xden/hvertex.num;
      hvec[0].diry = lpoint.y - hvertex.yden/hvertex.num;
      hvec[1].dirx = rpoint.x - hvertex.xden/hvertex.num;
      hvec[1].diry = rpoint.y - hvertex.yden/hvertex.num;

      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));
      hrad = sqrt(SY_SQUARE(hvec[0].dirx) + SY_SQUARE(hvec[0].diry));

      if (left_label != right_label){

 	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = SY_POS_INFTY;
	  vsk->eresid[i].head_residual = SY_POS_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = 0.0;
	  vsk->eresid[i].same_boundary = 0;
	}

	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	Sy_Insert(vsk, head_vertex, SY_POS_INFTY, hrad);
      }
      else{	
	dist  = SY_ALV(pot->elem_[left_point].val - 
		    pot->elem_[right_point].val);
	dist  = min(dist, pot->obj_tab_[left_label].len - dist);
	dist  = SY_ALV(dist);
 
	targ = (tvec[0].dirx*tvec[1].dirx + tvec[0].diry*tvec[1].diry)/
	       (trad*trad);

	harg = (hvec[0].dirx*hvec[1].dirx + hvec[0].diry*hvec[1].diry)/
	        (hrad*hrad);

        if (targ >  1.0) targ =  1.0;
        if (targ < -1.0) targ = -1.0;
        if (harg >  1.0) harg =  1.0;
        if (harg < -1.0) harg = -1.0;

	tper = trad * acos(targ);
	hper = hrad * acos(harg);
	dev = min(dist-tper, dist-hper);

	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = dev;
	  vsk->eresid[i].head_residual = dev;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = dist;
	  vsk->eresid[i].same_boundary = 1;
	}

	Sy_Insert (vsk, tail_vertex, dev, trad);
	Sy_Insert (vsk, head_vertex, dev, hrad);
      }
      break;
      case 1:
        left_point  = vsk->edge[i].left_point;
        right_point = vsk->edge[i].right_point;
        head_vertex = vsk->edge[i].head_vertex;
	tail_vertex = vsk->edge[i].tail_vertex;
	left_label  = pot->elem_[left_point].label;
	right_label = pot->elem_[right_point].label;
	
	lpoint  = vsk->point[left_point];
	rpoint  = vsk->point[right_point];
	tvertex = vsk->vertex[tail_vertex];
        hvertex = vsk->vertex[head_vertex];
      
	tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
	tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
	tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
	tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;

	trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));

	if (left_label != right_label){

	  if (residualCompMode == RES_FULL){
	    vsk->eresid[i].tail_residual = SY_POS_INFTY;
	    vsk->eresid[i].head_residual = SY_POS_INFTY;
	    vsk->eresid[i].tail_corr     = 0;
	    vsk->eresid[i].head_corr     = 0;
	    vsk->eresid[i].dist          = 0.0;
	    vsk->eresid[i].same_boundary = 0;
	  }

	  Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	}
	else{	
	  dist  = SY_ALV(pot->elem_[left_point].val - 
		      pot->elem_[right_point].val);
	  dist  = min(dist, pot->obj_tab_[left_label].len - dist);
	  dist  = SY_ALV(dist);

	  targ = (tvec[0].dirx*tvec[1].dirx + tvec[0].diry*tvec[1].diry)/
	         (trad*trad);

	  if (targ >  1.0) targ =  1.0;
	  if (targ < -1.0) targ = -1.0;

	  tper = trad * acos(targ);
	  dev  = dist-tper;

	  if (residualCompMode == RES_FULL){
	    vsk->eresid[i].tail_residual = dev;
	    vsk->eresid[i].head_residual = SY_NEG_INFTY;
	    vsk->eresid[i].tail_corr     = 0;
	    vsk->eresid[i].head_corr     = 0;
	    vsk->eresid[i].dist          = dist;
	    vsk->eresid[i].same_boundary = 1;
	  }
	  Sy_Insert (vsk, tail_vertex, dev, trad);
	  break;
	}
    }
  }
}
 
void Sy_ChordResidual (SySkelDAC *vsk, SyIpoDAC *ipd, SyRegpot *pot, 
		       int is_verbose)
{
  int    i;
  SyPoint  lpoint, rpoint;
  SyVertex tvertex, hvertex;
  double trad, hrad; 
  float  dist, dev;
  double tsec, hsec;
  SyVector tvec[2], hvec[2];
  int    left_point, right_point;
  int    tail_vertex, head_vertex;
  int    left_label, right_label;

  if (is_verbose) Sy_Message (SY_FMTNL, "Chord Residual ");
  vsk->prune_method = SKE_CHO;

  for (i = 1; i <= vsk->vdim; i++){
    vsk->regul[i].birth[0] = SY_NEG_INFTY;
    vsk->regul[i].birth[1] = SY_NEG_INFTY;
    vsk->regul[i].birth[2] = SY_NEG_INFTY;
  }

  for (i = 1; i <= vsk->edim; i++){
    switch(vsk->edge[i].rank){
    case 2:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      tail_vertex = vsk->edge[i].tail_vertex;
      head_vertex = vsk->edge[i].head_vertex;
      left_label  = pot->elem_[left_point].label;
      right_label = pot->elem_[right_point].label;

      lpoint  = vsk->point[left_point];
      rpoint  = vsk->point[right_point];
      tvertex = vsk->vertex[tail_vertex];
      hvertex = vsk->vertex[head_vertex];
      
      tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
      tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
      tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
      tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;
      hvec[0].dirx = lpoint.x - hvertex.xden/hvertex.num;
      hvec[0].diry = lpoint.y - hvertex.yden/hvertex.num;
      hvec[1].dirx = rpoint.x - hvertex.xden/hvertex.num;
      hvec[1].diry = rpoint.y - hvertex.yden/hvertex.num;

      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));
      hrad = sqrt(SY_SQUARE(hvec[0].dirx) + SY_SQUARE(hvec[0].diry));

      if (left_label != right_label){

 	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = SY_POS_INFTY;
	  vsk->eresid[i].head_residual = SY_POS_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = 0.0;
	  vsk->eresid[i].same_boundary = 0;
	}

	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	Sy_Insert(vsk, head_vertex, SY_POS_INFTY, hrad);
      }
      else{	
	dist  = SY_ALV(pot->elem_[left_point].val - 
		   pot->elem_[right_point].val);
	dist  = min(dist, pot->obj_tab_[left_label].len - dist);
	dist  = SY_ALV(dist);

	tsec = sqrt(SY_SQUARE(tvec[0].dirx-tvec[1].dirx) + 
                    SY_SQUARE(tvec[0].diry-tvec[1].diry));
	hsec = sqrt(SY_SQUARE(hvec[0].dirx-hvec[1].dirx) + 
                    SY_SQUARE(hvec[0].diry-hvec[1].diry));

	dev  = min(dist-tsec, dist-hsec);

	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = dist-tsec;
	  vsk->eresid[i].head_residual = dist-hsec;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = dist;
	  vsk->eresid[i].same_boundary = 1;
	}

	Sy_Insert (vsk, tail_vertex, dev, trad);
	Sy_Insert (vsk, head_vertex, dev, hrad);
      }
      break;
      case 1:
	left_point  = vsk->edge[i].left_point;
	right_point = vsk->edge[i].right_point;
	tail_vertex = vsk->edge[i].tail_vertex;
	left_label  = pot->elem_[left_point].label;
	right_label = pot->elem_[right_point].label;
	
	lpoint  = vsk->point[left_point];
	rpoint  = vsk->point[right_point];
	tvertex = vsk->vertex[tail_vertex];
      
	tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
	tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
	tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
	tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;

	trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));

	if (left_label != right_label){

	  if (residualCompMode == RES_FULL){
	    vsk->eresid[i].tail_residual = SY_POS_INFTY;
	    vsk->eresid[i].head_residual = SY_POS_INFTY;
	    vsk->eresid[i].tail_corr     = 0;
	    vsk->eresid[i].head_corr     = 0;
	    vsk->eresid[i].dist          = 0.0;
	    vsk->eresid[i].same_boundary = 0;
	  }

	  Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	}
	else{	
	  dist  = SY_ALV(pot->elem_[left_point].val - 
		      pot->elem_[right_point].val);
	  dist  = min(dist, pot->obj_tab_[left_label].len - dist);
	  dist  = SY_ALV(dist);

	  tsec = sqrt(SY_SQUARE(tvec[0].dirx-tvec[1].dirx) + 
                    SY_SQUARE(tvec[0].diry-tvec[1].diry));

	  dev  = dist-tsec;

	  if (residualCompMode == RES_FULL){
	    vsk->eresid[i].tail_residual = dev;
	    vsk->eresid[i].head_residual = SY_NEG_INFTY;
	    vsk->eresid[i].tail_corr     = 0;
	    vsk->eresid[i].head_corr     = 0;
	    vsk->eresid[i].dist          = dist;
	    vsk->eresid[i].same_boundary = 1;
	  }
	  Sy_Insert (vsk, tail_vertex, dev, trad);
	  break;
	}
    }
  }
}

void Sy_PotResidual (SySkelDAC *vsk, SyIpoDAC *ipd, SyRegpot *pot, 
		     int is_verbose)
{
  int    i;
  SyPoint  lpoint;
  SyVertex tvertex, hvertex;
  double trad, hrad;
  float  dist;
  SyVector tvec[2], hvec[2];
  int    left_point, right_point;
  int    tail_vertex, head_vertex;
  int    left_label, right_label;

  if (is_verbose) Sy_Message (SY_FMTNL, "Potential Residual ");

  vsk->prune_method = SKE_POT;

  for (i = 1; i <= vsk->vdim; i++){
    vsk->regul[i].birth[0] = SY_NEG_INFTY;
    vsk->regul[i].birth[1] = SY_NEG_INFTY;
    vsk->regul[i].birth[2] = SY_NEG_INFTY;
  }

  for (i = 1; i <= vsk->edim; i++){
    switch(vsk->edge[i].rank){
    case 2:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      tail_vertex = vsk->edge[i].tail_vertex;
      head_vertex = vsk->edge[i].head_vertex;
      left_label  = pot->elem_[left_point].label;
      right_label = pot->elem_[right_point].label;

      lpoint  = vsk->point[left_point];
      tvertex = vsk->vertex[tail_vertex];
      hvertex = vsk->vertex[head_vertex];
      
      tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
      tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
      hvec[0].dirx = lpoint.x - hvertex.xden/hvertex.num;
      hvec[0].diry = lpoint.y - hvertex.yden/hvertex.num;

      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));
      hrad = sqrt(SY_SQUARE(hvec[0].dirx) + SY_SQUARE(hvec[0].diry));

      if (left_label != right_label){

 	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = SY_POS_INFTY;
	  vsk->eresid[i].head_residual = SY_POS_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = 0.0;
	  vsk->eresid[i].same_boundary = 0;
	}

	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	Sy_Insert(vsk, head_vertex, SY_POS_INFTY, hrad);
      }
      else{	
	dist  = SY_ALV(pot->elem_[left_point].val - 
		   pot->elem_[right_point].val);
	dist  = min(dist, pot->obj_tab_[left_label].len - dist);
	dist  = SY_ALV(dist);

	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = dist;
	  vsk->eresid[i].head_residual = dist;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = dist;
	  vsk->eresid[i].same_boundary = 1;
	}

	Sy_Insert (vsk, tail_vertex, dist, trad);
	Sy_Insert (vsk, head_vertex, dist, hrad);
      }
      break;
      case 1:
	left_point  = vsk->edge[i].left_point;
	right_point = vsk->edge[i].right_point;
	tail_vertex = vsk->edge[i].tail_vertex;
	left_label  = pot->elem_[left_point].label;
	right_label = pot->elem_[right_point].label;
	
	lpoint  = vsk->point[left_point];
	tvertex = vsk->vertex[tail_vertex];
      
	tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
	tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;

	trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));

	if (left_label != right_label){

	  if (residualCompMode == RES_FULL){
	    vsk->eresid[i].tail_residual = SY_POS_INFTY;
	    vsk->eresid[i].head_residual = SY_POS_INFTY;
	    vsk->eresid[i].tail_corr     = 0;
	    vsk->eresid[i].head_corr     = 0;
	    vsk->eresid[i].dist          = 0.0;
	    vsk->eresid[i].same_boundary = 0;
	  }

	  Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	}
	else{	
	  dist  = SY_ALV(pot->elem_[left_point].val - 
		      pot->elem_[right_point].val);
	  dist  = min(dist, pot->obj_tab_[left_label].len - dist);
	  dist  = SY_ALV(dist);

	  if (residualCompMode == RES_FULL){
	    vsk->eresid[i].tail_residual = dist;
	    vsk->eresid[i].head_residual = SY_NEG_INFTY;
	    vsk->eresid[i].tail_corr     = 0;
	    vsk->eresid[i].head_corr     = 0;
	    vsk->eresid[i].dist          = dist;
	    vsk->eresid[i].same_boundary = 1;
	  }
	  Sy_Insert (vsk, tail_vertex, dist, trad);
	  break;
	}
    }
  }
}


static void Sy_Insert (SySkelDAC *vsk, int vindex, float dev, float rad)
{
  float*  birth;

  vsk->regul[vindex].radius = rad;
  birth = vsk->regul[vindex].birth;

  if (birth[2] < dev){
    birth[0] = birth[1];
    birth[1] = birth[2];
    birth[2] = dev;
    return;
  }
  if (birth[1] < dev){
    birth[0] = birth[1];
    birth[1] = dev;
    return;
  }
  if (birth[0] < dev){
    birth[0] = dev;
  }
}

void Sy_SetResidualCompMode (enum SyResidualCompMode mode)
{
  residualCompMode = mode;
}

void Sy_SetResidualBorderMode (enum SyResidualBorderMode mode)
{
  residualBorderMode = mode;
}

static float ske_res_eps = SY_RES_EPS;
static float ske_loc_eps = SY_LOC_EPS;

void Sy_SetResEpsilon (float eps)
{
  ske_res_eps = eps;
}

void Sy_SetLocEpsilon (float eps)
{
  ske_loc_eps = eps;
}

int Sy_SimpleResidualCut (SySkelDAC *vsk, int edge_ref, float threshold, 
			  SyVertex *tvertex, SyVertex *hvertex)
{
  int   retval;
  float head_radius, head_residual,
        tail_radius, tail_residual;

  retval = Sy_ResidualCut (vsk, edge_ref, threshold, tvertex, hvertex,
			     &tail_radius, &tail_residual, 
			     &head_radius, &head_residual);

  return(retval);
}
			 

int Sy_ResidualCut (SySkelDAC *vsk, int edge_ref, float threshold, 
		    SyVertex *tvertex, SyVertex *hvertex, 
		    float *ret_tail_radius, float *ret_tail_residual, 
		    float *ret_head_radius, float *ret_head_residual)
{
  SyVector   grad, eq, etail, tvec, hvec, ehead, qvec[2];
  SyDblPoint tv, hv, lov, hiv;
  SyPoint    lpoint, rpoint;
  SyDblPoint q;
  int      tail_vertex, head_vertex;
  double   tail_residual, head_residual, 
           lo_residual, hi_residual;
  double   k_lo, k_hi, k;
  double   qresidual, qrad, qarg, qperim,
           gradnorm, norm, dweight;
  int      tail_corr, head_corr;
  double   tmp;
  int      tmpint;
  int      iter = 0;
  
  if (vsk->erdim == 0) return(2);

  norm    = (vsk->prune_method == SKE_CIR) ? 1.0 : SY_FRAC_2_PI;
  dweight = (vsk->prune_method == SKE_CIR) ? 1.0 : SY_PI_2;
 
  tail_vertex   = vsk->edge[edge_ref].tail_vertex;
  head_vertex   = vsk->edge[edge_ref].head_vertex;
  tail_residual = vsk->eresid[edge_ref].tail_residual;
  head_residual = vsk->eresid[edge_ref].head_residual;

  if ((tail_residual >= threshold) && (head_residual >= threshold)){
    *tvertex           = vsk->vertex[tail_vertex];
    *hvertex           = vsk->vertex[head_vertex];
    *ret_tail_residual = tail_residual;
    *ret_tail_radius   = vsk->regul[tail_vertex].radius;
    *ret_head_residual = head_residual;
    *ret_head_radius   = vsk->regul[head_vertex].radius;
    return(0);
  }
  
  if ((tail_residual < threshold) && (head_residual < threshold)){
    return(2);
  }

  if (vsk->edge[edge_ref].rank < 2){
    *tvertex           = vsk->vertex[tail_vertex];
    *hvertex           = vsk->vertex[head_vertex];
    *ret_tail_residual = tail_residual;
    *ret_tail_radius   = vsk->regul[tail_vertex].radius;
    *ret_head_residual = head_residual;
    *ret_head_radius   = vsk->regul[head_vertex].radius;
    return(0);
  }
  
  if (tail_residual < threshold){
    tmp    = tail_residual; tail_residual = head_residual; head_residual = tmp;
    tmpint = tail_vertex; tail_vertex = head_vertex; head_vertex = tmpint;
    tail_corr = vsk->eresid[edge_ref].head_corr;
    head_corr = vsk->eresid[edge_ref].tail_corr;
    lpoint = vsk->point[vsk->edge[edge_ref].right_point];
    rpoint = vsk->point[vsk->edge[edge_ref].left_point];
  }
  else{
    tail_corr = vsk->eresid[edge_ref].tail_corr;
    head_corr = vsk->eresid[edge_ref].head_corr;
    lpoint = vsk->point[vsk->edge[edge_ref].left_point];
    rpoint = vsk->point[vsk->edge[edge_ref].right_point];
  }
    
  tv.x   = vsk->vertex[tail_vertex].xden/vsk->vertex[tail_vertex].num;
  tv.y   = vsk->vertex[tail_vertex].yden/vsk->vertex[tail_vertex].num;
  hv.x   = vsk->vertex[head_vertex].xden/vsk->vertex[head_vertex].num;
  hv.y   = vsk->vertex[head_vertex].yden/vsk->vertex[head_vertex].num;

  lov    = hv;
  hiv    = tv;
  hi_residual = tail_residual;
  lo_residual = head_residual;

  grad.dirx = tv.x - hv.x;
  grad.diry = tv.y - hv.y;
  gradnorm  = sqrt(grad.dirx*grad.dirx + grad.diry*grad.diry);

  tvec.dirx = lpoint.x - tv.x;
  tvec.diry = lpoint.y - tv.y;
  hvec.dirx = lpoint.x - hv.x;
  hvec.diry = lpoint.y - hv.y;
  
  if ((grad.dirx * tvec.dirx + grad.diry * tvec.diry) < 0){
    etail.dirx = - grad.dirx; etail.diry = - grad.diry;
  } else etail = grad;

  if ((grad.dirx * hvec.dirx + grad.diry * hvec.diry) < 0){
    ehead.dirx = - grad.dirx; ehead.diry = - grad.diry;
  } else ehead = grad;

  *tvertex  = vsk->vertex[tail_vertex];
  qresidual = head_residual;
  k = (threshold-head_residual)/(tail_residual-head_residual);

  while (((k*gradnorm)>ske_loc_eps) && (((1.0-k)*gradnorm) > ske_loc_eps)
	 &&(SY_ALV(threshold-qresidual)>ske_res_eps)){

    q.x = k*grad.dirx + lov.x;
    q.y = k*grad.diry + lov.y;

    qvec[0].dirx = lpoint.x - q.x;
    qvec[0].diry = lpoint.y - q.y;
    qvec[1].dirx = rpoint.x - q.x;
    qvec[1].diry = rpoint.y - q.y;
   
    qrad = sqrt(qvec[0].dirx*qvec[0].dirx + qvec[0].diry*qvec[0].diry);

    if ((qvec[0].dirx*grad.dirx + qvec[0].diry*grad.diry) < 0){
      eq.dirx = -grad.dirx;
      eq.diry = -grad.diry;
    } else eq = grad;
  
    qarg = (qvec[0].dirx*qvec[1].dirx + qvec[0].diry*qvec[1].diry)/
	         (qrad*qrad);

    if (qarg >  1.0) qarg =  1.0;
    if (qarg < -1.0) qarg = -1.0;

    if ((eq.dirx*etail.dirx + eq.diry*etail.diry) > 0){
      qperim = qrad * ((tail_corr != 0) ? (2*SY_PI-acos(qarg)): acos(qarg));
    }
    else{
      qperim = qrad * ((head_corr != 0) ? (2*SY_PI-acos(qarg)): acos(qarg));      
    }
    qresidual = norm * (dweight*vsk->eresid[edge_ref].dist - qperim);

    if (qresidual < threshold){
      lov.x = q.x;
      lov.y = q.y;
      lo_residual = qresidual;
    }
    else{
      hiv.x  = q.x;
      hiv.y  = q.y;
      hi_residual = qresidual;
    }   

    grad.dirx = hiv.x - lov.x;
    grad.diry = hiv.y - lov.y;
    gradnorm  = sqrt(grad.dirx*grad.dirx + grad.diry*grad.diry);
    k         = (threshold-lo_residual)/(hi_residual-lo_residual);
    iter++;
  }
  if (iter != 0){
    hvertex->xden      = k*grad.dirx + lov.x;
    hvertex->yden      = k*grad.diry + lov.y;
    hvertex->num       = 1.0;
    *ret_head_radius   = qrad;
    *ret_head_residual = qresidual;
  }
  else{
    *hvertex           = vsk->vertex[head_vertex];
    *ret_head_radius   = vsk->regul[head_vertex].radius;
    *ret_head_residual = lo_residual;
  }

  *ret_tail_radius   = vsk->regul[tail_vertex].radius;
  *ret_tail_residual = tail_residual;
  return(0);
}
    
void Sy_SetInfix (Boolean bool)
{
  is_residual_fix = bool;
}

void Sy_SetMaxfix (Boolean bool)
{
  is_maxfix = bool;
}

void Sy_CleanupIpoSelect (SyIpoDAC *ipd, SyRegpot *pot) 
{ 
  int i;

  if (syIsVerbose) sprintf (syStrBuf, "Cleanup %s ", sy_RegularId);
  Sy_Message (SY_FMTNL, syStrBuf);

  Sy_AllocRegpot (pot, ipd->ipodim);
  Sy_ComPot (ipd, pot, syIsVerbose, Sy_IpoTrace);
}
