/*
 *  $Id: Sy_ShmAlloc.c,v 1.1 1993/12/09 12:46:58 rlo Exp rlo $
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#if !defined(convex) || defined(NO_SHARED_MEM)
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <assert.h>

#include "Sy_ShmAlloc.h"

#define SY_USER_RW  00600
#define SY_SHM_INIT_KEY 3

typedef struct _SyMemDescr {
  int   type;
  int   alloc_bytes;
  int   shmid;
  key_t key;
  char* addr;
  struct _SyMemDescr *next;
} SyMemDescr;

typedef struct _SyShmInfo {
  int        dim;
  SyMemDescr *mem;
} SyShmInfo;

static int default_type = SY_ALLOC_SHM;
static SyShmInfo shminfo = {0, NULL};

static SyMemDescr* Sy_InsertKey (key_t key)
{
  SyMemDescr *memptr, *lastptr, *tmpptr;

  if (shminfo.mem == (SyMemDescr*) NULL){
    shminfo.mem = (SyMemDescr*) ck_malloc (sizeof(SyMemDescr));
    shminfo.mem->key  = key;
    shminfo.mem->alloc_bytes = 0;
    shminfo.mem->addr = (char*) NULL;
    shminfo.mem->next = (SyMemDescr*) NULL;
    shminfo.dim       = 1;
    return (shminfo.mem);
  }
  memptr = shminfo.mem; lastptr = (SyMemDescr*) NULL;
  while (memptr && (memptr->key < key)){
    lastptr = memptr;
    memptr = memptr->next;
  }

  if (memptr->key == key){
    perror("key collision"); return ( (SyMemDescr*) NULL);
  }

  if (lastptr == (SyMemDescr*) NULL){
    shminfo.mem       = (SyMemDescr*) ck_malloc (sizeof(SyMemDescr));
    shminfo.mem->key  = key; shminfo.mem->addr = (char*) NULL;
    shminfo.mem->alloc_bytes = 0;
    shminfo.mem->next = memptr;
    shminfo.dim++;
    return(shminfo.mem);
  }
  else{
    tmpptr = lastptr->next = (SyMemDescr*) ck_malloc (sizeof(SyMemDescr));
    tmpptr->key = key; tmpptr->alloc_bytes = 0; tmpptr->addr = (char*) NULL;
    tmpptr->next = (memptr == (SyMemDescr*) NULL) ? (SyMemDescr*) NULL :
      memptr;
    shminfo.dim++;
    return(tmpptr);
  }
}

static int Sy_DeleteKey (key_t key)
{  
  SyMemDescr *memptr, *lastptr, *tmpptr;
  memptr = shminfo.mem; lastptr = (SyMemDescr*) NULL;
  while (memptr && (memptr->key != key)){
    lastptr = memptr;
    memptr = memptr->next;
  }
  if (memptr != (SyMemDescr*) NULL){
    if (lastptr == (SyMemDescr*) NULL) shminfo.mem = memptr->next;
    else                               lastptr->next = memptr->next;
      
    free(memptr);
    return(0);
  }
  else return(-1);
}

static SyMemDescr* Sy_LookupKey (key_t key)
{
  SyMemDescr *memptr, *lastptr, *tmpptr;
  memptr = shminfo.mem; lastptr = (SyMemDescr*) NULL;
  while (memptr && (memptr->key != key)){
    lastptr = memptr;
    memptr = memptr->next;
  }
  return(memptr);
}

void Sy_SetAllocType (int type)
{
  switch(type){
  case SY_ALLOC_STD:
  case SY_ALLOC_SHM:     
    default_type = type;
    break;
  default:
    perror ("Wrong allocation type");
  }
}


char* Sy_ShmAlloc (int nbytes, key_t key)
{
  SyMemDescr *ctrl;
  char   *tmp_ptr;
  struct shmid_ds buf;

  if (key < SY_SHM_INIT_KEY){
    perror("key should be larger than 2");
    return ( (char*)NULL);
  }

  switch (default_type){
  case SY_ALLOC_STD:
    ctrl = Sy_InsertKey(key);
    assert(ctrl != (SyMemDescr*)NULL);
    ctrl->type = default_type;
    ctrl->alloc_bytes = nbytes;
    ctrl->shmid = -1;
    ctrl->addr = (char*) ck_malloc (ctrl->alloc_bytes);
    if (ctrl->addr == (char*) NULL){
      perror ("Sy_ShmAlloc failed"); ctrl->alloc_bytes = 0;
      return(NULL);
    }
    else return ctrl->addr;
    break;
  
  case SY_ALLOC_SHM:
    ctrl = Sy_InsertKey(key);
    assert(ctrl != (SyMemDescr*)NULL);
    ctrl->alloc_bytes = nbytes;
    ctrl->type = default_type;
    ctrl->shmid = shmget (ctrl->key, ctrl->alloc_bytes, IPC_CREAT);
    if (ctrl->shmid < 0 /* failure */){
      perror ("Sy_ShmAlloc failed"); ctrl->alloc_bytes = 0;
      return(NULL);
    }
    shmctl (ctrl->shmid, IPC_STAT, &buf);
    buf.shm_perm.uid  = getuid();
    buf.shm_perm.gid  = getgid();
    buf.shm_perm.mode = SY_USER_RW;
    shmctl (ctrl->shmid, IPC_SET, &buf);
    ctrl->addr = (char*) shmat(ctrl->shmid, (char*) NULL, (int) NULL); 
    if (ctrl->addr == (char*)-1){
      perror ("Sy_ShmAlloc failed"); ctrl->alloc_bytes = 0;
      return(NULL);
    }
    else return ctrl->addr;    
  }    
}

int Sy_ShmUnmap (key_t key)
{
  SyMemDescr *ctrl;
  ctrl = Sy_LookupKey(key);
  assert(ctrl != (SyMemDescr*)NULL);
  switch(ctrl->type){
  case SY_ALLOC_STD: return (0); break;
  case SY_ALLOC_SHM: 
    if (shmdt(ctrl->addr) < 0){
      perror ("Sy_ShmUnmap failed"); return (-1);
    }
    else return 0; break;
  }
}

int Sy_ShmUnmapAll (void)
{
  int retval = 0;
  SyMemDescr* memptr;

  memptr = shminfo.mem; 
  while (memptr){
    if (memptr->addr == (char*) NULL) continue;
    switch(memptr->type){
    case SY_ALLOC_STD: return (0); break;
    case SY_ALLOC_SHM: 
      if (shmdt(memptr->addr) < 0){
	perror ("Sy_ShmUnmap failed on entry"); retval = -1;
      }
      break;
    } /* switch */
    memptr = memptr->next;
  } /* while */
 

  return (retval);
}

char* Sy_ShmAddr (key_t key)
{
  SyMemDescr *ctrl;
  ctrl = Sy_LookupKey(key);
  assert(ctrl != (SyMemDescr*)NULL);
  switch(ctrl->type){
  case SY_ALLOC_STD: return (ctrl->addr); break;
  case SY_ALLOC_SHM: 
    ctrl->addr = (char*)shmat (ctrl->shmid, (char*) NULL, (int) NULL); 
    if (ctrl->addr == (char*)-1){
      perror ("Sy_ShmAddr failed"); return(NULL);
    }
    else return(ctrl->addr);    
    break;
  }
}

int Sy_ShmFree (key_t key)
{
  int retval = 0;
  SyMemDescr *ctrl;
  ctrl = Sy_LookupKey(key);
  assert(ctrl != (SyMemDescr*)NULL);
  switch(ctrl->type){
  case SY_ALLOC_STD: 
    free((char*)ctrl->addr); 
    break;
  case SY_ALLOC_SHM: 
    if (shmctl (ctrl->shmid, IPC_RMID, (struct shmid_ds*)NULL) < 0){
      perror ("Sy_Memfree failed"); retval = -1;
    }
  }
  if (retval == 0) assert(Sy_DeleteKey(key) == 0);
  shminfo.dim--;
  return(0);
}    

int Sy_ShmFreeAll (void)
{
  int retval = 0, tmpflag = 0;
  SyMemDescr* memptr;

  memptr = shminfo.mem; 
  while (memptr){
    tmpflag = 0;
    if (memptr->addr == (char*) NULL) continue;
    switch(memptr->type){
    case SY_ALLOC_STD: free((char*)memptr->addr); break;
    case SY_ALLOC_SHM: 
      if (shmctl (memptr->shmid, IPC_RMID, (struct shmid_ds*)NULL) < 0){
	perror ("Sy_MemfreeAll failed"); tmpflag = -1; retval = -1;
      }
      break;
    } /* switch */
    memptr = memptr->next; free(memptr);
  } /* while */
  shminfo.dim = 0;
  return (retval);
}

static char* Sy_ShmRestore (SyMemDescr *ctrl);

int Sy_ReadShmInfo (FILE *inp)
{
  SyMemDescr *memptr, *lastptr;
  int i;
  fscanf (inp, "%d\n", &shminfo.dim);
  lastptr = (SyMemDescr*) NULL;
  for (i = 0; i < shminfo.dim; i++){
    memptr  = (SyMemDescr*) ck_malloc (sizeof(SyMemDescr));
    if (shminfo.mem == (SyMemDescr*) NULL) shminfo.mem = memptr;
    if (lastptr != (SyMemDescr*) NULL) lastptr->next = memptr;
    if (fscanf (inp, "%d %d %d\n", &memptr->key, &memptr->alloc_bytes,
	    &memptr->type) != 3)
      panic("Could not read values from file");
    memptr->addr = (char*) NULL;
    if (memptr->alloc_bytes > 0){
      Sy_ShmRestore (memptr);
    }
    lastptr = memptr;
  }
  return (0);
}

int Sy_WriteShmInfo (FILE *outp)
{  
  SyMemDescr* memptr;
  ck_fprintf (outp, "%d \n\n", shminfo.dim);
  memptr = shminfo.mem; 
  while (memptr){
    ck_fprintf (outp, "%d %d %d\n", memptr->key, memptr->alloc_bytes,
	     memptr->type);
    memptr = memptr->next;
  }
  return (0);
}


static char* Sy_ShmRestore (SyMemDescr *ctrl)
{
  char   *tmp_ptr;
  struct shmid_ds buf;

  if (ctrl->alloc_bytes <= 0) return( (char*) NULL);

  switch (ctrl->type){
  case SY_ALLOC_STD:
    ctrl->shmid = -1;
    ctrl->addr = (char*) ck_malloc (ctrl->alloc_bytes);
    if (ctrl->addr == (char*) NULL){
      perror ("Sy_ShmMemRestore failed"); ctrl->alloc_bytes = 0;
      return((char*)NULL);
    }
    return (ctrl->addr);
    break;
  
  case SY_ALLOC_SHM:
    ctrl->shmid = shmget (ctrl->key, ctrl->alloc_bytes, (int) NULL);
    if (ctrl->shmid < 0 /* failure */){
      perror ("Sy_ShmMemRestore failed"); ctrl->alloc_bytes = 0;
      return(NULL);
    }
    ctrl->addr = (char*) shmat(ctrl->shmid, (char*) NULL, (int) NULL); 
    if (ctrl->addr == (char*)-1){
      perror ("Sy_ShmAlloc failed"); ctrl->alloc_bytes = 0;
      return(NULL);
    }
    else return ctrl->addr;    
  }    
}

#endif



