/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * 
 * Low Level IO for Voronoi Skeletons
 *
 */

static char sy_SkelIOId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include "Sy_Lang.h"
#include "Sy_Yacc.h"

static SyVSkeletonSet *cur_vss  = NULL;

static int rankptr = 1, elementptr = 1,
           nodeptr = 1, branchptr = 1, skelptr = 1;



/*********************************************************************
 * function int Sy_Skel_fprintf (fp, vsk)
 * 
 * purpose: sends Voronoi-skeleton data to a file referenced through fp
 *          (ASCII format)
 *
 * param:   [input] SySkelDAC* vsk  -  Voronoi-skeleton data
 *          [input] FILE*   fp     -  file pointer
 *
 * retval:  EOF (-1) indicates an error during write 
 *********************************************************************/

int Sy_Skel_fprintf (FILE *fp, char *indata, unsigned int iomode)
{
  SySkelDAC* vsk;
  int	    i, io_result, hierdim;

  vsk = (SySkelDAC*) indata;
  hierdim = (vsk->is_hierarchy) ? vsk->edim : 0;

  io_result = fprintf (fp, 
     "/* Voronoi skeleton info: */\npointdim = %d; edgedim = %d; \
vertexdim = %d;\nresidualdim = %d;\n", 
		       (int) vsk->pdim, (int) vsk->edim, (int) vsk->vdim,
		       (int) vsk->erdim);
  if (io_result == EOF) return (io_result); 

  io_result = fprintf (fp, 
		       "border = %s;\nhierarchy = %s;\nprunemethod =%d;\n", 
		       vsk->is_border ? "true":"false",
		       vsk->is_hierarchy ? "true":"false",
		       (int) vsk->prune_method);
  if (io_result == EOF) return (io_result); 

  Sy_Wrbegload (fp, "Sy_GenLoader", SY_FALSE, SY_TRUE);
  fprintf (fp, "\n");

  for (i = 1; i <= vsk->pdim; i++){

    io_result =
      fprintf (fp, "%d %d\n", 
	       (int) vsk->point[i].x, 
	       (int) vsk->point[i].y);
    
    if (io_result == EOF) return (io_result); 
  }
  
  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)vsk->point);
    vsk->point = (SyPoint*) NULL;
    vsk->pdim  = 0;
  }
  
  for (i = 1; i <= vsk->edim; i++){
    
    io_result =
      fprintf (fp, "%6d %6d %6d %6d %6d %6d %2d %6d %6d %6d %6d\n",
	       (int) vsk->edge[i].head_edge,   
	       (int) vsk->edge[i].tail_edge,
	       (int) vsk->edge[i].head_vertex, 
	       (int) vsk->edge[i].tail_vertex,
	       (int) vsk->edge[i].left_point,  
	       (int) vsk->edge[i].right_point,
	       (int) vsk->edge[i].rank,
	       (int) vsk->edge[i].dirx, 
	       (int) vsk->edge[i].diry,
	       (int) vsk->edge[i].xoff, 
	       (int) vsk->edge[i].yoff);
    
    if (io_result == EOF) return (io_result); 
  }

  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)vsk->edge);
    vsk->edge = (SyEdge*) NULL;
    vsk->edim = 0;
  }

  for (i = 1; i <= vsk->vdim; i++) {
    io_result =
      fprintf (fp, "%20.0lf %20.0lf %20.0lf\n", 
	       vsk->vertex[i].xden, 
	       vsk->vertex[i].yden,
	       vsk->vertex[i].num);
    
    if (io_result == EOF) return (io_result); 
  }

  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)vsk->vertex);
    vsk->vertex = (SyVertex*) NULL;
  }

  for (i = 1; i <= vsk->vdim; i++) {
    io_result =
      fprintf (fp, "%lf %lf %lf %lf\n", 
	       vsk->regul[i].birth[0], 
	       vsk->regul[i].birth[1],
	       vsk->regul[i].birth[2],
	       vsk->regul[i].radius);
    
    if (io_result == EOF) return (io_result); 
  }
  
  if (SY_IS_FREE_AFTER_WRITE(iomode)) {
    Sy_Free ((SY_VOID*)vsk->regul);
    vsk->regul = (SyRegul*) NULL;
    vsk->vdim   = 0;
  }

  if (vsk->erdim != 0){
    for (i = 1; i <= vsk->erdim; i++){
      io_result =
	fprintf (fp, "%f %f %f %d %d %d\n", 
		 vsk->eresid[i].head_residual,
		 vsk->eresid[i].tail_residual,
		 vsk->eresid[i].dist,
		 vsk->eresid[i].head_corr,
                 vsk->eresid[i].tail_corr,
		 vsk->eresid[i].same_boundary);
    
      if (io_result == EOF) return (io_result); 
    }

    if (SY_IS_FREE_AFTER_WRITE(iomode)) {
      Sy_Free ((SY_VOID*)vsk->eresid);
      vsk->eresid = (SyEresid*) NULL;
      vsk->erdim  = 0;
    }
  }
  
  if (vsk->is_hierarchy){
    for (i = 1; i <= hierdim; i++){
      io_result =
	fprintf (fp, "%d\n", vsk->hier[i]);
      if (io_result == EOF) return (io_result); 
    }
    if (SY_IS_FREE_AFTER_WRITE(iomode)) {
      Sy_Free ((SY_VOID*)vsk->hier);
      vsk->hier = NULL;
    }
  }

  return (1);
}


/**********************************************************************
 * function int Sy_Skel_fscanf (fp, vsk)
 * 
 * purpose: gets Voronoi-skeleton data from a file referenced through fp
 *          (ASCII format)
 *
 * param:   [inout] SySkelDAC* vsk   -  Voronoi-skeleton data
 *          [input] FILE*   fp   -  file pointer
 *
 * retval:  EOF (-1) indicates an error during read 
 *********************************************************************/

int Sy_Skel_fscanf (FILE *fp, SySkelDAC **ret_vsk, unsigned int iomode)
{
  SySkelDAC *vsk;
  int      i, io_result;
  int      buff [SY_NBUFF];
  double   fbuff[SY_NBUFF];
  int      bytes = 0;
  SyDataObject *dobj;

  vsk = (SySkelDAC*) Sy_Malloc (sizeof(SySkelDAC));
  Sy_InitSkeleton(vsk);

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "pointdim");
  if (dobj == NULL){
    Sy_Error ("Sy_SkelIO", "Sy_Skel_fscanf", "missing pointdim info...\n","");}
  vsk->pdim = dobj->t.i_val; 
  
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "edgedim");
  if (dobj == NULL){
    Sy_Error ("Sy_SkelIO", "Sy_Skel_fscanf", "missing edgedim info...\n","");}
  vsk->edim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "vertexdim");
  if (dobj == NULL){
    Sy_Error ("Sy_SkelIO","Sy_Skel_fscanf", "missing vertexdim info...\n","");}
  vsk->vdim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "pointdim");
  if (dobj == NULL){
    Sy_Error ("Sy_SkelIO", "Sy_Skel_fscanf", "missing pointdim info...\n","");}
  vsk->pdim = dobj->t.i_val; 
  
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "residualdim");
  if (dobj == NULL){
    Sy_Error ("Sy_SkelIO", "Sy_Skel_fscanf", 
	      "missing residualdim info...\n","");}
  vsk->erdim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "border");
  if (dobj == NULL) vsk->is_border = SY_FALSE;
  else vsk->is_border = strcmp(dobj->t.s_val, "true") ? SY_FALSE : SY_TRUE; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "hierarchy");
  if (dobj == NULL) vsk->is_hierarchy = SY_FALSE;
  else vsk->is_hierarchy = strcmp(dobj->t.s_val, "true") ? SY_FALSE : SY_TRUE; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "prunemethod");
  if (dobj == NULL){
    Sy_Error ("Sy_SkelIO", "Sy_Skel_fscanf", 
	      "missing prune method info...\n","");}
  vsk->prune_method = dobj->t.i_val; 
  
  vsk->point  = Sy_AllocPoint (vsk->pdim);
  vsk->edge   = Sy_AllocEdge  (vsk->edim);
  vsk->vertex = Sy_AllocVertex(vsk->vdim);
  vsk->regul  = Sy_AllocRegul (vsk->vdim);
  
  if (vsk->erdim != 0){
    vsk->eresid = Sy_AllocEresid (vsk->erdim);
  }

  for (i = 1; i <= vsk->pdim; i++){
    
    io_result = fscanf (fp, "%d %d", &buff[0], &buff[1]);  
    vsk->point[i].x = buff[0]; 
    vsk->point[i].y = buff[1];
    
    if (io_result != 2) return (EOF); 
  }

  for (i = 1; i <= vsk->edim; i++){

    io_result =
      fscanf (fp, "%d %d %d %d %d %d %d %d %d %d %d",
	      &buff[0], &buff[1], &buff[2], &buff[3],
	      &buff[4], &buff[5], &buff[6], &buff[7], 
	      &buff[8], &buff[9], &buff[10]);
    
    vsk->edge[i].head_edge    = buff[0];   
    vsk->edge[i].tail_edge    = buff[1];
    vsk->edge[i].head_vertex  = buff[2]; 
    vsk->edge[i].tail_vertex  = buff[3];
    vsk->edge[i].left_point   = buff[4];  
    vsk->edge[i].right_point  = buff[5];
    vsk->edge[i].rank         = buff[6];
    vsk->edge[i].dirx         = buff[7]; 
    vsk->edge[i].diry         = buff[8];
    vsk->edge[i].xoff         = buff[9]; 
    vsk->edge[i].yoff         = buff[10];
    
    if (io_result != 11) return (EOF); 
  }
  
  for (i = 1; i <= vsk->vdim; i++) {
    
    io_result =
      fscanf (fp, "%lf %lf %lf", &fbuff[0], &fbuff[1], &fbuff[2]); 
    vsk->vertex[i].xden = fbuff[0];
    vsk->vertex[i].yden = fbuff[1];
    vsk->vertex[i].num  = fbuff[2];
    
    if (io_result != 3) return (EOF); 
  }
  
  for (i = 1; i <= vsk->vdim; i++) {
    
    io_result =
      fscanf (fp, "%lf %lf %lf %lf", &fbuff[0], &fbuff[1], &fbuff[2], &fbuff[3]); 
    vsk->regul[i].birth[0] = fbuff[0];
    vsk->regul[i].birth[1] = fbuff[1];
    vsk->regul[i].birth[2] = fbuff[2];
    vsk->regul[i].radius   = fbuff[3];

    if (io_result != 4) return (EOF); 
  }
  
  if (vsk->erdim != 0){
    for (i = 1; i <= vsk->erdim; i++) {
      
      io_result =
	fscanf (fp, "%lf %lf %lf %d %d %d", 
		&fbuff[0], &fbuff[1], &fbuff[2], 
		&buff[0], &buff[1], &buff[2]); 
      vsk->eresid[i].head_residual = fbuff[0];
      vsk->eresid[i].tail_residual = fbuff[1];
      vsk->eresid[i].dist          = fbuff[2];
      vsk->eresid[i].head_corr     = buff[0];
      vsk->eresid[i].tail_corr     = buff[1];
      vsk->eresid[i].same_boundary = buff[2];
      if (io_result != 6) return (EOF); 
    }
    vsk->is_eresid = 1;
  }

  if (vsk->is_hierarchy){
    vsk->hier = (int*) Sy_Malloc(sizeof(int)*(vsk->edim+1));
    for (i = 1; i <= vsk->edim; i++){
     io_result = 
       fscanf (fp, "%d", &buff[0]);
     vsk->hier[i] = buff[0];
     if (io_result != 1) return (EOF);
   }
  }
  else vsk->hier = NULL;

  *ret_vsk = vsk; 
  return (bytes);
}


/*********************************************************************
 * function int Sy_Skel_fwrite (fp, vsk)
 * 
 * purpose: sends Voronoi-skeleton data to a file referenced through fp
 *          (BINARY format)
 *
 * param:   [input] SySkelDAC* vsk   -  Voronoi-skeleton data
 *          [input] FILE*   fp   -  file pointer
 *
 * retval:  EOF (-1) indicates an error during write 
 *********************************************************************/

int Sy_Skel_fwrite (FILE *fp, char *indata, unsigned int iomode)
{
  SySkelDAC* vsk;
  int	    i, io_result, hierdim;

  vsk = (SySkelDAC*) indata;
  hierdim = (vsk->is_hierarchy) ? vsk->edim : 0;

  io_result = fprintf (fp, 
     "/* Voronoi skeleton info: */\npointdim = %d; edgedim = %d; \
vertexdim = %d;\nresidualdim = %d;\n", 
		       (int) vsk->pdim, (int) vsk->edim, (int) vsk->vdim,
		       (int) vsk->erdim);
  if (io_result == EOF) return (io_result); 

  io_result = fprintf (fp, 
		       "border = %s;\nhierarchy = %s;\nprunemethod =%d;\n", 
		       vsk->is_border ? "true":"false",
		       vsk->is_hierarchy ? "true":"false",
		       (int) vsk->prune_method);
  if (io_result == EOF) return (io_result); 

  Sy_Wrbegload (fp, "Sy_GenLoader", SY_FALSE, SY_FALSE);

  io_result = fwrite ((SY_VOID*)vsk->point, sizeof(SyPoint), vsk->pdim+1, fp);
  if (io_result == 0) return (EOF); 
  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)vsk->point);
    vsk->point = (SyPoint*) NULL;
    vsk->pdim  = 0;
  }
  
  io_result = fwrite ((SY_VOID*)vsk->edge, sizeof(SyEdge), vsk->edim+1, fp);  
  if (io_result == 0) return (EOF); 
  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)vsk->edge);
    vsk->edge = (SyEdge*) NULL;
    vsk->edim = 0;
  }
  
  io_result = fwrite ((SY_VOID*)vsk->vertex, sizeof(SyVertex), vsk->vdim+1, fp);  
  if (io_result == 0) return (EOF); 
  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)vsk->vertex);
    vsk->vertex = (SyVertex*) NULL;
  }
  
  io_result = fwrite ((SY_VOID*)vsk->regul, sizeof(SyRegul), vsk->vdim+1, fp);
  if (io_result == 0) return (EOF);
    if (SY_IS_FREE_AFTER_WRITE(iomode)) {
    Sy_Free ((SY_VOID*)vsk->regul);
    vsk->regul = (SyRegul*) NULL;
    vsk->vdim  = 0;
  }

  if (vsk->erdim != 0){
    io_result =fwrite ((SY_VOID*)vsk->eresid, sizeof(SyEresid), vsk->erdim+1, fp);
    if (io_result == 0) return (EOF);
    if (SY_IS_FREE_AFTER_WRITE(iomode)) {
      Sy_Free ((SY_VOID*)vsk->eresid);
      vsk->eresid = (SyEresid*) NULL;
      vsk->erdim  = 0;
    }
  }  

  if (vsk->is_hierarchy){
      io_result = fwrite ((SY_VOID*)vsk->hier, sizeof(int), hierdim+1, fp);
      if (io_result == 0) return (EOF); 
      if (SY_IS_FREE_AFTER_WRITE(iomode)) {
	Sy_Free ((SY_VOID*)vsk->hier);
	vsk->hier = NULL;
      }
  }

  return (1);
}

/**********************************************************************
 * function int Sy_Skel_fread (fp, vsk)
 * 
 * purpose: gets Voronoi-skeleton data from a file referenced through fp
 *          (BINARY format)
 *
 * param:   [inout] SySkelDAC* vsk   -  Voronoi-skeleton data
 *          [input] FILE*    fp   -  file pointer
 *
 * retval:  EOF (-1) indicates an error during read 
 *********************************************************************/

int Sy_Skel_fread (FILE *fp, SySkelDAC **ret_vsk, unsigned int iomode)
{
  SySkelDAC *vsk;
  int      io_result;
  int      bytes = 0;
  SyDataObject *dobj;

  vsk = (SySkelDAC*) Sy_Malloc (sizeof(SySkelDAC));
  Sy_InitSkeleton(vsk);

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "pointdim");
  if (dobj == NULL){
    Sy_Error ("Sy_SkelIO", "Sy_Skel_fscanf", "missing pointdim info...\n","");}
  vsk->pdim = dobj->t.i_val; 
  
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "edgedim");
  if (dobj == NULL){
    Sy_Error ("Sy_SkelIO", "Sy_Skel_fscanf", "missing edgedim info...\n","");}
  vsk->edim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "vertexdim");
  if (dobj == NULL){
    Sy_Error ("Sy_SkelIO","Sy_Skel_fscanf", "missing vertexdim info...\n","");}
  vsk->vdim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "pointdim");
  if (dobj == NULL){
    Sy_Error ("Sy_SkelIO", "Sy_Skel_fscanf", "missing pointdim info...\n","");}
  vsk->pdim = dobj->t.i_val; 
  
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "residualdim");
  if (dobj == NULL){
    Sy_Error ("Sy_SkelIO", "Sy_Skel_fscanf", 
	      "missing residualdim info...\n","");}
  vsk->erdim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "border");
  if (dobj == NULL) vsk->is_border = SY_FALSE;
  else vsk->is_border = strcmp(dobj->t.s_val, "true") ? SY_FALSE : SY_TRUE; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "hierarchy");
  if (dobj == NULL) vsk->is_hierarchy = SY_FALSE;
  else vsk->is_hierarchy = strcmp(dobj->t.s_val, "true") ? SY_FALSE : SY_TRUE; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "prunemethod");
  if (dobj == NULL){
    Sy_Error ("Sy_SkelIO", "Sy_Skel_fscanf", 
	      "missing prune method info...\n","");}
  vsk->prune_method = dobj->t.i_val; 

  vsk->point  = Sy_AllocPoint (vsk->pdim);
  vsk->edge   = Sy_AllocEdge  (vsk->edim);
  vsk->vertex = Sy_AllocVertex(vsk->vdim);
  vsk->regul  = Sy_AllocRegul(vsk->vdim);

  if (vsk->erdim != 0){
    vsk->eresid = Sy_AllocEresid (vsk->erdim);
  }
  
  io_result = fread ((SY_VOID*)vsk->point, sizeof(SyPoint), vsk->pdim+1, fp);
  
  if (io_result == 0) return (EOF); 
  
  io_result = fread ((SY_VOID*)vsk->edge, sizeof(SyEdge), vsk->edim+1, fp);
  
  if (io_result == 0) return (EOF); 
  
  io_result = fread ((SY_VOID*)vsk->vertex, sizeof(SyVertex), vsk->vdim+1, fp);
  
  if (io_result == 0) return (EOF);
  
  io_result = fread ((SY_VOID*)vsk->regul, sizeof(SyRegul), vsk->vdim+1, fp);
  
  if (io_result == 0) return (EOF);

  if (vsk->erdim != 0){
    io_result = fread ((SY_VOID*)vsk->eresid, sizeof(SyEresid), vsk->erdim+1, fp);
    vsk->is_eresid = 1;
    if (io_result == 0) return (EOF);
  }
  
  if (vsk->is_hierarchy){
    vsk->hier = (int*) Sy_Malloc(sizeof(int)*(vsk->edim+1));
    io_result = fread((SY_VOID*)vsk->hier, sizeof(int), vsk->edim+1, fp);
    if (io_result == 0){
      vsk->is_hierarchy = SY_FALSE;
      Sy_Free ((SY_VOID*)vsk->hier);
      vsk->hier = NULL;
    }
  }
  else vsk->hier = NULL;

  *ret_vsk = vsk;
  return (bytes);
}

int Sy_SkeletonSet_fprintf (FILE *fp, char *indata, unsigned int iomode) 
{
  int i;
  SyVSkeletonSet *vss;
  vss = (SyVSkeletonSet*) indata;
  
  for (i = 1; i <= vss->dim; i++){
    if (vss->skel[i].branchdim == 0 && vss->skel[i].nodedim == 0)
      continue;
    Sy_Skeleton_fprintf (fp, &vss->skel[i], NULL);
  }
  return SY_TRUE;
}

static char skel_out_template[]= 
"\n\
object skeleton_%d {\n\
type          = skeleton;\n\
id            = %d; \n\
user_id       = %d; \n\
branchdim     = %d; \n\
nodedim       = %d; \n\
is_foreground = %s; \n\n";

static char branch_out_template[] = 
"\n\
object branch_%d { /* skeleton no %d */\n\
type            = skeletonbranch;\n\
binary_io       = %s;\n\
id              = %d;\n\
user_id         = %d;\n\
skel_id         = %d;\n\
hierarchy_level = %d;\n\
is_foreground   = %s;\n\
has_boundary    = %s;\n\
strength        = %f;\n\
length          = %f;\n\
max_strength    = %f; min_strength = %f;\n\
mean_strength   = %f; var_strength = %f;\n\
max_radius      = %f; min_radius   = %f;\n\
mean_radius     = %f; var_radius   = %f;\n\
tail_branch     = %d; head_branch  = %d;\n\
tail_node       = %d; head_node    = %d;\n\
elementdim      = %d;\n\
axistype        = %s;\n";

static char node_out_template[]=
"\n\
object node_%d { /* skeleton no %d */\n\
type            = skeletonnode;\n\
binary_io       = %s;\n\
id              = %d;\n\
user_id         = %d;\n\
skel_id         = %d;\n\
hierarchy_level = %d;\n\
is_foreground   = %s;\n\
nodetype        = %s;\n\
strength        = %f;\n\
radius          = %f;\n\
rank            = %d;\n\
xytype          = %s;\n\
/* position: */\n\
%s\n";

char* Sy_MapBranchElementType2String (SyBranchElementType etype)
{
  switch (etype){
  case SY_REAL:       return "real"; break;
  case SY_RATIONAL:   return "rational"; break;
  case SY_INTEGER:    return "integer"; break;
  case SY_NILBRANCH:  return "nil"; break;
  default:            return "nil";
  }
}


char* Sy_MapNodeLoc2String(SyBranchElementType etype, union SyGenericPoint *loc)
{
  static char buf[256];

  switch (etype){
  case SY_REAL:       sprintf (buf, "x = %f; y = %f;", loc->rp.x, loc->rp.y);
                      return buf; break;
  case SY_RATIONAL:   sprintf (buf, "x = %d; y = %d; xnum = %d; ynum = %d;", 
			       loc->qp.xden, loc->qp.yden, 
			       loc->qp.xnum, loc->qp.ynum);
		      return buf; break;
  case SY_INTEGER:    sprintf (buf, "x = %d; y = %d;", loc->np.x, loc->np.y);
                      return buf; break;
  default:            sprintf (buf, "x = %f; y = %f;", loc->rp.x, loc->rp.y);
                      return buf;
  }
}


int Sy_Skeleton_fprintf (FILE *fp, SyVSkeleton *vs, SyControl *ctrl)
{
  int skptr, i, k;
  SyVSkeletonNode *n;
  SyVSkeletonBranch *b;

  fprintf (fp, skel_out_template,
	   vs->id,
	   vs->id,
	   vs->user_id,
	   vs->branchdim,
	   vs->nodedim,
	   vs->is_foreground ? "true" : "false");
  
  for (i = 1; i <= vs->nodedim; i++){
    n = &vs->node[i];
    fprintf (fp, node_out_template, 
	     i,
	     vs->id,
	     syIsBinaryIO ? "true" : "false",
	     n->id,
	     n->user_id,
	     vs->id,
	     n->hierarchy_level,
	     n->is_foreground ? "true" : "false",
	     n->ntype == SY_INTERNAL ? "internal" : "endpoint",
	     n->strength,
	     n->radius,
	     n->rank,
	     Sy_MapBranchElementType2String (n->etype),
	     Sy_MapNodeLoc2String(n->etype, &n->loc));

    fprintf (fp, "/* list of incident branches: */\n");
    Sy_Wrbegload (fp, "Sy_ReadIncList", SY_FALSE, SY_TRUE); 
    /* list of incident branches */
    
    for (k = 1; k <= n->rank; k++){
      fprintf (fp, "%d\n", n->inbranch[k]);
    }
  
    Sy_Wrendload (fp, SY_FALSE);

    fprintf (fp, "\n} /* end node no %d */ \n", n->id);
  } /* for all nodes */
  
  for (i = 1; i <= vs->branchdim; i++){
    b = &vs->branch[i];
    fprintf (fp, branch_out_template, 
	     i,
	     vs->id,
	     syIsBinaryIO ? "true" : "false",
	     b->id,
	     b->user_id,
	     vs->id,
	     b->hierarchy_level,
	     b->is_foreground ? "true" : "false",
	     b->has_boundary ? "true" : "false",
	     b->strength,
	     b->length,
	     b->max_strength, b->min_strength, 
	     b->mean_strength, b->var_strength,
	     b->max_radius, b->min_radius,
	     b->mean_radius, b->var_radius,
	     b->tail_branch, b->head_branch,
	     b->tail_node, b->head_node,
	     b->elementdim,
	     Sy_MapBranchElementType2String (b->etype));

    fprintf (fp, "/* symmetry axis: */\n");
    Sy_Wrbegload (fp, "Sy_ReadSymmAxis", SY_FALSE, !syIsBinaryIO); 

    if (syIsBinaryIO){
      fwrite (b->axis, sizeof(union SyGenericPoint), b->elementdim+1, fp);
    }
    else { /* ascii */
      switch (b->etype){
      case SY_RATIONAL:   
	for (k = 1; k <= b->elementdim; k++){ 
	  fprintf (fp, "%d %d %d %d\n",b->axis[k].qp.xden, b->axis[k].qp.xnum,
		   b->axis[k].qp.yden, b->axis[k].qp.ynum); 
	}
	break;
      case SY_INTEGER:   
	for (k = 1; k <= b->elementdim; k++){ 
	  fprintf (fp, "%d %d\n", b->axis[k].np.x, b->axis[k].np.y);}
	break;
      case SY_REAL:     
      default:            
	for (k = 1; k <= b->elementdim; k++){ 
	  fprintf (fp, "%f %f\n", b->axis[k].rp.x, b->axis[k].rp.y);}
	break;
      }
    } /* else */
    
    Sy_Wrendload (fp, SY_FALSE);

    fprintf (fp, "/* radii: */\n");
    Sy_Wrbegload (fp, "Sy_ReadRadiusList", SY_FALSE, !syIsBinaryIO); 

    if (syIsBinaryIO){
      fwrite (b->radius, sizeof(float), b->elementdim+1, fp);
    }
    else {
      for (k = 1; k <= b->elementdim; k++){
	fprintf (fp, "%f\n", b->radius[k]);
      }
    } /* else */

    Sy_Wrendload (fp, SY_FALSE);

    fprintf (fp, "/* references to original boundary: */\n");
    Sy_Wrbegload (fp, "Sy_ReadReferences", SY_FALSE, !syIsBinaryIO); 

    if (syIsBinaryIO){
      fwrite (b->ref, sizeof(SyVSkeletonRef), b->elementdim+1, fp);
    }
    else {
      for (k = 1; k <= b->elementdim; k++){
	fprintf (fp, "%d %d %d %d\n", b->ref[k].left_contour, 
		                      b->ref[k].right_contour,
		                      b->ref[k].left_point,
	                              b->ref[k].right_point);
      }
    } /* else */

    Sy_Wrendload (fp, SY_FALSE);

    fprintf (fp, "/* boundary (offset vectors): */\n");
    Sy_Wrbegload (fp, "Sy_ReadBndryOffsetVectors", SY_FALSE, !syIsBinaryIO); 

    if (syIsBinaryIO){
      fwrite (b->left_bndry, sizeof(SyVSkeletonBndryOffset), 
	      b->elementdim, fp);
      fwrite (b->right_bndry, sizeof(SyVSkeletonBndryOffset), 
	      b->elementdim, fp);
    }
    else {
      for (k = 1; k <= b->elementdim-1; k++){
	fprintf (fp, "%f %f\n", b->left_bndry[k].xoff,
		 b->left_bndry[k].yoff);
      }
      for (k = 1; k <= b->elementdim-1; k++){
	fprintf (fp, "%f %f\n", b->right_bndry[k].xoff,
		 b->right_bndry[k].yoff);
      }
    } /* else */ 

    Sy_Wrendload (fp, SY_FALSE);

    fprintf (fp, "\n} /* end branch no %d */ \n", b->id);
  } /* for all branches */

  fprintf (fp, "\n} /* end skeleton no %d */ \n", vs->id);

  return SY_TRUE;
}

/* ***************** I/O of Skeleton Sets ***************** */

SyBranchElementType Sy_MapString2BranchElementType (char *string)
{
  if (strcmp(string,"real")==0) return SY_REAL;
  else
  if (strcmp(string,"rational")==0) return SY_RATIONAL;
  else
  if (strcmp(string,"integer")==0) return SY_INTEGER;
  else
    return SY_NILBRANCH;
}

SyBranchNodeType Sy_MapString2NodeElementType (char *string)
{
  if (strcmp(string,"internal")==0) return SY_INTERNAL;
  else
  if (strcmp(string,"endpoint")==0) return SY_END;
  else
    return SY_NILNODE;
}

SyBoolean Sy_MapString2Bool (char *string)
{
  if (strcmp(string,"true")==0) return SY_TRUE;
  else
  if (strcmp(string,"false")==0) return SY_FALSE;
  else
    return SY_FALSE;
}

SyVSkeletonSet* Sy_ManageSkeletonSet (SyVSkeletonSet *in_vss, int skelid)
{
  SyVSkeletonSet *vss;
  int diff, i, oldmaxdim;

  vss = in_vss;
  if (vss == NULL){
    vss = (SyVSkeletonSet*) Sy_Malloc(sizeof(SyVSkeletonSet));
    vss->maxdim = 0;
  }
  
  if (vss->maxdim == 0){
    vss->dim = 0;
    vss->maxdim = max(64,skelid);
    vss->is_foreground = SY_FALSE;
    vss->skel = (SyVSkeleton*) Sy_Malloc(sizeof(SyVSkeleton)*(vss->maxdim+1));
    
    for (i = 0; i <= vss->maxdim; i++){
      vss->skel[i].branchdim = vss->skel[i].nodedim = 0;
      vss->skel[i].branch = NULL;
      vss->skel[i].node = NULL;
    }

    return vss;
  }
  else if (skelid > vss->maxdim){
    oldmaxdim = vss->maxdim;
    diff = max((skelid-vss->maxdim),64);
    vss->maxdim += diff;
    vss->skel = (SyVSkeleton*) Sy_Realloc((SY_VOID*)vss->skel, 
					  sizeof(SyVSkeleton)*(vss->maxdim+1));

    for (i = oldmaxdim+1; i <= vss->maxdim; i++){
      Sy_InitVSkeleton(&vss->skel[i]);
    }
    if (skelid > vss->dim) vss->dim = skelid;
    return vss;
  }

  if (skelid > vss->dim) vss->dim = skelid;
  return vss;
}

SyVSkeleton* Sy_ManageSkeletonBranches (SyVSkeleton *in_vs, int branchid)
{
  SyVSkeleton *vs;
  int diff, oldmaxdim, i;
  
  vs = in_vs;
  if (vs == NULL){
    vs = (SyVSkeleton*) Sy_Malloc(sizeof(SyVSkeleton));
    Sy_InitVSkeleton(vs);
  }
  if (vs->maxbranchdim == 0){
    vs->maxbranchdim = max(64,branchid);
    vs->branch = (SyVSkeletonBranch*) 
      Sy_Malloc(sizeof(SyVSkeletonBranch)*(vs->maxbranchdim+1));
    
    for (i = 0; i <= vs->maxbranchdim; i++){
      Sy_InitVSkeletonBranch(&vs->branch[i]);
    }

    return vs;
  }
  else if (branchid > vs->maxbranchdim){
    oldmaxdim = vs->maxbranchdim;
    diff = max((branchid-vs->maxbranchdim),64);
    vs->maxbranchdim += diff;
    vs->branch = (SyVSkeletonBranch*) 
      Sy_Realloc((SY_VOID*)vs->branch, sizeof(SyVSkeletonBranch)
		                    *(vs->maxbranchdim+1));

    for (i = oldmaxdim+1; i <= vs->maxbranchdim; i++){
      Sy_InitVSkeletonBranch(&vs->branch[i]);
    }
    return vs;
  }
  return vs;
}

SyVSkeleton* Sy_ManageSkeletonNodes (SyVSkeleton *in_vs, int nodeid)
{
  SyVSkeleton *vs;
  int diff, oldmaxdim, i;

  vs = in_vs;
  if (vs == NULL){
    vs = (SyVSkeleton*) Sy_Malloc(sizeof(SyVSkeleton));
    Sy_InitVSkeleton(vs);
  }
  if (vs->maxnodedim == 0){
    vs->maxnodedim = max(64,nodeid);
    vs->node = (SyVSkeletonNode*) 
      Sy_Malloc(sizeof(SyVSkeletonNode)*(vs->maxnodedim+1));

    for (i = 0; i <= vs->maxnodedim; i++){
      Sy_InitVSkeletonNode(&vs->node[i]);
    }

    return vs;
  }
  else if (nodeid > vs->maxnodedim){
    oldmaxdim = vs->maxnodedim;
    diff = max((nodeid-vs->maxnodedim),64);
    vs->maxnodedim += diff;
    vs->node = (SyVSkeletonNode*) 
      Sy_Realloc((SY_VOID*)vs->node, sizeof(SyVSkeletonNode)
		                    *(vs->maxnodedim+1));

    for (i = oldmaxdim+1; i <= vs->maxnodedim; i++){
      Sy_InitVSkeletonNode(&vs->node[i]);
    }
    return vs;
  }
  return vs;
}

SyVSkeletonNode* Sy_ManageNodeData (SyVSkeletonNode *in_node, int rank)
{
  SyVSkeletonNode *cn;
  int diff;
  
  cn = in_node;

  if (cn == NULL){
    cn = (SyVSkeletonNode*) Sy_Malloc(sizeof(SyVSkeletonNode));
    Sy_InitVSkeletonNode(cn);
  }
  
  if (cn->maxrank == 0){
    cn->maxrank = max(3,rank);
    cn->inbranch = (int*)Sy_Malloc(sizeof(int)*(cn->maxrank+1));
    return cn;
  }
  else if (rank > cn->maxrank){
    diff = max((rank-cn->maxrank),3);
    cn->maxrank += diff;
    cn->inbranch = (int*)Sy_Realloc((SY_VOID*)cn->inbranch, sizeof(int)
		                    *(cn->maxrank+1));
    return cn;
  }
  return cn;
}

SyVSkeletonBranch* Sy_ManageBranchData (SyVSkeletonBranch *in_br, int elemptr)
{
  SyVSkeletonBranch *br;
  int diff;

  br = in_br;
  if (br == NULL){
    br = (SyVSkeletonBranch*) Sy_Malloc(sizeof(SyVSkeletonBranch));
    Sy_InitVSkeletonBranch(br);
  }

  if (br->maxelementdim == 0){
    br->maxelementdim = max(32,elemptr);
    br->axis = (union SyGenericPoint*)Sy_Malloc(sizeof(union SyGenericPoint)
						*(br->maxelementdim+1));
    br->radius = (float*)Sy_Malloc(sizeof(float)*(br->maxelementdim+1));
    br->ref = (SyVSkeletonRef*)Sy_Malloc(sizeof(SyVSkeletonRef)
					 *(br->maxelementdim+1));
    br->left_bndry = (SyVSkeletonBndryOffset*)
      Sy_Malloc(sizeof(SyVSkeletonBndryOffset)*(br->maxelementdim+1));
    br->right_bndry = (SyVSkeletonBndryOffset*)
      Sy_Malloc(sizeof(SyVSkeletonBndryOffset)*(br->maxelementdim+1));

    return br;
  }
  else if (elemptr > br->maxelementdim){
    diff = max((elemptr-br->maxelementdim),32);
    br->maxelementdim += diff;

    br->axis = (union SyGenericPoint*)
      Sy_Realloc((SY_VOID*)br->axis, sizeof(union SyGenericPoint)
		 *(br->maxelementdim+1));

    br->radius = (float*)
      Sy_Realloc((SY_VOID*)br->radius, sizeof(float)*(br->maxelementdim+1));

    br->ref = (SyVSkeletonRef*)
      Sy_Realloc((SY_VOID*)br->ref, sizeof(SyVSkeletonRef)*(br->maxelementdim+1));

    br->left_bndry = (SyVSkeletonBndryOffset*) 
      Sy_Realloc((SY_VOID*)br->left_bndry, 
		 sizeof(SyVSkeletonBndryOffset)*(br->maxelementdim+1));

    br->right_bndry = (SyVSkeletonBndryOffset*)
      Sy_Realloc((SY_VOID*)br->right_bndry, sizeof(SyVSkeletonBndryOffset)
		 *(br->maxelementdim+1));
    return br;
  }
  return in_br;
}

SyVSkeletonSet *Sy_ReadSkeletonSet (FILE *fp)
{ 

  SyVSkeletonSet *local_vss = NULL;
  cur_vss                   = NULL;

  if (fp == NULL) return NULL;

  /* initialization */

  Sy_AvlTreeInit(&sy_yacc_tree);
  Sy_AvlTreeInit(&sy_func_tree);

  Sy_AddIoObject(&sy_func_tree, "callback");

  Sy_AddDataSameObject(&sy_func_tree, "Sy_ReadIncList", SY_LOADFUNC_T, 
		       Sy_ReadIncList);
  Sy_AddDataSameObject(&sy_func_tree, "Sy_ReadSymmAxis", SY_LOADFUNC_T, 
		       Sy_ReadSymmAxis);
  Sy_AddDataSameObject(&sy_func_tree, "Sy_ReadRadiusList", SY_LOADFUNC_T, 
		       Sy_ReadRadiusList);
  Sy_AddDataSameObject(&sy_func_tree, "Sy_ReadReferences", SY_LOADFUNC_T, 
		       Sy_ReadReferences);
  Sy_AddDataSameObject(&sy_func_tree, "Sy_ReadBndryOffsetVectors", 
		       SY_LOADFUNC_T, Sy_ReadBndryOffsetVectors);

  Sy_AddDataSameObject(&sy_func_tree, "Sy_Load_skeletonnode", 
		       SY_LOADFUNC_T, Sy_Load_skeletonnode);

  Sy_AddDataSameObject(&sy_func_tree, "Sy_Load_skeletonbranch", 
		       SY_LOADFUNC_T, Sy_Load_skeletonbranch);

  Sy_AddDataSameObject(&sy_func_tree, "Sy_Load_skeleton", 
		       SY_LOADFUNC_T, Sy_Load_skeleton);

  Sy_AddIoObject(&sy_yacc_tree, "world");
 
  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;
  sy_yacc_iotoken.func_tree = sy_func_tree;

  yyin = fp;
  sy_yacc_iotoken.in = yyin;

  yyparse();
  fclose (yyin);   

  Sy_DeleteIoObjectTree(&sy_yacc_tree);
  Sy_DeleteIoObjectTree(&sy_func_tree);

  local_vss = cur_vss;
  cur_vss   = NULL;

  return local_vss;
}

int Sy_ReadIncList(SyIoToken *tok)
{
  SyVSkeleton        *cur_vs   = NULL;
  SyVSkeletonNode    *cur_nd   = NULL;
  SyVSkeletonBranch  *cur_br   = NULL;

  int is_sync, rank, node_id, skel_id, i;
  SyDataObject *dobj;

#ifdef SY_DEBUG
  printf ("Entering %s...\n", tok->funcname);
#endif

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "rank");
  if (dobj == NULL){
    Sy_Error("Sy_SkelIO", "Sy_ReadIncList", "missing `rank' field.\n", "");  
  } else rank = dobj->t.i_val;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "id");
  if (dobj == NULL){
    Sy_Error("Sy_SkelIO", "Sy_ReadIncList", "missing `id' field.\n", "");  
  } else node_id = dobj->t.i_val;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "skel_id");
  if (dobj == NULL){
    Sy_Error("Sy_SkelIO", "Sy_ReadIncList", "missing `skel_id' field.\n","");  
  } else skel_id = dobj->t.i_val;

  is_sync = Sy_SyncWithStream(yyin, SY_SYNC_COOKIE_BEGIN);
  
  if (is_sync){
    cur_vss = Sy_ManageSkeletonSet(cur_vss, skel_id);
    cur_vs  = &cur_vss->skel[skel_id];
    cur_vs  = Sy_ManageSkeletonNodes(cur_vs, node_id);
    cur_nd  = &cur_vs->node[node_id];
    cur_nd  = Sy_ManageNodeData(cur_nd, rank);
    
    for (i = 1; i <= rank; i++){
      fscanf (yyin, "%d", &cur_nd->inbranch[i]);
    }
    
  }/* if sync */  
  else{ return is_sync; }

  is_sync = Sy_SyncWithStream(yyin, SY_SYNC_COOKIE_END);

#ifdef SY_DEBUG
  printf ("Exiting %s...\n", tok->funcname);
#endif
  return is_sync;  
}

void Sy_GetBranchInfo (SyIoToken *tok, int *nelem, int *branch_id, 
		       int *skel_id, SyBoolean *is_binary_io, 
		       char* funcname)
{
  SyDataObject *dobj;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "elementdim");
  if (dobj == NULL){
    Sy_Error("Sy_SkelIO", funcname, "missing `elementdim' field.\n", "");  
  } else *nelem = dobj->t.i_val;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "id");
  if (dobj == NULL){
    Sy_Error("Sy_SkelIO", funcname, "missing `id' field.\n", "");  
  } else *branch_id = dobj->t.i_val;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "skel_id");
  if (dobj == NULL){
    Sy_Error("Sy_SkelIO", funcname, "missing `skel_id' field.\n","");  
  } else *skel_id = dobj->t.i_val;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "binary_io");
  if (dobj == NULL) *is_binary_io = SY_FALSE;
  else *is_binary_io = strcmp(dobj->t.s_val, "true") ? SY_FALSE : SY_TRUE; 

}

int Sy_ReadSymmAxis(SyIoToken *tok)
{
  SyVSkeleton        *cur_vs   = NULL;
  SyVSkeletonNode    *cur_nd   = NULL;
  SyVSkeletonBranch  *cur_br   = NULL;

  SyBoolean is_binary_io;
  int is_sync, nelem, branch_id, skel_id, i;
  SyDataObject *dobj;

#ifdef SY_DEBUG
  printf ("Entering %s...\n", tok->funcname);
#endif

  Sy_GetBranchInfo (tok, &nelem, &branch_id, &skel_id, &is_binary_io, 
		    "Sy_ReadSymmAxis");

  is_sync = Sy_SyncWithStream(yyin, SY_SYNC_COOKIE_BEGIN);
  
  if (is_sync){
    cur_vss = Sy_ManageSkeletonSet(cur_vss, skel_id);
    cur_vs  = &cur_vss->skel[skel_id];
    cur_vs  = Sy_ManageSkeletonBranches(cur_vs, branch_id);
    cur_br  = &cur_vs->branch[branch_id];
    cur_br  = Sy_ManageBranchData(cur_br, nelem);
    
    if (is_binary_io){
      fread (cur_br->axis, sizeof(union SyGenericPoint), nelem+1, yyin);
    }
    else {
      for (i = 1; i <= nelem; i++){
	fscanf (yyin, "%f %f", &cur_br->axis[i].rp.x, &cur_br->axis[i].rp.y);
      }
    } /* else */
    
  }/* if sync */  
  else{ return is_sync; }

  is_sync = Sy_SyncWithStream(yyin, SY_SYNC_COOKIE_END);

#ifdef SY_DEBUG
  printf ("Exiting %s...\n", tok->funcname);
#endif
  return is_sync;  
}

int Sy_ReadRadiusList(SyIoToken *tok)
{
  SyVSkeleton        *cur_vs   = NULL;
  SyVSkeletonNode    *cur_nd   = NULL;
  SyVSkeletonBranch  *cur_br   = NULL;

  SyBoolean is_binary_io;
  int is_sync, nelem, branch_id, skel_id, i;
  SyDataObject *dobj;

#ifdef SY_DEBUG
  printf ("Entering %s...\n", tok->funcname);
#endif

  Sy_GetBranchInfo (tok, &nelem, &branch_id, &skel_id, &is_binary_io,
		    "Sy_ReadRadiusList");

  is_sync = Sy_SyncWithStream(yyin, SY_SYNC_COOKIE_BEGIN);
  
  if (is_sync){
    cur_vss = Sy_ManageSkeletonSet(cur_vss, skel_id);
    cur_vs  = &cur_vss->skel[skel_id];
    cur_vs  = Sy_ManageSkeletonBranches(cur_vs, branch_id);
    cur_br  = &cur_vs->branch[branch_id];
    cur_br  = Sy_ManageBranchData(cur_br, nelem);

    if (is_binary_io){
      fread (cur_br->radius, sizeof(float), nelem+1, yyin);
    }
    else {    
      for (i = 1; i <= nelem; i++){
	fscanf (yyin, "%f", &cur_br->radius[i]);
      }
    } /* else */
    
  }/* if sync */  
  else{ return is_sync; }

  is_sync = Sy_SyncWithStream(yyin, SY_SYNC_COOKIE_END);

#ifdef SY_DEBUG
  printf ("Exiting %s...\n", tok->funcname);
#endif
  return is_sync;  
}

int Sy_ReadReferences(SyIoToken *tok)
{
  SyVSkeleton        *cur_vs   = NULL;
  SyVSkeletonNode    *cur_nd   = NULL;
  SyVSkeletonBranch  *cur_br   = NULL;

  SyBoolean is_binary_io;
  int is_sync, nelem, branch_id, skel_id, i;
  SyDataObject *dobj;

#ifdef SY_DEBUG
  printf ("Entering %s...\n", tok->funcname);
#endif

  Sy_GetBranchInfo (tok, &nelem, &branch_id, &skel_id, &is_binary_io, 
		    "Sy_ReadReferences");

  is_sync = Sy_SyncWithStream(yyin, SY_SYNC_COOKIE_BEGIN);
  
  if (is_sync){
    cur_vss = Sy_ManageSkeletonSet(cur_vss, skel_id);
    cur_vs  = &cur_vss->skel[skel_id];
    cur_vs  = Sy_ManageSkeletonBranches(cur_vs, branch_id);
    cur_br  = &cur_vs->branch[branch_id];
    cur_br  = Sy_ManageBranchData(cur_br, nelem);

    if (is_binary_io){
      fread (cur_br->ref, sizeof(SyVSkeletonRef), nelem+1, yyin);
    }
    else {    
      for (i = 1; i <= nelem; i++){
	fscanf (yyin, "%d %d %d %d", 
		&cur_br->ref[i].left_contour, &cur_br->ref[i].right_contour,
		&cur_br->ref[i].left_point, &cur_br->ref[i].right_point);
      }
    } /* else */
    
  }/* if sync */  
  else{ return is_sync; }

  is_sync = Sy_SyncWithStream(yyin, SY_SYNC_COOKIE_END);

#ifdef SY_DEBUG
  printf ("Exiting %s...\n", tok->funcname);
#endif
  return is_sync;  
}

int Sy_ReadBndryOffsetVectors(SyIoToken *tok)
{
  SyVSkeleton        *cur_vs   = NULL;
  SyVSkeletonNode    *cur_nd   = NULL;
  SyVSkeletonBranch  *cur_br   = NULL;

  SyBoolean is_binary_io;
  int is_sync, nelem, branch_id, skel_id, i;
  SyDataObject *dobj;

#ifdef SY_DEBUG
  printf ("Entering %s...\n", tok->funcname);
#endif

  Sy_GetBranchInfo (tok, &nelem, &branch_id, &skel_id, &is_binary_io,  
		    "ReadBndryOffsetVectors");

  is_sync = Sy_SyncWithStream(yyin, SY_SYNC_COOKIE_BEGIN);
  
  if (is_sync){
    cur_vss = Sy_ManageSkeletonSet(cur_vss, skel_id);
    cur_vs  = &cur_vss->skel[skel_id];
    cur_vs  = Sy_ManageSkeletonBranches(cur_vs, branch_id);
    cur_br  = &cur_vs->branch[branch_id];
    cur_br  = Sy_ManageBranchData(cur_br, nelem);

    if (is_binary_io){
      fread (cur_br->left_bndry, sizeof(SyVSkeletonBndryOffset), 
	     nelem, yyin);
      fread (cur_br->right_bndry, sizeof(SyVSkeletonBndryOffset), 
	     nelem, yyin);
    }
    else {    
      for (i = 1; i <= nelem-1; i++){
	fscanf (yyin, "%f %f", &cur_br->left_bndry[i].xoff,
	      	               &cur_br->left_bndry[i].yoff);
      }
      for (i = 1; i <= nelem-1; i++){
	fscanf (yyin, "%f %f", &cur_br->right_bndry[i].xoff,
	                       &cur_br->right_bndry[i].yoff);
      }
    } /* else */
    
  }/* if sync */  
  else{ return is_sync; }

  is_sync = Sy_SyncWithStream(yyin, SY_SYNC_COOKIE_END);

#ifdef SY_DEBUG
  printf ("Exiting %s...\n", tok->funcname);
#endif
  return is_sync;  
}


int Sy_Load_skeletonnode (SyIoToken *tok)
{
  SyVSkeleton        *cur_vs   = NULL;
  SyVSkeletonNode    *cur_nd   = NULL;
  SyVSkeletonBranch  *cur_br   = NULL;

  int rank, node_id, skel_id;
  SyBoolean is_sync, is_error = SY_FALSE;
  SyDataObject *dobj;

#ifdef SY_DEBUG
  printf ("Entering %s...\n", tok->funcname);
#endif

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "id");
  if (dobj == NULL){
    Sy_Error("Sy_SkelIO", "Sy_ReadIncList", "missing `id' field.\n", "");  
  } else node_id = dobj->t.i_val;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "skel_id");
  if (dobj == NULL){
    Sy_Error("Sy_SkelIO", "Sy_ReadIncList", "missing `skel_id' field.\n","");  
  } else skel_id = dobj->t.i_val;

  cur_vss = Sy_ManageSkeletonSet(cur_vss, skel_id);
  cur_vs  = &cur_vss->skel[skel_id];
  cur_vs  = Sy_ManageSkeletonNodes(cur_vs, node_id);
  cur_nd  = &cur_vs->node[node_id];

  cur_nd->id      = node_id;
  cur_nd->skel_id = skel_id;
  
  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "user_id");
  cur_nd->user_id = dobj != NULL ? dobj->t.i_val : 0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "hierarchy_level");
  cur_nd->hierarchy_level = dobj != NULL ? dobj->t.i_val : 0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "is_foreground");
  cur_nd->is_foreground = 
    dobj != NULL ? Sy_MapString2Bool(dobj->t.s_val): SY_TRUE;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "nodetype");
  cur_nd->ntype = 
    dobj != NULL ? Sy_MapString2NodeElementType(dobj->t.s_val) : SY_END;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "strength");
  cur_nd->strength = dobj != NULL ? dobj->t.f_val : 0.0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "radius");
  cur_nd->radius = dobj != NULL ? dobj->t.f_val : 0.0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "rank");
  cur_nd->rank = dobj != NULL ? dobj->t.i_val : 0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "xytype");
  cur_nd->etype = 
    dobj != NULL ? Sy_MapString2BranchElementType(dobj->t.s_val) :SY_NILBRANCH;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "x");
  cur_nd->loc.rp.x = dobj != NULL ? dobj->t.f_val : 0.0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "y");
  cur_nd->loc.rp.y = dobj != NULL ? dobj->t.f_val : 0.0;

#ifdef SY_DEBUG
  printf ("Exiting %s...\n", tok->funcname);
#endif

  return SY_TRUE;  
}

int Sy_Load_skeletonbranch (SyIoToken *tok)
{
  SyVSkeleton        *cur_vs   = NULL;
  SyVSkeletonNode    *cur_nd   = NULL;
  SyVSkeletonBranch  *cur_br   = NULL;

  int nelem, branch_id, skel_id;
  SyBoolean is_sync, is_error = SY_FALSE, 
            is_binary_io;
  SyDataObject *dobj;

#ifdef SY_DEBUG
  printf ("Entering %s...\n", tok->funcname);
#endif

  Sy_GetBranchInfo (tok, &nelem, &branch_id, &skel_id, &is_binary_io, 
		    "ReadBndryOffsetVectors");

  cur_vss = Sy_ManageSkeletonSet(cur_vss, skel_id);
  cur_vs  = &cur_vss->skel[skel_id];
  cur_vs  = Sy_ManageSkeletonBranches(cur_vs, branch_id);
  cur_br  = &cur_vs->branch[branch_id];
  cur_br  = Sy_ManageBranchData(cur_br, nelem);
    
  cur_br->id         = branch_id;
  cur_br->skel_id    = skel_id;
  cur_br->elementdim = nelem;
  
  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "user_id");
  cur_br->user_id = dobj != NULL ? dobj->t.i_val : 0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "hierarchy_level");
  cur_br->hierarchy_level = dobj != NULL ? dobj->t.i_val : 0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "is_foreground");
  cur_br->is_foreground = 
    dobj != NULL ? Sy_MapString2Bool(dobj->t.s_val): SY_TRUE;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "has_boundary");
  cur_br->has_boundary = 
    dobj != NULL ? Sy_MapString2Bool(dobj->t.s_val): SY_TRUE;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "strength");
  cur_br->strength = dobj != NULL ? dobj->t.f_val : 0.0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "length");
  cur_br->length = dobj != NULL ? dobj->t.f_val : 0.0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "max_strength");
  cur_br->max_strength = dobj != NULL ? dobj->t.f_val : 0.0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "min_strength");
  cur_br->min_strength = dobj != NULL ? dobj->t.f_val : 0.0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "mean_strength");
  cur_br->mean_strength = dobj != NULL ? dobj->t.f_val : 0.0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "var_strength");
  cur_br->var_strength = dobj != NULL ? dobj->t.f_val : 0.0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "max_radius");
  cur_br->max_radius = dobj != NULL ? dobj->t.f_val : 0.0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "min_radius");
  cur_br->min_radius = dobj != NULL ? dobj->t.f_val : 0.0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "mean_radius");
  cur_br->mean_radius = dobj != NULL ? dobj->t.f_val : 0.0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "var_radius");
  cur_br->var_radius = dobj != NULL ? dobj->t.f_val : 0.0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "tail_branch");
  cur_br->tail_branch = dobj != NULL ? dobj->t.i_val : 0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "head_branch");
  cur_br->head_branch = dobj != NULL ? dobj->t.i_val : 0;
  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "tail_node");
  cur_br->tail_node = dobj != NULL ? dobj->t.i_val : 0;
  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "head_node");
  cur_br->head_node = dobj != NULL ? dobj->t.i_val : 0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "axistype");
  cur_br->etype = 
    dobj != NULL ? Sy_MapString2BranchElementType(dobj->t.s_val) :SY_NILBRANCH;

#ifdef SY_DEBUG
  printf ("Exiting %s...\n", tok->funcname);
#endif

  return SY_TRUE;  
}

int Sy_Load_skeleton (SyIoToken *tok)
{
  SyVSkeleton        *cur_vs   = NULL;
  SyVSkeletonNode    *cur_nd   = NULL;
  SyVSkeletonBranch  *cur_br   = NULL;

  int skel_id;
  SyBoolean is_sync, is_error = SY_FALSE;
  SyDataObject *dobj;

#ifdef SY_DEBUG
  printf ("Entering %s...\n", tok->funcname);
#endif

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "id");
  if (dobj == NULL){
    Sy_Error("Sy_SkelIO", "Sy_Load_skeleton", "missing `id' field.\n", "");  
  } else skel_id = dobj->t.i_val;

  cur_vss = Sy_ManageSkeletonSet(cur_vss, skel_id);
  cur_vs  = &cur_vss->skel[skel_id];

  cur_vs->id = skel_id;
  
  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "user_id");
  cur_vs->user_id = dobj != NULL ? dobj->t.i_val : 0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "branchdim");
  cur_vs->branchdim = dobj != NULL ? dobj->t.i_val : 0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "nodedim");
  cur_vs->nodedim = dobj != NULL ? dobj->t.i_val : 0;

  dobj = Sy_GetDataSameObject (&tok->yacc_tree, "is_foreground");
  cur_vs->is_foreground = 
    dobj != NULL ? Sy_MapString2Bool(dobj->t.s_val): SY_TRUE;

#ifdef SY_DEBUG
  printf ("Exiting %s...\n", tok->funcname);
#endif

  return SY_TRUE;  
}

SyVSkeletonSet *Sy_GetCurSkeletonSetPtr (void)
{
  return cur_vss;
}

void Sy_ClearCurSkeletonSetPtr (void)
{
  cur_vss = NULL;
}

void Sy_WriteSkeletonSet (FILE *stream, SyVSkeletonSet *vss, 
			  int is_ascii)
{ 
  SyIOhandle *iohandle;  
  if (stream == NULL || vss == NULL) return;

  iohandle = Sy_CreateIOhandle ();
  Sy_SetIOproperty (iohandle, FORMAT, (is_ascii) ? (SY_VOID*) SY_ASCII: 
		    (SY_VOID*) SY_BINARY);
  Sy_SetIOproperty (iohandle, TYPE, (SY_VOID*) SY_SKELETON_SET);
  Sy_SetIOproperty (iohandle, DATA, (SY_VOID*) vss);
  Sy_SetIOproperty (iohandle, IO_MODE, (SY_VOID*) SY_FREE_AFTER_WRITE);
  Sy_PutIOhandle   (iohandle);
  Sy_ManageOutputStream (stream);
}





