/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * 
 * all-purpose skeleton utilities
 *
 */

static char sy_SkelUtilId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"

int Sy_IpoConnected (SyIpoDAC *ipd, int first_point, int second_point)
{
  int conn[4], conn_flag;
 
  conn[0]   = ((ipd->ipl[first_point].tail_icn == ipd->ipl[second_point].tail_icn) &&
	       (ipd->ipl[first_point].tail_icn != 0));
  
  conn[1]   = ((ipd->ipl[first_point].tail_icn == ipd->ipl[second_point].head_icn) &&
	       (ipd->ipl[first_point].tail_icn != 0));
  
  conn[2]   = ((ipd->ipl[first_point].head_icn == ipd->ipl[second_point].tail_icn) &&
	       (ipd->ipl[first_point].head_icn != 0));
  
  conn[3]   = ((ipd->ipl[first_point].head_icn == ipd->ipl[second_point].head_icn) &&
	       (ipd->ipl[first_point].head_icn != 0));
  
  conn_flag = (conn[0] || conn[1] || conn[2] || conn[3]);
  
  return(conn_flag);
}

void Sy_ResetWalktask (SyWalktask *wt)
{
  int i;
  for (i = 0; i < SY_WALK_USERFUNCTIONS; i++){
    wt->user_function[i] = (void(*)())NULL;
  }
  wt->user_file = (FILE**) NULL;
  wt->fildim    = 0;
  wt->retval    = 0;
  wt->error     = (int*) NULL;
  wt->verbose   = 0;

  wt->thres       = 0.0;
  wt->group_thres = 0.0;
  wt->narr_thres  = 0.0;
  wt->pro_thres   = 0.0;

  wt->is_loc_preset = SY_FALSE;
}

int Sy_IsResidualCut (SySkelDAC *vsk, SyIpoDAC *ipd, int edge_ref, float thres)
{
  int is_cut,
      is_border,
      is_connected;

  if (edge_ref == 0) return (0);

  if (vsk->edge[edge_ref].rank < 2) return(0);

  is_cut = vsk->is_eresid && 
    ((vsk->eresid[edge_ref].tail_residual > thres) ^
     (vsk->eresid[edge_ref].head_residual > thres));

  if (!is_cut) return (0);

  is_border = (vsk->edge[edge_ref].tail_edge == 0) || 
              (vsk->edge[edge_ref].head_edge == 0);

  is_connected = Sy_IpoConnected (ipd, vsk->edge[edge_ref].left_point,
				    vsk->edge[edge_ref].right_point);


  return (is_cut && (!is_connected || is_border));
}

void Sy_MakeTriPointref (SyTriDAC *tri)
{
  int i;

  tri->is_pointref = 1;
  tri->pref = Sy_AllocRef (tri->pdim);

  for (i = 1; i <= tri->edim; i++){
    tri->pref[tri->edge[i].tail_point] = 
    tri->pref[tri->edge[i].head_point] = i;
  }
}

int Sy_TrianConnected (SyTriDAC *tri, int first_point, int second_point)
{
  int next_edge, ref_edge, ref_point, opp_point;

  if ((tri->pref[first_point]  == 0) ||
      (tri->pref[second_point] == 0)) return(0);

  ref_edge  = tri->pref[first_point]; ref_point = first_point;
  next_edge = ref_edge;

  if (tri->edge[ref_edge].tail_point == ref_point){
    next_edge = tri->edge[ref_edge].tail_edge;
    opp_point = tri->edge[ref_edge].head_point;
  }
  else{
    next_edge = tri->edge[ref_edge].head_edge;
    opp_point = tri->edge[ref_edge].tail_point;
  }

  if (next_edge == 0){
    return (opp_point == second_point);
  }

  if (opp_point == second_point) return(1);

  while ((next_edge != ref_edge) && (next_edge != 0)){

    if (ref_point == tri->edge[next_edge].tail_point){
      opp_point = tri->edge[next_edge].head_point;
      next_edge = tri->edge[next_edge].tail_edge;
    }
    else{
      opp_point = tri->edge[next_edge].tail_point;
      next_edge = tri->edge[next_edge].head_edge;
    }
  
    if (opp_point == second_point) return(1);

  } 

  return(0);
}
  
int Sy_SkeletonType (SySkelDAC *vsk, SyIpoDAC *ipd, int current_head_vertex, 
		     int next_edge)
{
  SyVertex   v, vt, vh;
  SyIpo      p;
  SyFloatVector grad, prev_grad, succ_grad;
  SyIntVector  k1, k2, crack1, crack2;
  float    dt, dh, x1, y1;
  int      curr_point, right_point, prev_ipo, succ_ipo, tail_icn, head_icn;
  int      foreground, foreground1, foreground2, conn_flag, is_convex;

  if (vsk->edge[next_edge].rank < 1) return (foreground = 0); 

  curr_point  = vsk->edge[next_edge].left_point; 
  right_point = vsk->edge[next_edge].right_point;

  tail_icn   = ipd->ipl[curr_point].tail_icn;
  head_icn   = ipd->ipl[curr_point].head_icn;
  prev_ipo   = (ipd->icn[tail_icn].head_ipo == curr_point) ? 
                ipd->icn[tail_icn].tail_ipo : ipd->icn[tail_icn].head_ipo;
  succ_ipo   = (ipd->icn[head_icn].head_ipo == curr_point) ? 
                ipd->icn[head_icn].tail_ipo : ipd->icn[head_icn].head_ipo;

  if (prev_ipo == succ_ipo) return(foreground = 1);

  conn_flag   = Sy_IpoConnected (ipd, curr_point, right_point);

  if ((prev_ipo == 0) || (succ_ipo == 0)){
    curr_point = vsk->edge[next_edge].right_point;
    tail_icn   = ipd->ipl[curr_point].tail_icn;
    head_icn   = ipd->ipl[curr_point].head_icn;
    prev_ipo   = (ipd->icn[tail_icn].head_ipo == curr_point) ? 
                  ipd->icn[tail_icn].tail_ipo : ipd->icn[tail_icn].head_ipo;
    succ_ipo   = (ipd->icn[head_icn].head_ipo == curr_point) ? 
                  ipd->icn[head_icn].tail_ipo : ipd->icn[head_icn].head_ipo;
  }

  p  = ipd->ipo[curr_point];

  if (vsk->edge[next_edge].rank >= 2){
    if (!conn_flag){
      vt = vsk->vertex[vsk->edge[next_edge].tail_vertex];
      vh = vsk->vertex[vsk->edge[next_edge].head_vertex];
      dt = SY_SQUARE(p.x - vt.xden/vt.num)+
	SY_SQUARE(p.y - vt.yden/vt.num);
      dh = SY_SQUARE(p.x - vh.xden/vh.num)+
	SY_SQUARE(p.y - vh.yden/vh.num);
      v  = (dh > dt) ? vh : vt; 
    }
    else{
      v = vsk->vertex[current_head_vertex];
    }
  }
  else{
    v  = vsk->vertex[vsk->edge[next_edge].tail_vertex];
  }

  k1.diry = -(crack1.dirx  = p.x - ipd->ipo[prev_ipo].x);
  k1.dirx =   crack1.diry  = p.y - ipd->ipo[prev_ipo].y;
  k2.diry = -(crack2.dirx  = ipd->ipo[succ_ipo].x - p.x);
  k2.dirx =   crack2.diry  = ipd->ipo[succ_ipo].y - p.y;

  prev_grad.x  = ipd->ipl[curr_point].grad.x + ipd->ipl[prev_ipo].grad.x;
  prev_grad.y  = ipd->ipl[curr_point].grad.y + ipd->ipl[prev_ipo].grad.y;
  succ_grad.x  = ipd->ipl[curr_point].grad.x + ipd->ipl[succ_ipo].grad.x;
  succ_grad.y  = ipd->ipl[curr_point].grad.y + ipd->ipl[succ_ipo].grad.y;

  if ((k1.dirx * prev_grad.x + k1.diry * prev_grad.y) < 0){
    k1.dirx = - k1.dirx; k1.diry = - k1.diry;
  }
  if ((k2.dirx * succ_grad.x + k2.diry * succ_grad.y) < 0){
    k2.dirx = - k2.dirx; k2.diry = - k2.diry;
  }

  is_convex = ((k2.dirx - k1.dirx)*crack1.dirx + 
	       (k2.diry - k1.diry)*crack1.diry) < 0;
  foreground1 = (k1.dirx *(v.xden-v.num*p.x)) >= (k1.diry *(v.num*p.y-v.yden));
  foreground2 = (k2.dirx *(v.xden-v.num*p.x)) >= (k2.diry *(v.num*p.y-v.yden));
  
  foreground  = (is_convex) ? foreground1 && foreground2 : 
    foreground1 || foreground2;

  return(foreground);
}

/* absolute value of a */
#define ABS(a)		(((a)<0) ? -(a) : (a))

/* take binary sign of a, either -1, or 1 if >= 0 */
#define SGN(a)		(((a)<0) ? -1 : 1)

void Sy_DrawDigitalLine (int x1, int y1, int x2, int y2, 
			 void (*dotproc)(int, int))
{
  int d, x, y, ax, ay, sx, sy, dx, dy;

  dx = x2-x1;  ax = ABS(dx)<<1;  sx = SGN(dx);
  dy = y2-y1;  ay = ABS(dy)<<1;  sy = SGN(dy);

  x = x1;
  y = y1;
  if (ax>ay) {		/* x dominant */
    d = ay-(ax>>1);
    for (;;) {
      (*dotproc)(x, y);
      if (x==x2) return;
      if (d>=0) {
	y += sy;
	d -= ax;
      }
      x += sx;
      d += ay;
    }
  }
  else {			/* y dominant */
    d = ax-(ay>>1);
    for (;;) {
      (*dotproc)(x, y);
      if (y==y2) return;
      if (d>=0) {
	x += sx;
	d -= ay;
      }
      y += sy;
      d += ax;
    }
  }
}
