/*
 *  $Id: Sy_Uparse.c,v 1.1 1993/12/12 16:36:49 rlo Exp rlo $
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

 /*
  ---------------------------------------------------------------------------

   Command Line Parsing Routines 
   Robert L. Ogniewicz, 30 August 89
   
   int option (char ch) 
        
      Returns 1 if option 'ch' is part of the command line, 0 otherwise
      Example: if (option ('a')) { ...; ...; }
   
   char* argopt (char ch) 

      Returns a pointer to an (eventual) argument of option 'ch'
      Example: string_var = argopt('a');

   int uparse (argv, argc, optstring, noptarray, noptnr)
     char* argv[];
     char* noptarray[];  
     int   argc;
     char  optstring[];
     int*  noptnr;

     Performs parsing of a command line, returns 1 to indicate successful
     completion, 0 otherwise
     argv      : command line argument list
     argc      : nr of command line arguments
     optstring : string containing the options to parse for, option letters 
                 followed by a colon indicate that this option requires an
                 argument
     noptarray : array of pointers to strings holding the arguments after the
                 last option processed (possibly file names)
                 
     noptnr    : nr of elements in noptarray

  ---------------------------------------------------------------------------
*/
#include <stdio.h>
#include <assert.h>
#include "Sy_Stddef.h"
#include "Sy_Getopt.h"
#include "Sy_Uparse.h"

#define SY_OPT_DIM 256
#define SY_NIL       0

#ifndef min
#define min(x,y)  ((x)<(y) ? (x) : (y))
#endif

#ifndef max
#define max(x,y) ((x) > (y) ? (x) :  (y))
#endif

char* resarg[SY_OPT_DIM];
int   resopt[SY_OPT_DIM];

int           lresflag[SY_OPT_DIM];
Sy_CommLineOption lresopt[SY_OPT_DIM], lquery[SY_OPT_DIM];
int ldim; /* actual dimension of lquery */


int option (char ch)
{ return (resopt[(int)ch]); }

char* argopt (char ch)
{ return (resarg[(int) ch]); }

int uparse (char **argv, int argc, char *optstring, char **noptarray, 
	    int maxnopt, int *noptnr)
{
  extern char* sy_optarg;
  extern int sy_optind, sy_opterr;

  int res, i, errflag = 0;
  
  for (i = 0; i < SY_OPT_DIM; i++) resopt[i]= 0;  

  while ((res = Sy_Getopt(argc, argv, optstring)) != EOF){

    if (res == '?'){
      errflag++;
    }
    else{
      if (resopt[res]){
        fprintf (stderr,"%s: multiple occurences of option -%c,", 
	         argv[0], (char) res);
        fprintf (stderr,"superseeding \n");
      }/* if */

      resopt[res] = 1; 
      resarg[res] = sy_optarg;

      if (sy_optarg != SY_NIL){
        if (*sy_optarg == '-'){
          fprintf (stderr,"%s: wrong or missing argument of -%c -> %s \n",
		   argv[0], (char) res, sy_optarg); 
          fprintf (stderr,"%s: argument of -%c may not begin with '-' \n",
		   argv[0], (char) res);
          errflag++;
        }/* if */
      }/* if */

    }/* else */
  }/* while */

  for (i = 0; (sy_optind < argc) && (i < maxnopt); sy_optind++){
    noptarray[i++] = argv[sy_optind];
  }
  *noptnr = i;

  if (sy_optind < argc){
    fprintf (stderr, "%s: extraneous text on command line \n", argv[0]);
    fprintf (stderr, "%s: max nr of non option arguments: %d \n", argv[0], 
	     maxnopt);
    errflag++;
    for (; sy_optind < argc; sy_optind++){
      fprintf (stderr,"ignored --> %s \n", argv[sy_optind]);
    }
  }

  if (errflag) return (0); else return (1);
}   

static int optcompare (SY_CONST_VOID *a, SY_CONST_VOID *b){
  return (strcmp(((Sy_CommLineOption*)a)->name, ((Sy_CommLineOption*)b)->name));
}

int uparse_long (char **argv, int argc, char *inoptstring, 
		 SyOption *inlongopt, char **noptarray, int maxnopt, 
		 int *noptnr)
{
  extern char* sy_optarg;
  extern int sy_optind, sy_opterr;

  SyOption *lptr;
  Sy_CommLineOption *longopt;
  char *optstring;
  static char emptystring[] = "";

  int res, i, errflag = 0, lindex, indim;

  optstring = (inoptstring == (char*) NULL) ? emptystring : inoptstring;
    
  for (lptr = inlongopt, indim = 0; lptr->name; lptr++, indim++);
  longopt = (Sy_CommLineOption*) Sy_Malloc (sizeof(Sy_CommLineOption)*(indim+1));
  assert(longopt != (Sy_CommLineOption*)NULL);
  for (i = 0; i <= indim; i++){
    longopt[i].name    = inlongopt[i].name;
    longopt[i].has_arg = inlongopt[i].has_arg;
    longopt[i].flag    = (int*) NULL;
    longopt[i].val     = i+1;
  }

  for (i = 0; i < SY_OPT_DIM; i++) {
    resopt[i]= lresflag[i] = 0;  
    lresopt[i].name = (char*)NULL;
  }
  
  while(1){
    lindex = -1;
    res = Sy_GetoptLongOnly(argc, argv, optstring, longopt, &lindex); 
    if (res == EOF) break;

    if (res == '?'){
      errflag++;
    }
    else{
      if ((lindex == -1) ? resopt[res] : lresflag[lindex]){
        (lindex == -1) ? 
	  fprintf (stderr,"%s: multiple occurences of option -%c, ", 
		   argv[0], (char) res):
	  fprintf (stderr,"%s: multiple occurences of option -%s, ", 
		   argv[0], lresopt[lindex].name);
        fprintf (stderr,"superseeding... \n");
      }/* if */

      if (lindex == -1){
	resopt[res] = 1; resarg[res] = sy_optarg;}
      else{
	lresflag[lindex] = 1; lresopt[lindex] = longopt[lindex];
	lresopt[lindex].sy_optarg = sy_optarg;
      }

      if (sy_optarg != SY_NIL){
        if (*sy_optarg == '-'){
	  if (lindex == -1){ 
	    fprintf (stderr,"%s: wrong or missing argument of -%c -> %s \n",
		     argv[0], (char) res, sy_optarg); 
	    fprintf (stderr,"%s: argument of -%c may not begin with '-' \n",
		     argv[0], (char) res);
	  } else {
	    fprintf (stderr,"%s: wrong or missing argument of -%s -> %s \n",
		     argv[0], lresopt[lindex].name, sy_optarg); 
	    fprintf (stderr,"%s: argument of -%s may not begin with '-' \n",
		     argv[0], lresopt[lindex].name);
	  }
	  errflag++;
        }/* if */
      }/* if */

    }/* else */
  }/* while */

  for (i = 0; (sy_optind < argc) && (i < maxnopt); sy_optind++){
    noptarray[i++] = argv[sy_optind];
  }
  *noptnr = i;

  if (sy_optind < argc){
    fprintf (stderr, "%s: extraneous text on command line \n", argv[0]);
    fprintf (stderr, "%s: max nr of non option arguments: %d \n", argv[0], 
	     maxnopt);
    errflag++;
    for (; sy_optind < argc; sy_optind++){
      fprintf (stderr,"ignored --> %s \n", argv[sy_optind]);
    }
  }
  
  for (i = 0, ldim = 0; i < SY_OPT_DIM; i++){
    if (lresopt[i].name != (char*)NULL){
      lquery[ldim++] = lresopt[i];
    }
  }
  
  /* sort lresopt array */
  
  if (ldim > 0){
    qsort(lquery, ldim, sizeof(lquery[0]), optcompare); 
  }

  free(longopt);
  if (errflag) return (0); else return (1);
} 

int long_option (char* name)
{ 
  int upper, lower, ptr, val;

  if (ldim <= 0) return(0);
  lower = 0; upper = ldim - 1;
  do{
    ptr = (lower+upper)/2;
    val = strcmp(lquery[ptr].name, name);
    if (val == 0) return(1);
    if (lower == upper) return(0);
    else if (val > 0) upper = max(ptr-1,lower);
    else              lower = min(ptr+1,upper);
  } while (1);
    
}

char* long_argopt (char* name)
{
  int upper, lower, ptr, val;

  if (ldim <= 0) return((char*)NULL);

  lower = 0; upper = ldim - 1;
  do{
    ptr = (lower+upper)/2;
    val = strcmp(lquery[ptr].name, name);
    if (val == 0) return(lquery[ptr].sy_optarg);
    if (lower == upper) return((char*)NULL);
    else if (val > 0) upper = max(ptr-1,lower);
    else              lower = min(ptr+1,upper);
  } while (1);
}

#undef min
#undef max

#ifdef UPARSETEST

static SyOption longopt[] = {
{"threshold", SY_ARG_REQ},
{"scale", SY_ARG_REQ},
{"ascii", SY_ARG_NONE}, 
SY_NULL_OPTION};

#define MAXNOPT 2

main (int argc, char** argv)
{
  
  char* noptarray[MAXNOPT];
  float th, sc;
  int   noptnr, i;

  if ((!uparse_long (argv, argc, NULL, longopt, noptarray, 
		      MAXNOPT, &noptnr)) 
      ||
      (noptnr < 1)){
    fprintf(stderr, "Incorrect command line !\n");
    exit(1);
  }
  
  sscanf (long_option("threshold") ? long_argopt("threshold") : "2.0", "%f", 
	  &th);

  printf("threshold value is %f\n", th);

  
  sscanf (long_option("scale") ? long_argopt("scale") : "1.1", "%f", 
	  &sc);

  printf("scale value is %f\n", sc);

  if (long_option("ascii")) printf ("ascii mode set\n");

  printf ("non-option arguments:\n");
  for (i = 0; i < noptnr; i++){
    printf("<%d> %s\n", i, noptarray[i]);
  }
 
}
#endif /* UPARSETEST */











