#include <stdlib.h>
#include <stdio.h>
#include <sys/fcntl.h>
#include "solex.h"

/*
 *
 * Emulation of USLEEP
 *
 */


#if !defined(HAS_USLEEP) && (defined(SYSV) || defined(SVR4))
#include <unistd.h>
#include <sys/time.h>
#include <signal.h>

/* SIGALRM signal handler for `sleep'.  This does nothing but return,
   but SIG_IGN isn't supposed to break `pause'.  */
static void sleep_handler(int sig)
{
  return;
}

/* Sleep USECONDS microseconds, or until a previously set timer goes off.  */
int usleep(unsigned int useconds)
{
  struct itimerval otimer;
  struct timeval before, after, delta;
  void(*ohandler)();
  sigset_t omask, alarm, blockall;

#define MICRO 1000000

  if (getitimer(ITIMER_REAL, &otimer) < 0) return -1;
  sigfillset(&blockall); 
  sigprocmask(SIG_SETMASK, &blockall, &omask);
  ohandler = signal (SIGALRM, sleep_handler);
  sigrelse(SIGALRM);
  if (ohandler == SIG_ERR) return -1;

  if ((unsigned int) ((otimer.it_value.tv_sec * MICRO) +
		      otimer.it_value.tv_usec) < useconds)
    {
      /* Set up the timer.  */
      struct itimerval timer;
      timer.it_value.tv_sec = useconds / MICRO;
      timer.it_value.tv_usec = useconds % MICRO;
      timer.it_interval.tv_sec = 0;
      timer.it_interval.tv_usec = 0;

      /* Find the time beforehand so we can tell how much time elapsed.  */
      if (gettimeofday(&before, (struct timezone *) NULL) < 0)
	goto lose;

      if (setitimer(ITIMER_REAL, &timer, &otimer) < 0)
	goto lose;
    }

  /* Wait for the timer to expire.  */
  pause ();

  /* Find out what time it is now, and see how much time elapsed.  */
  if (gettimeofday(&after, (struct timezone *) NULL) < 0)
    {
      delta.tv_sec = 0;
      delta.tv_usec = 0;
    }
  else
    {
      delta.tv_sec = after.tv_sec - before.tv_sec;
      delta.tv_usec = after.tv_usec - before.tv_usec;
    }

  /* Adjust the old timer to account for the elapsed time.  */
  otimer.it_value.tv_sec -= delta.tv_sec;
  if (otimer.it_value.tv_sec < 0)
    otimer.it_value.tv_sec = 0;
  otimer.it_value.tv_usec -= delta.tv_usec; 
  if (otimer.it_value.tv_usec < 0)
    otimer.it_value.tv_usec = 0;

  /* Restore the old timer (which might have be zero).  */
  sighold(SIGALRM);
  signal (SIGALRM, ohandler);
  
  /* setitimer(ITIMER_REAL, &otimer, (struct itimerval *) NULL); */

  sigprocmask(SIG_SETMASK, &omask, NULL); 
  return useconds - ((delta.tv_sec * MICRO) + delta.tv_usec);

 lose:
  sighold(SIGALRM);
  signal (SIGALRM, ohandler);
  sigprocmask(SIG_SETMASK, &omask, NULL);
  return -1;
}

#endif /* HAS_USLEEP... */

/* 
 *  Emulation of GETRUSAGE 
 *
 */

#if !defined(HAS_GETRUSAGE) && (defined(SYSV) || defined(SVR4))
#if defined (hpux)
#include <syscall.h>
#else
#include <sys/resource.h>
#include <sys/procfs.h>
#endif

#ifdef PIOCUSAGE

/* rusage data structure */

#define	RUSAGE_SELF	0
#define	RUSAGE_CHILDREN	-1

int getrusage(int who, struct rusage *rusinfo)
{
#define CHARBUFSIZE 1024
  int             fd;
  char            proc[CHARBUFSIZE];
  prusage_t       prusage;
  prpsinfo_t      prpsinfo;

  if (who == RUSAGE_CHILDREN){
    perror("RUSAGE_CHILDREN not implemented"); exit(0);
  }

  sprintf(proc,"/proc/%d", getpid());
  if ((fd = open(proc, O_RDONLY)) == -1) { perror("open"); exit(2); }
  if (ioctl(fd, PIOCUSAGE, &prusage) == -1) { perror("ioctl"); exit(2);}
  if (ioctl(fd, PIOCPSINFO, &prpsinfo) == -1) { perror("ioctl"); exit(2);}
  
  rusinfo->ru_utime.tv_sec  = prusage.pr_utime.tv_sec;
  rusinfo->ru_utime.tv_usec = prusage.pr_utime.tv_nsec/1000.0;
  rusinfo->ru_stime.tv_sec  = prusage.pr_stime.tv_sec;
  rusinfo->ru_stime.tv_usec = prusage.pr_stime.tv_nsec/1000.0;
  rusinfo->ru_maxrss        = prpsinfo.pr_rssize;
  rusinfo->ru_idrss         = prpsinfo.pr_rssize;
  rusinfo->ru_minflt        = prusage.pr_minf;
  rusinfo->ru_majflt        = prusage.pr_majf;
  rusinfo->ru_nswap         = prusage.pr_nswap;
  rusinfo->ru_inblock       = prusage.pr_inblk;
  rusinfo->ru_oublock       = prusage.pr_oublk;
  rusinfo->ru_msgsnd        = prusage.pr_msnd;
  rusinfo->ru_msgrcv        = prusage.pr_mrcv;
  rusinfo->ru_nsignals      = prusage.pr_sigs;
  rusinfo->ru_nvcsw         = prusage.pr_vctx;
  rusinfo->ru_nivcsw        = prusage.pr_ictx;
}  

#endif /* PIOCUSAGE */  

#endif /* HAS_GETRUSAGE... */

/*
 *   Emulation of SETLINEBUF
 *
 */


#if !defined(HAS_SETLINEBUF) && (defined(SYSV) || defined(SVR4))   
int setlinebuf(FILE *iop){return setvbuf(iop, NULL, _IOLBF, 0);}
#endif /* HAS_SETLINEBUF... */

/*
 *      Emulation of GETHOSTID 
 *
 *      This example has a combination of high-level
 *      (SVR4) and (SI_HW_SERIAL) feature declarations.
 */
   
#if !defined(HAS_GETHOSTID) && (defined(SYSV) || defined(SVR4))
#if defined (sun)
#include <sys/systeminfo.h>
#elif defined (hpux)
#endif
#if defined(SI_HW_SERIAL)
long gethostid()
{
  char buf[128];                
  if (sysinfo(SI_HW_SERIAL, buf, 128) == -1) {
    perror("sysinfo");
    exit(1);
  }
  return(strtoul(buf,NULL,0));
}
#endif
#endif /* HAS_GETHOSTID... */

/*
 *      Emulation of GETDTABLESIZE
 *
 *      Several possibilites here. Note that while one
 *      can emulate getdtablesize with getrlimit on SVR4
 *      or 4.3BSD (or later), these systems should be
 *      POSIX.1 compliant, so sysconf is preferred.
 *
 */


#if !defined(HAS_GETDTABLESIZE) && (defined(SYSV) || defined(SVR4))
int getdtablesize()
{
  int tableSize;
#ifdef  _SC_OPEN_MAX	/* POSIX -- preferred */
  if ((tableSize = sysconf(_SC_OPEN_MAX)) == -1) {
    perror("sysconf");
    exit(2);
  }
#elif   RLIMIT_NOFILE	/* who is non POSIX but has this? */
  if (getrlimit(RLIMIT_NOFILE, &rlimit) == -1) {
    perror("getrlimit");
    exit(1);
  }
  tableSize = rlimit.rlim_max;
#endif  
  return tableSize;
}
#endif /* HAS_GETDTABLESIZE... */

/*
 *
 * Emulation of GETHOSTNAME 
 *
 */

#if !defined(HAS_GETHOSTNAME) && (defined(SYSV) || defined(SVR4))
#if defined (sun)
#include <sys/systeminfo.h>
#endif
int gethostname(char *name, int namelen)
{
  if (sysinfo(SI_HOSTNAME, name, namelen) <0) {
    perror("SI_HOSTNAME");
    return (-1);
  }
  return (0);
}
#endif /* HAS_GETHOSTNAME */


/*
 *
 * Emulation of GETPAGESIZE
 *
 */

#if !defined(HAS_GETPAGESIZE) && (defined(SYSV) || defined(SVR4))
#include <unistd.h>
int getpagesize (void)
{
#if defined (hpux)
  return (int)sysconf(_SC_PAGE_SIZE);
#else
  return (int)sysconf(_SC_PAGESIZE);
#endif
}
#endif




